%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Program     : Smith Waterman string matching algorithm Erlang version       %
% File        : smith.erl                                                     %
% Author      : Alexander Jean-Claude Bottema (alexb@csd.uu.se)               %
% Datum       : Sep 19 1995                                                   %
% Revidated   : Sep 19 1995                                                   %
% --------------------------------------------------------------------------- %
% Changes:                                                                    %
%    Optimization 1: Inline "match_weights"                                   %
%    Optimization 2: Guards on all int variables (explicit type information)  %
% --------------------------------------------------------------------------- %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(smith0).
-export([run/0]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% max(A,B) - The maximum value of A and B
%

% This is a bif now but I'll keep it. /S-O

maxx(A,B) when is_integer(A), is_integer(B) ->
    if A > B -> A
     ; true  -> B
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% alpha_beta_penalty(A,B) - The penalty value of A and B
%
alpha_beta_penalty(A,B) when is_integer(A), is_integer(B) -> maxx(A-4,B-1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% match_weights(A,B) - Weight function of A and B
%
% match_weights(A,B) -> if A == B -> 1 ; true -> 0 end.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate_sequence( Length, Seed ) - Generate a random sequence of length
%                                     'Length' using seed value 'Seed'.
%

generate_sequence( Length, R ) when is_integer(Length) ->
    if  Length == 0 -> []
      ; true        ->  [R rem 10 | generate_sequence( Length - 1,
						      (R * 11 + 1237501)
						      rem 10067)]
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate_sequences( No, Length, Seed )
%

generate_sequences( 0, _, _ ) -> [] ;
generate_sequences( N, Length, R ) when is_integer(N), is_integer(Length) ->
    [generate_sequence(Length, R) | generate_sequences(N-1,Length,R+1)].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Entry representation:
%
% Entry = {Left,Upper,UpperLeft,Maxx}
%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% match_entry(Top,Side,UpperLeft,Upper,Left) -
%     Match sequence entries with surrounding information
match_entry(Top,Side,UpperLeft,Upper,Left) when is_integer(Top), is_integer(Side) ->
    MeLeft = alpha_beta_penalty( element( 3, Left ), element( 1, Left ) ),
    MeUpper = alpha_beta_penalty( element( 3, Upper ), element( 2, Upper ) ),
    %
    % match weight removed
    %
    if Top == Side ->
        MeUpperLeft = maxx(MeLeft,
		           maxx(MeUpper,
			        maxx( element( 3, UpperLeft ) + 1, 0 )))
    ;   true -> MeUpperLeft = maxx(MeLeft,
		           maxx(MeUpper,
			        maxx( element( 3, UpperLeft ), 0 )))
    end,
    {MeLeft, MeUpper, MeUpperLeft,
     maxx(MeUpperLeft,
	  maxx(element(4, Left),
	       maxx(element(4, Upper), element(4, UpperLeft))))}.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% match_zero_entry( Top, Side, Left )
%
match_zero_entry( Top, Side, {Left,_,UpperLeft,Max} ) when is_integer(Top), is_integer(Side) ->
    ELeft = alpha_beta_penalty(UpperLeft, Left),
    Weight = maxx(1-abs(Side-Top),0),
    EUpperLeft = maxx(maxx(ELeft,maxx(1-abs(Side-Top),0)),0),
    EMax = maxx(maxx(Max,EUpperLeft),0),
    {ELeft, -1, EUpperLeft, EMax}.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% match(Top, Side, Prev, UpperLeft, Left )
%
match(Tops, Side, Prev, UpperLeft, Left) ->
    match0(Tops, Side, Prev, UpperLeft, Left, [], none).

match0([], _, _, _, _, Acc, Last) -> {Acc,Last} ;
match0([Top|Tops], Side, [Upper|Prev], UpperLeft, Left, Acc, Last) when
    is_integer(Top), is_integer(Side) ->
    E = match_entry(Top, Side, UpperLeft, Upper, Left),
    match0(Tops, Side, Prev, Upper, E, [E|Acc], E) ;
match0([Top|Tops], Side, none, UpperLeft, Left, Acc, Last) when
    is_integer(Top), is_integer(Side) ->
    E = match_zero_entry(Top, Side, Left ),
    match0(Tops, Side, none, UpperLeft, E, [E|Acc], E).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% match_two_seq(Side, Top, Prev)
%

match_two_seq(Side, Top, Prev) ->
    match_two_seq0(Side, Top, Prev, none).

match_two_seq0([], _, _, Result) -> Result ;
match_two_seq0([S|Side], Top, Prev, Acc) when is_integer(S) ->
    {Row,Result} = match(Top,S,Prev,{0,0,0,0},{0,0,0,0}),
    match_two_seq0(Side, Top, Row, Result).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% match_sequences(Tops, Side)
%

match_sequences(Tops, Side) ->
    match_sequences0(Tops, Side, -9999999).

match_sequences0([], _, MaxResult) -> MaxResult ;
match_sequences0([Top|Tops], Side, CrntResult) ->
%    Result = element(4, match_two_seq(Top, Side, none)),
    {_, _, _, Result} = match_two_seq(Top, Side, none),
    match_sequences0(Tops, Side, maxx(CrntResult, Result)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% test(N)
%

test(N) ->
    Tops = generate_sequences(N, 32, 1),
    Side = generate_sequence(32, 0),
    F = fun() ->
                match_sequences(Tops, Side)
        end,
    {Time, R} = timer:tc(F),
    io:format("~s : ~.2f seconds", ["Smith", Time/1.0E6]),
    {smith, R}.

%: +func run :: () -> {smith, integer()}.
run() -> 
    test(600).
