-module(measure).
% Code for performance measurements.

-compile(export_all).

modules() -> [agenda, coalesce, conn, convert, graph, match, poly,
    pos, program, reach, record_expand, rfilter, sanity, scfa_file,
    solver, walker, worklist].


configs() ->
    [{simplify, prefer_spec, true, local, 0},
     {no_stable, prefer_spec, true, local, 0},
     {no_simplify, prefer_spec, true, local, 0},
     {simplify, prefer_code, true, local, 0},
     {no_stable, prefer_code, true, local, 0},
     {no_simplify, prefer_code, true, local, 0}].

libs() ->
    [gb_trees, gb_sets, io, io_lib, ordsets, scfa_epp].


fast() ->
    AllModules = modules(),
    Config = {simplify, prefer_spec, true, no_local, 0},
    fast_run(AllModules, AllModules, Config).

fast_run([], _, _) ->
    [];
fast_run([Module|Modules], AllModules, Config) ->
    [{Module, run(Module, AllModules, Config)} |
     fast_run(Modules, AllModules, Config)].



run(Module, AllModules, {Simplify, Prefer, Verbose, Local, N}) ->
    SimpFlag =
        case Simplify of
            simplify ->
                [simplify];
            no_stable ->
                [simplify, nostable];
            no_simplify ->
                []
        end,
    PreferFlag =
        case Prefer of
            prefer_spec ->
                [prefer_spec];
            prefer_code ->
                [{prefer_code, Module}]
                %% use_spec(libs()) ++
                %% use_spec(AllModules--[Module])
        end,
    LocalFlag = 
        case Local of 
            local ->
                [local];
            no_local ->
                []
        end,
    io:format("~nFlags: ~w", [SimpFlag++PreferFlag]),
    TimeFlag = [time],
    Path = get_path(),
    {Time, {success, _}} = timer:tc(fun () -> 
                                 poly:top(Module, Path ++ ["../src"], 
                                          SimpFlag
                                          ++PreferFlag
                                          ++TimeFlag
                                          ++LocalFlag
                                          ++[verbose])
                    end),
    Text = io_lib:format("~.2f", [Time/1.0E6]),
    io:format(" Time: ~s", [Text]),
    Text.

use_spec(L) ->
    [{use_spec, F} || F<- L].

get_path() ->    
    case get(mpath) of
        undefined ->
            [];
        Path ->
            Path
    end.

run_all_configs(Module, AllModules) ->
    Configs = configs(),
    Name =  io_lib:format("~w", [Module]),
    R = run_all_configs(Module, AllModules, Configs),
    [Name|R].

run_all_configs(Module, AllModules, []) ->
    [];
run_all_configs(Module, AllModules, [Config|Configs]) ->
    [run(Module, AllModules, Config) |
     run_all_configs(Module, AllModules, Configs)].


run_all_modules() ->
    AllModules = modules(),
    run_all_modules(AllModules, AllModules).

run_all_modules([], _) ->
    [];
run_all_modules([Module|Modules], AllModules) ->
    io:format("~nChecking ~w~n", [Module]),
    [run_all_configs(Module, AllModules) |
     run_all_modules(Modules, AllModules) ].

print_row([Item|Row = [_|_]]) ->
    io:format("~s & ", [Item]),
    print_row(Row);
print_row([Item]) ->
    io:format("~s \\\\~n", [Item]).

print(Rows) ->
    io:format("~n~n", []),
    print_rows(Rows),
    io:format("~n~n", []).

print_rows([]) ->
    true;
print_rows([Row|Rows]) ->
    print_row(Row),
    print_rows(Rows).

%%%%


%%%% New version 200205. Check more modules, 

smodules() -> [{agenda,   200,  15},
               {coalesce, 220,  16},
               {conn,     181,  38},
               {convert,  815, 112},
               {graph,    137,  13},
               {match,    130,  30},
               {poly,     321,  20},
               {pos,       14,   7},
               {program,  310,  49},
               {reach,    398,  26},
               {record_expand, 88, 22},
               {rfilter, 139,   14},
               {sanity,  197,   12},
               {scfa_file, 44,   5},
               {solver,  419,   65},
               {walker,  695,  102},
               {worklist, 26,    8}].
              
emodules() -> [{barnes,   182,   4},
               {smith0,    76,   1},
               {smith,     76,  11},
               {gb_trees, 303,  30},
               {gb_sets,  512,  31},
               {ordsets,  148,  21},
               {complex,   31,  10},
               {mandelbrot,37,   8},
               {image,     59,  12},
               {render,    22,   2}].

% measure:to_latex(measure:smodules(), ["."]).
% measure:to_latex(measure:emodules(), [".", "../bench", "../bench/mandelbrot"]).

to_latex(Modules, Path) ->
    put(mpath, Path),
    Lines = to_latex_run_all(Modules),
    latex_lines(Lines).

to_latex_run_all(Modules) ->
    [{Module, LOC, LOD, latex_time(Module)} || {Module, LOC, LOD} <- Modules].

latex_time(Module) ->
    Flags = [simplify, prefer_spec],
    Path = get_path(),
    {Time, {success, _}} = 
        timer:tc(fun () -> 
                         poly:top(Module, Path, Flags)
                 end),
    Text = io_lib:format("~.2f", [Time/1.0E6]),
    io:format(" Time: ~s", [Text]),
    Time/1.0E6.

latex_lines([]) ->
    true;
latex_lines([{Module, LOC, LOD, Time}|Lines]) ->
    io:format("~n~w& ~w& ~w& ~.2f\\\\", [Module, LOC, LOD, Time]),
    latex_lines(Lines).
