-module(agenda).

%: +type item() 
%:   = {top, fnid()} 
%:   + {def, fnid()} 
%:   + {spec, fnid(), dir()} 
%:  + {typedef, tyid(), dir()} 
%:  + {callee, fnid()}.
%: +type agenda() = gb_sets:tree(item()).
%: +type fnid() = {fnid, atom(), atom(), integer()}.
%: +type tyid() = {tyid, atom(), atom()}.
%: +type dir() = walker:dir().
%: +type list(X) = erlang:list(X).

-export([top/2, top_list/2, to_list/1]).

%: +func top :: fnid() * program:program() -> agenda().                  
top(Fnid, P) ->
    A = agenda_new(),
    case add_top(Fnid, A) of
	{added, A1} ->
	    A2 = def(Fnid, A1, P),
	    spec(Fnid, upper, A2, P);
	already_present ->
	    A
    end.

%: +func top_list :: list(fnid()) * program:program() -> agenda().                  
top_list(FnidList, P) ->
    A = agenda_new(),
    A1 = top_list(FnidList, P, A),
    A1.

top_list([], _, A) ->
    A;
top_list([Fnid|FnidList], P, A) ->
    case add_top(Fnid, A) of
	{added, A1} ->
	    A2 = def(Fnid, A1, P),
	    A3 = spec(Fnid, upper, A2, P);
	already_present ->
	    A3 = A
    end,
    top_list(FnidList, P, A3).
    

% %: +func def :: fnid() * agenda() * program:program() -> agenda().
def(Fnid, A, P) ->
   case add_def(Fnid, A) of 
       {added, A1} ->                       
          {def, _, Fnid, _, E}  = program:def(Fnid, P),
          exp(E, A1, P);
       already_present ->
          A
   end.

spec(Fnid, Dir, A, P) ->
   case add_spec(Fnid, Dir, A) of
       {added, A1} ->
	   case program:spec(Fnid, P) of
	       {spec, _, Fnid, TE} ->
		   type_exp(TE, Dir, A1, P);
	       {spec_bounded, _, Fnid, _, TE} ->
		   type_exp(TE, Dir, A1, P)
	   end;
       already_present ->
           A
   end.

typedef(Tyid, Dir, A, P) ->
    case add_typedef(Tyid, Dir, A) of
	{added, A1} ->
	    {typedef, _, _, _, TExp} = program:typedef(Tyid, P),
	    type_exp(TExp, Dir, A1, P);
	already_present ->
	    A
    end.

callee(Fnid = {fnid, M, _, _}, A, P) ->
    P1 = program:ensure_loaded(M, P),
    case add_callee(Fnid, A) of
	{added, A1} ->
	    case program:def_or_spec(Fnid, P1) of
		def ->
		    def(Fnid, A1, P1);
		spec ->
		    spec(Fnid, lower, A1, P1)
	    end;	   
	already_present ->
	    A
    end.



%: +func exp :: convert:exp() * agenda() * program:program() -> agenda().
exp({var, _}, A, _P) ->
    A;
exp({c, _C, ExpList},  A, P) ->
    exp_list(ExpList, A, P);
exp({op, _,  _, E1, E2}, A, P) ->
    exp_list([E1, E2], A, P);
exp({op,  _, E1, E2}, A, P) ->
    exp_list([E1, E2], A, P);
exp({eq, E1, E2}, A, P) ->
    exp_list([E1, E2], A, P);
exp({match, _, _, Exp}, A, P) ->
    exp(Exp, A, P);
exp({ccase, _Line, KeyList, ClauseList, _}, A, P) ->
    A1 = exp_list(KeyList, A, P),
    clause_list(ClauseList, A1, P);
exp({lambda, _VarList, Exp}, A, P) ->
    exp(Exp, A, P);
exp({lfun, _, Fnid}, A, P) ->
    callee(Fnid, A, P);
exp({local_call, _, Fnid, ExpList}, A, P) ->
    A1 = exp_list(ExpList, A, P),
    callee(Fnid, A1, P);
exp({call, _, Fnid, ExpList}, A, P) ->
    A1 = exp_list(ExpList, A, P),
    callee(Fnid, A1, P);
exp({apply, _, Exp, ExpList}, A, P) ->
    exp_list([Exp|ExpList], A, P);
exp({lc, _, Exp, GenList}, A, P) ->
    A1 = exp(Exp, A, P),
    gen_list(GenList, A1, P);
exp({promise, _, Exp, TExp}, A, P) ->
    A1 = exp(Exp, A, P),
    type_exp(TExp, lower, A1, P);
exp({block, _, Block}, A, P) ->
    exp_list(Block, A, P);
exp({'try', _, Block, CRClauseList, ClauseList, After}, A, P) ->
    A1 = exp_list(Block, A, P),
    A2 = clause_list(CRClauseList, A1, P),
    A3 = clause_list(ClauseList, A2, P),
    exp_list(After, A3, P);
exp({'catch', _, E}, A, P) ->
    exp(E, A, P);
exp({'receive', _, ClauseList, _}, A, P) ->
    clause_list(ClauseList, A, P);
exp({'receive', _, ClauseList, _, After, Final}, A, P) ->
    A1 = clause_list(ClauseList, A, P),
    A2 = exp(After, A1, P),
    exp_list(Final, A2, P);
exp({check, E}, A, P) ->
    exp(E, A, P).




exp_list([], A, _) ->
    A;
exp_list([E|EList], A, P) ->
    A1 = exp(E, A, P),
    exp_list(EList, A1, P).

gen_list([Gen|GenList], A, P) ->
    A1 = gen(Gen, A, P),
    gen_list(GenList, A1, P);
gen_list([], A, _) ->
    A.

gen({generate, _, _, Exp}, A, P) ->
    exp(Exp, A, P);
gen({b_generate, _, _, Exp}, A, P) ->
    exp(Exp, A, P);
gen(Exp, A, P) ->
    exp(Exp, A, P).


clause_list([], A, _) ->
    A;
clause_list([Clause|ClauseList], A, P) ->
    A1 = clause(Clause, A, P),
    clause_list(ClauseList, A1, P).

clause({clause, _, _PairList, Body}, A, P) ->
    exp_list(Body, A, P).

type_exp({var, _}, _, A, _) ->
    A;

type_exp({alias, _, Var, E}, Dir, A, P) ->
    A1 = type_exp(Var, Dir, A, P),
    type_exp(E, Dir, A1, P);

%% type_exp({c, {fun_fixed, _N}, [E| EList]}, Dir, A, P) ->
%%     A1 = type_exp_list(EList, walker:reverse(Dir), A, P),
%%     type_exp(E, Dir, A1, P);

type_exp({c, C, TExpList}, Dir, A, P) ->
    {CoList, CoVarList} = conn:variance(C, TExpList),
    DirR = walker:reverse(Dir),
    A1 = type_exp_list(CoList, Dir, A, P),
    type_exp_list(CoVarList, DirR, A1, P);
type_exp({number, _, TE}, Dir, A, P) ->
    type_exp(TE, Dir, A, P);
type_exp({union, _, Exp1, Exp2}, Dir, A, P) ->
    A1 = type_exp(Exp1, Dir, A, P),
    type_exp(Exp2, Dir, A1, P);
type_exp({any}, _, A, _) ->
    A;
type_exp({none}, _, A, _) ->
    A;
type_exp({type_call, _, Tyid, _, TEList}, Dir, A, P) ->
    A1 = type_exp_list(TEList, Dir, A, P),
    typedef(Tyid, Dir, A1, P);
type_exp({type_extern_call, _, Tyid, TEList}, Dir, A, P) ->
    A1 = type_exp_list(TEList, Dir, A, P),
    typedef(Tyid, Dir, A1, P);
type_exp({watch, _, Exp}, Dir, A, P) ->
    type_exp(Exp, Dir, A, P);
type_exp({'when', _, TE, Constraint}, Dir, A, P) ->
    A1 = type_exp(TE, Dir, A, P),
    constraint(Constraint, A1, P).

constraint({'=<', _, TE1, TE2}, A, P) ->
    A1 = type_exp(TE1, lower, A, P),
    type_exp(TE2, upper, A1, P).

type_exp_list([], _, A, _) ->
    A;
type_exp_list([E|EList], Dir, A, P) ->
    A1 = type_exp(E, Dir, A, P),
    type_exp_list(EList, Dir, A1, P).


%% is_primitive_type(atom, 0) ->
%%     {value, {c, atom, []}};
%% is_primitive_type(float, 0) ->
%%     {value, {c, float, []}};
%% is_primitive_type(integer, 0) ->
%%     {value, {c, integer, []}};
%% is_primitive_type(tuple, 0) ->
%%     {value, {c, tuple, []}};
%% is_primitive_type(_, _) ->
%%     none.

agenda_new() ->
    gb_sets:empty().

add(Item, A) ->
    case gb_sets:is_member(Item, A) of
	true ->
	    already_present;
	false ->
	    A1 = gb_sets:add(Item, A),
	    {added, A1}
    end.

add_top(Fnid, A) ->
    add({top, Fnid}, A).

add_def(Fnid, A) ->
    add({def, Fnid}, A).

add_spec(Fnid, Dir, A) ->
    add({spec, Fnid, Dir}, A).

add_typedef(Tyid, Dir, A) ->
    add({typedef, Tyid, Dir}, A).

add_callee(Fnid, A) ->
    add({callee, Fnid}, A).

%: +func to_list :: agenda() -> gb_trees:list(item()).		      
to_list(A) ->
    gb_sets:to_list(A).

%% reverse(upper) ->
%%     lower;
%% reverse(lower) ->
%%     upper.
