-module(coalesce).

-export([coalesce/1]).

% Taking a gb_tree of predicates, coalesce disconvers sets of
% predicates that form strongly connected components and outputs a
% list of predicates where every predicate only refers to predicates
% earlier in the list.

%: +type list(X) = erlang:list(X).
%: +type predicate() = walker:predicate().
%: +type pname() = walker:pname().
%: +type adjust() = gb_trees:tree(pname(), {pname(), integer(), integer()}).
%: +type env() = gb_trees:tree(pname(), predicate()).

%: +func coalesce ::  env() -> list(predicate()).
coalesce(Predicates) ->
    Graph = make_graph(Predicates),
    Cycles = lists:reverse(graph:cycles(Graph)),
    Dag = cycles_to_dag(Cycles, Predicates, gb_trees:empty()),
    Dag1 = poly:shorten_predicate_list(Dag),
    Dag1.

%%%% Make graph

%: +func make_graph :: env() -> graph:graph(pname()).
make_graph(Predicates) ->
    G = graph:empty_graph(),
    insert_predicates(gb_trees:values(Predicates), G).

insert_predicates([], G)   ->
    G;
insert_predicates([D|L], G)->
    G1 = insert_predicate(D, G),
    insert_predicates(L, G1).

%: +func insert_predicate :: predicate() * graph:graph(pname()) 
%:                           -> graph:graph(pname()).
insert_predicate({predicate, Name, _, Phi}, G)     ->  
    Children = [Name|children(Phi)],
    G1 = graph:new_node(Name, G),
    insert_children(Name, Children, G1).

children(Phi)->
    children(Phi, []).

children([], Acc)->
    Acc;
children([Phi1|Phi2], Acc) ->
    children(Phi1, children(Phi2, Acc));
children({include, _, _, _}, Acc)->
    Acc;
children({call, Name, _}, Acc)->
    [Name|Acc];
children({watch, _, _}, Acc)  ->
    Acc.

insert_children(_, [], G)  ->
    G;
insert_children(Parent, [Child|Children], G) ->
    G1 = graph:insert_son(Parent, Child, G),
    insert_children(Parent, Children, G1).

%%%% Build dag

%: +func cycles_to_dag :: list(list(pname())) * env() * adjust() 
%:                     -> list(predicate()).
cycles_to_dag([Cycle|RCycles], Predicates, Adjust)     ->
    {Phi, S} = merge_cycle(Cycle, Predicates),
    Phi1 = adjust(Phi, Adjust), 
    {Adjust1, Name} = update_adjust(Adjust, Predicates, Cycle),
    ArgList = new_arg_list(Cycle, S, Predicates),
    Predicate = {predicate, Name, ArgList, Phi1},
    Predicates1 = remove_cycle(Cycle, Predicates),
    [Predicate | cycles_to_dag(RCycles, Predicates1, Adjust1)];
cycles_to_dag([], _, _)    ->
    [].

remove_cycle([_|Cycle], Predicates) ->
   remove_all(Cycle, Predicates).

remove_all([], Predicates) ->
   Predicates;
remove_all([Name|Cycle], Predicates) ->
   remove_all(Cycle, gb_trees:delete(Name, Predicates)).


%: +func merge_cycle :: list(pname()) * env() 
%:                   -> {walker:constraints(), synonyms()}.
merge_cycle(Cycle, Predicates) ->
    Predicates1 = [gb_trees:get(Name, Predicates) || Name <- Cycle], 
    CS = gb_sets:from_list(Cycle),
    {Phi, S} = merge_predicates(Predicates1, CS),
    Phi1 = drop_cycle_calls(Phi, CS),
    {[Phi1], S}.

drop_cycle_calls([], _)->
    [];
drop_cycle_calls([Phi1|Phi2], CS) ->
    [drop_cycle_calls(Phi1, CS) |
     drop_cycle_calls(Phi2, CS)]; 
drop_cycle_calls(Phi={include, _, _, _}, _CS)   ->
   Phi;
drop_cycle_calls(Phi={call, Name, _}, CS)->
    case gb_sets:is_element(Name, CS) of
	true ->
	    [];
	false ->
	    Phi
    end;

drop_cycle_calls(Phi={watch, _, _}, _CS)->
    Phi.

   
merge_predicates(PList, CS)->
    S = synonyms(PList, CS, synonyms_empty()),
    G = transform(PList, S),
    {lists:flatten(G), S}.


new_arg_list([Name|Cycle], S, Predicates)->
    {predicate, P, ArgList, _} = gb_trees:get(Name, Predicates),
    NewArgList = [deref(global(Arg, P), S) || Arg <- ArgList],    
    NewArgList ++ new_arg_list(Cycle, S, Predicates);
new_arg_list([], _, _) ->
    [].

% adjust. Change calls to reflect merging of cycles.
adjust(Phi, Adjust) ->
    {Psi, _} = adjust_1(Phi, Adjust, 0),
    [Psi].

adjust_1([], _, P)    ->
    {[], P};
adjust_1([Phi1|Phi2], Adjust, P)  ->
    {Psi1, P1} = adjust_1(Phi1, Adjust, P),
    {Psi2, P2} = adjust_1(Phi2, Adjust, P1),
    {[Psi1 | Psi2], P2};
adjust_1(Phi = {include, _, _, _}, _, P) ->
    {Phi, P};
adjust_1({call, Name, ArgList}, Adjust, P)   ->
    {Name1, Offset, Size} = gb_trees:get(Name, Adjust),
    {ArgList1, P1} = adjust_arg_list(ArgList, Offset, Size, P),
    {{call, Name1, ArgList1}, P1};
adjust_1(Phi = {watch, _, _}, _, P) ->
    {Phi, P}.

adjust_arg_list(ArgList, Offset, Size, P)  ->
    Left = new_temps(Offset-1, P),
    Right = new_temps(Size-(Offset-1)-length(ArgList), P+Offset),
    {Left++ArgList++Right, P+Size}.

update_adjust(Adjust, Predicates, Cycle)->
    [NewName|_] = Cycle,
    Size = lists:sum(
	     [length(predicate_arg_list(gb_trees:get(Name, Predicates))) ||
		 Name <- Cycle]),
    update_offsets(Cycle, Predicates, NewName, 1, Size, Adjust).


predicate_arg_list({predicate, _, ArgList, _})     ->
    ArgList.


update_offsets([], _, NewName, _, _, Adjust) ->
    {Adjust, NewName};
update_offsets([Name|NameList], Predicates, NewName, Offset, Size, Adjust) ->
    Adjust1 = gb_trees:insert(Name, {NewName, Offset, Size}, Adjust),
    {predicate, _, ArgList, _} = gb_trees:get(Name, Predicates),
    update_offsets(NameList, Predicates, NewName, Offset+length(ArgList), 
                   Size, Adjust1).



% synonyms. When merging a cycle.

global(zero, _)->
    zero;
global(Name, Pred) ->
    list_to_atom("global"++walker:name_to_string(Name)++ atom_to_list(Pred)).

%: +type synonyms() = gb_trees:tree(walker:name(), walker:name()).
%: +func synonyms :: list(predicate()) * gb_sets:tree(pname()) * synonyms() 
%:                -> synonyms().
                          
synonyms([], _, S)  ->
    S;
synonyms([{predicate, Name, ArgList, G}| Predicates], CS, S)     ->
    S1 = synonyms_args(ArgList, 1, Name, Name, S),
    S2 = synonyms_constraint(G, Name, CS, S1),
    synonyms(Predicates, CS, S2).

synonyms_args([], _, _, _, S) ->
    S;
synonyms_args([Name|EList], N, Dest, Source, S)    ->
    S1 = synonyms_add(global(Name, Source), arg(Dest, N), S),
    synonyms_args(EList, N+1, Dest, Source, S1).

synonyms_constraint([], _, _, S) ->
    S;
synonyms_constraint([G1|G2], Pred, CS, S)   ->
    S1 = synonyms_constraint(G1, Pred, CS, S),
    synonyms_constraint(G2, Pred, CS, S1);
synonyms_constraint({include, _, _, _}, _, _, S)   ->
    S;                              
synonyms_constraint({call, Other, EList}, Pred, CS, S) ->  
    case gb_sets:is_element(Other, CS) of
	true ->
	    synonyms_args(EList, 1, Other, Pred, S);
	false ->
	    S
    end;
synonyms_constraint({watch, _, _}, _, _, S)-> 
    S.

% Synonyms is a gb_tree. Of course. 

synonyms_empty() ->
    gb_trees:empty().


synonyms_add(N1, N2, S)->
    N1d = deref(N1, S),
    N2d = deref(N2, S),
    if N1d == N2d ->
         S;
       N1d < N2d ->
         gb_trees:insert(N2d, N1d, S);
       N2d < N1d ->
         gb_trees:insert(N1d, N2d, S)
    end.

%% So this solution could allow long chains to form (which might
%% affect complexity). But I feel confident it will work, so I'll use
%% it for now.

deref(N, S)  ->
    case gb_trees:lookup(N, S) of
        {value, N1} ->
          deref(N1, S);
        none ->
          N
    end.


%% Transform. Change variables to cycle-wide names.

transform([], _) ->
    [];
transform([{predicate, Pred, _, G} | Predicates], S)   ->
    G1 = transform_constraint(G, Pred, S),
    G2 = transform(Predicates, S), 
    [G1|G2].

transform_constraint([], _, _)->
    [];
transform_constraint([G1| G2], Pred, S) ->
    G1t = transform_constraint(G1, Pred, S),                      
    G2t = transform_constraint(G2, Pred, S),
    [G1t|G2t];
transform_constraint({include, E1, E2, Source}, Pred, S)   ->
    E1t = transform_exp(E1, Pred, S),
    E2t = transform_exp0(E2, Pred, S),
    {include, E1t, E2t, Source};
transform_constraint({call, Name, ArgList}, Pred, S)   ->
    NewArgList = [deref(global(Arg, Pred), S) || Arg <- ArgList],    
    {call, Name, NewArgList};
transform_constraint({watch, Name, M}, Pred, S)    -> 
    Nt = transform_name(Name, Pred, S),
    {watch, Nt, M}.

transform_exp({var, Name}, Pred, S) ->
    N = deref(global(Name, Pred), S), 
    {var, N};
transform_exp({c, C, EList}, Pred, S)   ->
    EListT = transform_exp0_list(EList, Pred, S),
    {c, C, EListT};
transform_exp({sift, N, F}, Pred, S)->
    Nt = transform_name(N, Pred, S),
    {sift, Nt, F}.

transform_exp0({var, Name}, Pred, S) ->
    N = deref(global(Name, Pred), S), 
    {var, N};
transform_exp0({c, C, EList}, Pred, S)   ->
    EListT = transform_exp0_list(EList, Pred, S),
    {c, C, EListT}.


transform_exp0_list(EList, Pred, S)->
    [transform_exp0(E, Pred, S) || E <- EList].


transform_name(Name, Pred, S) ->
    deref(global(Name, Pred), S).


% position

new_temps(0, _)  ->
    [];
new_temps(N, P) when N>0 ->
    [new_temp(P) | new_temps(N-1, P+1)].

new_temp(P) ->
    list_to_atom(integer_to_list(P)).

arg(Dest, N) ->
    list_to_atom("arg "++atom_to_list(Dest)++integer_to_list(N)).
