-module(convert).

-export([fun_def/3, type_def/2, type_def_external/2]).
-export([type_exp/2]).
-include_lib("stdlib/include/assert.hrl").
%: +type exp() = 
%:          {var, name()}
%:        + {c, econ(), list(exp())}
%:        + {match, line(), pattern(), exp()}
%:        + {check, exp()}
%:        + {lambda, list(name()), exp()}
%:        + {remote, line(), exp(), exp()}
%:        + {ccase, line(), list(exp()), 
%:                  list(cclause()), list(name())}
%:        + {call, line(), fnid(), list(exp())} 
%:        + {local_call, line(), fnid(), list(exp())}
%:        + {extern_call, line(), fnid(), list(exp())}
%:        + {apply, line(), exp(), list(exp())}
%:        + {lc, line(), exp(), list(lc_gen())}
%:        + {promise, line(), exp(), texp()}
%:        + {block, line(), list(exp())}
%:        + {'catch', line(), exp()}
%:        + {'try', line(), list(exp()), 
%:                          list(cclause()), 
%:                          list(cclause()), list(exp())}
%:        + {lfun, line(), fnid()}
%:        + {'receive', line(), list(cclause()), list(name())}
%:        + {'receive', line(), list(cclause()), list(name()), exp(), list(exp())}.

%: +type lc_gen() 
%:    = {generate, line(), pattern(), exp()}
%:    + {b_generate, line(), pattern(), exp()}
%:    + exp().


%: +type texp() = 
%:    {var, name()} 
%:  + {c, con(), list(texp())} 
%:  + {alias, line(), texp(), texp()}
%:  + {number, line(), texp()}
%:  + {union, line(), texp(), texp()}
%:  + {any} 
%:  + {none}
%:  + {type_call, line(), tyid(), local+extern, list(texp())}
%:  + {watch, line(), texp()}.

%: +type pattern() = 
%:       {empty, list(name())} 
%:     + {c, con(), list(pattern()), list(name())}
%:     + unmatchable.
%: +type fnid() = {fnid, atom(), atom(), integer()}.
%: +type tyid() = {tyid, atom(), atom()}.
%: +type con() = conn:con().
%: +type econ() = {integer_fixed, integer()} + {float_fixed, float()} + con().
%: +type name() =
%:     atom().
%: +type var() = {var, name()}.
%: +type line() = program:line().
%: +type head() = list(pattern()).
%: +type guard() = list(exp()).
%: +type cclause() = {clause, line(), list({head(), guard()}), list(exp())}.

%: +type env() = {env, atom(), record_env(), gb_trees:tree(name(), name())}.
%: +type record_env() = gb_trees:tree(program:rid(), program:recorddef()).
%: +type bind() = gb_trees:tree(name(), name()).
%: +type match() = {check, exp()}.

%: +type list(X) = erlang:list(X).

%: +type pos() = pos:pos().


% Guards

check_eq(Line, E1, E2) -> 
    [{check, {call, Line, {fnid, 'Tintsm', '=:=', 2}, [E1, E2]}}].

check_eq_integer(Line, Name, N) ->
    [{check, {call, Line, {fnid, 'Tintsm', '=:=', 2}, 
	      [Name, {c, {integer_fixed, N}, []}]}}].

% functions

%: +func fun_def :: absyn:form()
%:                * atom() 
%:                * record_env()
%:               -> !{def, !absyn:line(), !fnid(), !list(exp()), !exp()}.

fun_def({function, Line, Name, Arity, Clauses}, Module, Records) ->
    ArgList = invent_arg_list(Arity),
    Env = env_new(Module, Records),
    {RClauses, BKeyList}  = clause_list(Clauses,  pos:pos_zero(), Env),
    Fnid = {fnid, Module, Name, Arity},
    Body = {ccase, Line, ArgList, RClauses, BKeyList},
    {def, Line, Fnid, ArgList, Body}.

%: +func clause_list :: list(absyn:clause()) * pos:pos() * env()
%:                   -> {list(cclause()), list(name())}.
 
clause_list([Clause|ClauseList], Pos, Env) ->
    {RC, BKeyList1} = clause(Clause, pos:left(Pos), Env),
    case ClauseList of
        [] ->
	    {[RC], BKeyList1};
        ClauseList ->
	    {RCList, BKeyList2} = 
		clause_list(ClauseList, pos:right(Pos), Env),	    
	    {[RC|RCList], bkeylist_intersect(BKeyList1, BKeyList2)}
    end.

%: +func clause :: absyn:clause() * pos:pos() * env() 
%:              -> {cclause(), list(name())}.

clause({clause,Line, Head, Guard, Body}, Pos, Env) ->
    {PairList, B1} = 
        split_head_guard(Head, Guard, Env, pos:left(pos:left(Pos))),    
    Env1 = env_add_binding(Env, B1),    
    {Body1, B3} = body(Body, Env1, pos:right(Pos)),
    Clause = {clause, Line, PairList, Body1},
    {Clause, binding_to_keylist(B3)}.


invent_arg_list(N) ->
%    [{var, {arg, I}} || I <- lists:seq(1, N)].
    [{var, make_arg(I)} || I <- lists:seq(1, N)].

%: +func body :: list(absyn:exp()) * env() * pos() -> !{!list(!exp()), !bind()}.
body(Body, Env, Pos) ->
    {BR, B} = block(Body, Env, Pos),
    {BR, B}.

block([M={match, _, _, _},
      P={promise, _, _, _, _}|Body], Env, Pos) ->
    promise(M, P, Body, Env, Pos);
block([Exp|ExpList], Env, Pos) -> 
    {RExp, B1} = exp(Exp, Env, pos:left(Pos)),
    Env1 = env_add_binding(Env, B1),
    {RExpList, B2} = block(ExpList, Env1, pos:right(Pos)),
    B =  binding_union(B1, B2),
    {[RExp|RExpList], B};
block([], _, _) -> 
    {[], binding_empty()}.

promise(Match = {match, _, {var, _, Name}, _},
	{promise, _, {atom,Line,promise}, {var,_,Name1}, TE},
	Block, Env, Pos) -> 
    ?assert(Name==Name1),
    {RExp, B1} = exp(Match, Env, pos:left(pos:left(Pos))),
    Module = env_module(Env),
    RTE = type_exp(TE, Module, pos:right(pos:left(Pos))),
    Promise = {promise, Line, RExp, RTE},
    Env1 = env_add_binding(Env, B1),
    {RExpList, B2} = block(Block, Env1, pos:right(Pos)), 
    B = binding_union(B1, B2),
    {[Promise|RExpList], B}.

%: +func exp :: absyn:exp() * env() * pos() -> !{!exp(), !bind()}.
exp({var, Line, Name}, Env, _Pos) ->
    case env_lookup(Name, Env) of
	none ->
            Message = io_lib:format("Unbound variable at line ~w", [Line]),
	    error(lists:flatten(Message));
        {value, Name1} ->
	   {{var, Name1}, binding_empty()}
    end;
exp({nil, _}, _, _) ->
    {{c ,conn:nil(), []}, binding_empty()};
exp({atom, _, Name}, _, _) ->
    {{c, conn:atom_fixed(Name), []}, binding_empty()};
exp({integer, _, N}, _, _) ->
    {{c, {integer_fixed,N}, []}, binding_empty()};
exp({char, _, N}, _, _) ->
    {{c, {integer_fixed, N}, []}, binding_empty()};
exp({float, _, F}, _, _) ->
    {{c, {float_fixed, F}, []}, binding_empty()};
exp({cons, _, E1, E2}, Env, Pos) ->
    {RExp1, B1} = exp(E1, Env, pos:left(Pos)),
    {RExp2, B2} = exp(E2, Env, pos:right(Pos)),
    B = binding_union(B1, B2),
    {{c, conn:cons(), [RExp1, RExp2]}, B};
exp({tuple, _, [{atom, _, Name}|EList]}, Env, Pos) ->
    {REList, B} = exp_list(EList, Env, Pos),
    {{c, conn:tuple_tagged(Name, length(EList)+1), REList}, B};
exp({tuple, _, EList}, Env, Pos) ->
    {REList, B} = exp_list(EList, Env, Pos),
    {{c, conn:tuple_fixed(length(EList)), REList}, B};
exp({string, _, Data}, _, _) ->
    {exp_string(Data), binding_empty()};
exp({bin, Line, BinElementList}, Env, Pos) ->
    {L, B} = bin_element_list(BinElementList, Env, Pos),
    {{call, Line, {fnid, 'Tintsm', '$make_binary', 1}, L}, B};
exp({op, Line, OP, E1}, Env, Pos) ->
    {RExp1, B1} = exp(E1, Env, Pos),
    {{call, Line, {fnid, 'Tintsm', OP, 1}, [RExp1]}, B1};
exp({op, Line, OP, E1, E2}, Env, Pos) ->
    {RExp1, B1} = exp(E1, Env, pos:left(Pos)),
    {RExp2, B2} = exp(E2, Env, pos:right(Pos)),
    B = binding_union(B1, B2),
    {{call, Line, {fnid, 'Tintsm', OP, 2}, [RExp1, RExp2]}, B};
exp({match, Line, P, E}, Env, Pos) ->
    {RE, B1} = exp(E, Env, pos:right(Pos)),
    {RP, B2, _} = pattern(P, env_add_binding(Env, B1), pos:left(Pos)),
    B = binding_union(B1, B2),
    {{match, Line, RP, RE}, B};
exp({'case', Line, E, ClauseList}, Env, Pos) ->
    {RE, B1} = exp(E, Env, pos:left(pos:left(Pos))),
    Env1 = env_add_binding(Env, B1),
    {RClauseList, BList} = 
        cclause_list(ClauseList, Env1, pos:right(pos:left(Pos))),
    {RClauseList1, BNew} = 
        make_clauses_uniform(RClauseList, BList, pos:right(Pos)),
    B = binding_union(B1, BNew),
    {{ccase, Line, [RE], RClauseList1, gb_trees:values(BNew)}, B};
exp({'if', Line, ClauseList}, Env, Pos) ->
    {RClauseList, BList} = 
        cclause_list(ClauseList, Env, pos:left(Pos)),
    {RClauseList1, BNew} = 
        make_clauses_uniform(RClauseList, BList, pos:right(Pos)),
    {{ccase, Line, [], RClauseList1, gb_trees:values(BNew)}, BNew};
exp({'fun', Line, {function, Name, Arity}}, Env, _Pos) ->
    Module = env_module(Env),
    {{lfun, Line, {fnid, Module, Name, Arity}}, binding_empty()};
exp({'fun', Line, {function, Module, Name, Arity}}, _Env, _Pos) ->
    {{lfun, Line, {fnid, Module, Name, Arity}}, binding_empty()};
exp({remote, Line, Mod, F}, Env, Pos) ->
    {[RMod, RF], B, _M} = exp_list([Mod, F], Env, Pos),
    {{remote, Line, RMod, RF}, B};
exp({'fun', Line, {clauses, FunClauseList}}, Env, Pos) ->
    KeyList = fun_key_list(FunClauseList, pos:left(Pos)),
    {RClauseList, BList} = 
        cclause_list(FunClauseList, Env, pos:left(pos:right(Pos))),
    {RClauseList1, _} = 
        make_clauses_uniform(RClauseList, BList, pos:right(pos:right(Pos))),
    {{lambda, KeyList, 
      {ccase, Line, [{var, Key} || Key <- KeyList], RClauseList1, []}}, 
     binding_empty()};

exp({call, Line, {atom, _, F}, ExpList}, Env, Pos) ->
    Module = env_module(Env),
    {REList, B} = exp_list(ExpList, Env, Pos),
    case is_bif(F, length(ExpList)) of
    	true -> 
    	    {{call, Line, {fnid, 'Tintsm', F, length(ExpList)}, REList}, B};
    	false ->
    	    {{local_call, Line, {fnid, Module, F, length(ExpList)}, REList}, B}
    end;
exp({call, Line, {remote, _, {atom, _, Module}, {atom, _, F}}, ExpList}, Env, Pos) ->
    {REList, B} = exp_list(ExpList, Env, Pos),
    {{call, Line, {fnid, Module, F, length(ExpList)}, REList}, B};
exp({call, Line, E, EList}, Env, Pos) ->
    {RE, B1} = exp(E, Env, pos:left(Pos)),
    {REList, B2} = exp_list(EList, Env, pos:right(Pos)),
    B = binding_union(B1, B2),
    {{apply, Line, RE, REList}, B};
exp(LC={lc, _, _, _}, Env, Pos) ->
    lc(LC, Env, Pos);
exp({'catch', Line, Exp}, Env, Pos) ->
    {RE, B} = exp(Exp, Env, Pos),
    {{'catch', Line, RE}, B};

exp({'try', Line, Body, CRClauseList, ClauseList, After}, Env, Pos) ->
    [Pos1, Pos2, Pos3, Pos4] = pos:pos_list(4, Pos),
    {RBody, _} = block(Body, Env, Pos1),
    {RCRClauseList, _} = cclause_list_u(CRClauseList, Env, Pos2),
    {RClauseList, _} = cclause_list_u(ClauseList, Env, Pos3),
    {RAfter, _} = block(After, Env, Pos4),
    {{'try', Line, RBody, RCRClauseList, RClauseList, RAfter}, 
     binding_empty()};
exp({'receive', Line, ClauseList}, Env, Pos) ->
    {RClauseList, BList} = 
        cclause_list(ClauseList, Env, pos:left(Pos)),
    {RClauseList1, BNew} = 
        make_clauses_uniform(RClauseList, BList, pos:right(Pos)),
    {{'receive', Line, RClauseList1, gb_trees:values(BNew)}, binding_empty()};
exp({'receive', Line, ClauseList, After, Final}, Env, Pos) ->
    {RClauseList, BList} = 
        cclause_list(ClauseList, Env, pos:left(pos:left(Pos))),
    {RClauseList1, BNew} = 
        make_clauses_uniform(RClauseList, BList, pos:right(pos:left(Pos))),
    {RE1, B1} = exp(After, Env, pos:left(pos:right(Pos))),
    {RE2, B2} = block(Final, Env, pos:right(pos:right(Pos))),
    B = binding_union(B1, B2),
    {{'receive', Line, RClauseList1, gb_trees:values(BNew), 
      RE1, RE2}, B};
exp(R = {record, _, Name, _}, Env, Pos) ->
    E = record_expand:exp(R, env_record(Name, Env), pos:left(Pos)),
    exp(E, Env, pos:right(Pos));    
exp(R = {record, _, _, Name, _}, Env, Pos) ->
    E = record_expand:exp(R, env_record(Name, Env), pos:left(Pos)),
    exp(E, Env, pos:right(Pos));    
exp(R={record_field, _, _, Name, _}, Env, Pos) ->
    E = record_expand:exp(R, env_record(Name, Env), pos:left(Pos)),
    exp(E, Env, pos:right(Pos));
exp({record_index, _, _, _}, _, _) ->
    {{c, conn:integer(), []}, binding_empty()};

exp({block, Line, Block}, Env, Pos) ->
    {BlockR, B} = block(Block, Env, Pos),
    {{block, Line, BlockR}, B}.

%: +func bin_element_list :: list(absyn:bin_element()) * env() * pos()
%:                        -> !{!list(exp()), !bind()}.
bin_element_list([], _, _) ->
    {[], binding_empty()};
bin_element_list([BE|BEList], Env, Pos) ->
    {RBE, B1} = bin_element(BE, Env, pos:left(Pos)),    
    {RBEList, B2} = bin_element_list(BEList, Env, pos:right(Pos)),
    B = binding_union(B1, B2),
    {[RBE|RBEList], B}.

bin_element({bin_element, _Line, Exp, Size, OptBitTypeList}, Env, Pos) ->
    {RE, B} = exp(Exp, Env, Pos),
    {{c, conn:tuple_fixed(3), [RE, bin_element_size(Size), 
		      bin_element_type(OptBitTypeList)]}, B}.


bin_element_size(default) ->
    {c, conn:atom_fixed(default), []};
bin_element_size({integer, _, _Value}) ->
    {c, conn:integer(), []}.

bin_element_type(default) ->
    {c, conn:nil(), []};
bin_element_type([binary]) ->
    {c, conn:cons(), [{c, conn:atom_fixed(binary), []}, {c, conn:nil(), []}]};
bin_element_type([signed]) ->
    {c, conn:cons(), [{c, conn:atom_fixed(signed), []}, {c, conn:nil(), []}]};
bin_element_type([unsigned]) ->
    {c, conn:cons(), [{c, conn:atom_fixed(unsigned), []}, {c, conn:nil(), []}]};
bin_element_type([utf8]) ->
    {c, conn:cons(), [{c, conn:atom_fixed(utf8), []}, {c, conn:nil(), []}]};
bin_element_type([utf16]) ->
    {c, conn:cons(), [{c, conn:atom_fixed(utf16), []}, {c, conn:nil(), []}]};
bin_element_type([utf16, little]) ->
    {c, conn:cons(), [{c, conn:atom_fixed(utf16), []}, 
               {c, conn:cons(), [{c, conn:atom_fixed(little), []},
                          {c, conn:nil(), []}]}]};
bin_element_type([utf32]) ->
    {c, conn:cons(), [{c, conn:atom_fixed(utf32), []}, {c, conn:nil(), []}]};
bin_element_type([utf32, little]) ->
    {c, conn:cons(), [{c, conn:atom_fixed(utf32), []}, 
               {c, conn:cons(), [{c, conn:atom_fixed(little), []},
                          {c, conn:nil(), []}]}]}.

%: +func exp_list :: list(absyn:exp()) * env() * pos() 
%:     -> !{!list(!exp()), !bind()}.
exp_list([], _, _) ->
    {[], binding_empty()};
exp_list([E|EList], Env, Pos) ->
    {RE, B1} = exp(E, Env, pos:left(Pos)),
    {REList, B2} = exp_list(EList, Env, pos:right(Pos)),
    B = binding_union(B1, B2),
    {[RE|REList], B}.

%: +func exp_string :: list(integer()) -> exp().
exp_string("") ->
    {c, conn:nil(), []};
exp_string([_|L]) ->
    {c, conn:cons(), [{c, conn:integer(), []}, 
                      exp_string(L)]}.


fun_key_list([{clause, _, PatternList, _, _}|_],  Pos) ->
    key_list(PatternList, Pos).

%: +func key_list :: list(any()) * pos:pos() -> list(name()).
key_list([], _) ->
    [];
key_list([_|L], Pos) ->
    [make_name_p(arg, pos:left(Pos))|key_list(L, pos:right(Pos))].


%: +func is_bif :: atom() * integer() -> true + false.
is_bif(abs, 1) -> true;
is_bif(atom_to_list, 1) -> true;
is_bif(binary_to_list, 1) -> true;
is_bif(byte_size, 1) -> true;
is_bif(element, 2) -> true;
is_bif(error, 1) -> true;
is_bif(exit, 1) -> true;
is_bif(float, 1) -> true;
is_bif(iolist_to_binary, 1) -> true;
is_bif(integer_to_list, 1) -> true;
is_bif(is_atom, 1) -> true;
is_bif(is_integer, 1) -> true;
is_bif(is_list, 1) -> true;
is_bif(is_process_alive, 1) -> true;
is_bif(is_record, 2) -> true;
is_bif(is_tuple, 1) -> true;
is_bif(hd, 1) -> true;
is_bif(length, 1) -> true;
is_bif(list_to_atom, 1) -> true;
is_bif(list_to_binary, 1) -> true;
is_bif(list_to_integer, 1) -> true;
is_bif(list_to_tuple, 1) -> true;
is_bif(make_ref, 0) -> true;
is_bif(monitor_node, 2) -> true;
is_bif(node, 1) -> true;
is_bif(process_flag, 2) -> true;
is_bif(round, 1) -> true;
is_bif(self, 0) -> true;
is_bif(setelement, 3) -> true;
is_bif(spawn, 1) -> true;
is_bif(spawn, 3) -> true;
is_bif(throw, 1) -> true;
is_bif(trunc, 1) -> true;
is_bif(tuple_to_list, 1) -> true;
is_bif(tuple_size, 1) -> true;
is_bif(whereis, 1) -> true;
is_bif(_, _) -> false.


%: +func lc :: absyn:exp() * env() * pos() -> !{!exp(), !bind()}.
lc({lc, Line, E, GenList}, Env, Pos) ->
    {RGenList, B1} = gen_list(GenList, Env, pos:left(Pos)),
    {RE, _B2} = exp(E, env_add_binding(Env, B1), pos:right(Pos)),
    {{lc, Line, RE, RGenList}, binding_empty()}.

gen_list([], _, _) ->
    {[], binding_empty()};
gen_list([Gen|GenList], Env, Pos) ->
    {RGen, B1} = gen(Gen, Env, pos:left(Pos)),
    {RGenList, B2} = gen_list(GenList, env_add_binding(Env,  B1), pos:right(Pos)),
    B = binding_union(B1, B2),
    {[RGen|RGenList], B}.

gen({generate, Line, P, E}, Env, Pos) ->
    {RE, B1} = exp(E, Env, pos:right(Pos)),
    {RP, B2, _} = pattern(P, env_add_binding(Env, B1), pos:left(Pos)),
    B = binding_union(B1, B2),
    {{generate, Line, RP, RE}, B};
gen({b_generate, Line, P, E}, Env, Pos) ->
    {RE, B1} = exp(E, Env, pos:right(Pos)),
    {RP, B2, _} = pattern(P, env_add_binding(Env, B1), pos:left(Pos)),
    B = binding_union(B1, B2),
    {{b_generate, Line, RP, RE}, B};
gen(E, Env, Pos) ->
    {RE, _} = exp(E, Env, Pos),
    {RE, binding_empty()}.

%Head

head([P|PList], Env, Pos) -> 
    {RP, B1, M1} = pattern(P, Env, pos:left(Pos)),
    Env1 = env_add_binding(Env, B1),
    {RPList, B2, M2} = head(PList, Env1, pos:right(Pos)),
    B = binding_union(B1, B2),
    {[RP|RPList], B, M1++M2};  
head([], _, _) -> 
    {[], binding_empty(), []}.

unuglify([]) ->
    [[]];
unuglify(Guard) ->
    Guard.

guard(Guard, Env, Pos) ->
    [element(1, branch(Branch, Env, Pos)) || Branch <- Guard].

branch(Branch, Env, Pos) ->
    {RL, B} = exp_list(Branch, Env, Pos),
    {RL, B, []}.
    

% pattern

%: +func pattern :: absyn:pattern() * env() * pos() -> !{!pattern(), !bind(), !list(!match())}.		      

pattern({var, _Line, '_'}, _Env, _Pos) ->
    {{empty, []}, binding_empty(), []};
pattern({var, Line, Name}, Env, Pos) ->
    case env_lookup(Name, Env) of
	none ->
	    Name1 = make_name_p(Name, Pos),
	    B = binding_singleton(Name, Name1),
            {{empty, [Name1]}, B, []};
        {value, Name1} ->
	    Name2 = make_name_p(Name, Pos),
	    M = check_eq(Line, {var, Name2}, {var, Name1}),
  	    {{empty, [Name2]}, binding_empty(), M}
    end;
pattern({nil, _}, _, _) ->
    {{c, conn:nil(), [], []}, binding_empty(), []};
pattern({atom, _, Name}, _, _) ->
    {{c, conn:atom_fixed(Name), [], []}, binding_empty(), []};
pattern({integer, Line, N}, _, Pos) ->
    Name = make_name_p(intvar, Pos),
    M = check_eq_integer(Line, {var, Name}, N),
    {{empty, [Name]}, binding_empty(), M};
pattern({char, Line, N}, _, Pos) ->
    Name = make_name_p(intvar, Pos),
    M = check_eq_integer(Line, {var, Name}, N),
    {{empty, [Name]}, binding_empty(), M};
pattern({cons, _, P1, P2}, Env, Pos) ->
    {RP1, B1, M1} = pattern(P1, Env, pos:left(Pos)),
    {RP2, B2, M2} = pattern(P2, Env, pos:right(Pos)),
    B = binding_union(B1, B2),
    {{c, conn:cons(), [RP1, RP2], []}, B, M1++M2};
pattern({tuple, _, [{atom, _, Name}|PList]}, Env, Pos) ->
    {RPList, B, M} = pattern_list(PList, Env, Pos),
    {{c, conn:tuple_tagged(Name, length(PList)+1), RPList, []}, B, M};
pattern({tuple, _, PList}, Env, Pos) ->
    {RPList, B, M} = pattern_list(PList, Env, Pos),
    {{c, conn:tuple_fixed(length(PList)), RPList, []}, B, M};
pattern({string, _, S}, _Env, _Pos) ->
    P=pattern_string(S),
    {P, binding_empty(), []};
pattern({op, _, '++', {string, _, S}, P}, Env, Pos) ->
    {RP1, B, M} = pattern(P, Env, Pos),             
    RP2 = pattern_string(S, RP1),
    {RP2, B, M};        
% "hel"++Z = "hello". 
pattern({op, Line, _, _}, _, Pos) ->
    Name = make_name_p(intvar, Pos),
    N = 4711, %silly
    M = check_eq_integer(Line, {var, Name}, N),
    {{empty, [Name]}, binding_empty(), M};
% -(1.0) = -1.0. 100+200=300. 100*2=200.
pattern({bin, _, BEList}, Env, Pos) -> %leave binaries for now...
    {B, M} = pattern_bin_element_list(BEList, Env, Pos),
    {{c, conn:bitstring(), [], []}, B, M};   
pattern({match, _, P1, P2}, Env, Pos) ->
    {RP1, B1, M1} = pattern(P1, Env, pos:left(Pos)),
    {RP2, B2, M2} = pattern(P2, Env, pos:right(Pos)),
    B = binding_union(B1, B2),
    RP = pattern_match(RP1, RP2),
    {RP, B, M1++M2};
pattern(R = {record, _Line, Name, _FieldList}, Env, Pos) ->
    P = record_expand:pattern(R,  env_record(Name, Env), pos:left(Pos)),
    pattern(P, Env, pos:right(Pos));
pattern({record_index, Line, _, _}, _, Pos) ->
    Name = make_name_p(intvar, Pos),
    N = 4711, % also silly
    M = check_eq_integer(Line, {var, Name}, N),
   {{empty, [Name]}, binding_empty(), M}.


pattern_string([]) ->
    {c, conn:nil(), [], []};
pattern_string([_|L]) ->
    {c, conn:cons(), [{c, conn:integer(), [], []}, pattern_string(L)], []}.

pattern_string([], RP) ->
    RP;
pattern_string([_|L], RP) ->
    {c, conn:cons(), [{c, conn:integer(), [], []}, pattern_string(L, RP)], []}.

pattern_list([], _, _) ->
    {[], binding_empty(), []};
pattern_list([P|PList], Env, Pos) ->
    {RP, B1, M1} = pattern(P, Env, pos:left(Pos)),
    {RPList, B2, M2} = pattern_list(PList, Env, pos:right(Pos)),
    B = binding_union(B1, B2),
    {[RP|RPList], B, M1++M2}.

pattern_bin_element_list([], _,_) ->
    {binding_empty(), []};
pattern_bin_element_list([BE|BEList], Env, Pos) ->
    {B1, M1} = pattern_bin_element(BE, Env, pos:left(Pos)),
    {B2, M2} = pattern_bin_element_list(BEList, Env, pos:right(Pos)),
    B  = binding_union(B1, B2),
    {B, M1++M2}.

pattern_bin_element({bin_element,_,P,_Size, _L}, Env, Pos) ->
    {_RP, B1, M1} = pattern(P, Env, pos:left(Pos)),
    {B1, M1}.


pattern_match({empty, VarList1}, {empty, VarList2}) ->
    {empty, VarList1++VarList2};
pattern_match({empty, VarList1}, {c, C, PList, VarList2}) ->
    {c, C, PList, VarList1++VarList2};
pattern_match({c, C, PList, VarList1}, {empty, VarList2}) ->
    {c, C, PList, VarList1++VarList2};
pattern_match({c, C, PList1, VarList1}, {c, C, PList2, VarList2}) ->
    PList = [pattern_match(P1, P2) || 
		{P1, P2} <- lists:zip(PList1, PList2)],
    {c, C, PList, VarList1++VarList2};
pattern_match({c, C1, _, _}, {c, C2, _, _}) when C1 /= C2->
    % pattern cannot possibly match
    unmatchable.

% split

%: +func split_head_guard :: absyn:head() * absyn:guard() * env() * pos() 
%:                        -> !{!list(!{!head(), !guard()}), !bind()}.    
split_head_guard(Head, Guard, Env, Pos) ->
    {RHead1, B, M} = head(Head, Env, pos:left(Pos)),
    Env1 = env_add_binding(Env, B),
    RGuard = guard(unuglify(Guard), Env1, Pos), 
    PairList = 
    [{RHead2, M1}
     || RGuardOne <- RGuard,
        {RHead2, M1} <- guard_split_one(RHead1, RGuardOne, M)],
    {PairList, B}.

tasty({call, _, Fnid, ArgList}) ->
    tasty_fnid(Fnid, ArgList);
tasty(_) ->
    none.

guard_split_one(RHead, One, M) ->
    {RMapList, New} = guard_split_1(One, [gb_trees:empty()], []),
    RHeadList =
        [fuse_head(RHead, RMap) || RMap <- RMapList],
    M1 = New++M,
    [ {RHead1, M1} || RHead1 <- RHeadList].

guard_split_1([], MapList, RGuard) ->
    {MapList, lists:reverse(RGuard)};
guard_split_1([Test|Guard], MapList, RGuard) ->
    case tasty(Test) of
        {value, RVarPList} ->
            MapList1 = [gb_trees:insert(RVar, P, Map) ||
                           {RVar, P} <- RVarPList,
                           Map <- MapList],
            guard_split_1(Guard, MapList1, RGuard);
        none ->
	    guard_split_1(Guard, MapList, [Test|RGuard])
    end.

tasty_fnid({fnid, 'Tintsm',is_atom, 1}, [{var, Name}]) ->
    {value, [{Name, {c, conn:atom(), [], []}}]};
tasty_fnid({fnid, 'Tintsm', is_tuple, 1}, [{var, Name}]) ->
    {value, [{Name, {c, conn:tuple(), [{empty, []}], []}}]};
tasty_fnid({fnid, 'Tintsm',is_integer, 1}, [{var, Name}]) ->
    {value, [{Name, {c, conn:integer(), [], []}}]};
tasty_fnid({fnid, 'Tintsm',is_list, 1}, [{var, Name}]) ->
    {value, [{Name, {c, conn:nil(), [], []}},
             {Name, {c, conn:cons(), [{empty, []}, {empty, []}], []}}]};
tasty_fnid(_Fnid, _) ->
    none.

fuse_head(Head, Map) ->
    [fuse(P, Map)||P<-Head].    

fuse({empty, VarList}, Map) ->
    TastyList = find_tasty_list(VarList, Map),
    lists:foldl(fun merge/2, {empty, VarList}, TastyList);
fuse({c, C, PList, VarList}, Map) ->
    PList1 = [fuse(P, Map) || P <-PList],
    TastyList = find_tasty_list(VarList, Map),	
    P = {c, C, PList1, VarList},
    lists:foldl(fun merge/2, P, TastyList).

find_tasty_list([], _) ->
    [];
find_tasty_list([Var|VarList], Map) ->
    case gb_trees:lookup(Var, Map) of
	none ->
	    find_tasty_list(VarList, Map);
	{value, P} ->
	    [P|find_tasty_list(VarList, Map)]
    end.


merge({empty, VarList}, {empty, VarList1}) ->
    {empty, VarList++VarList1};
merge({empty, VarList}, {c, C, PList, VarList1}) ->
    {c, C, PList, VarList++VarList1};
merge({c, C, PList, VarList}, {empty, VarList1}) ->
    {c, C, PList, VarList++VarList1};
merge({c, C, _, _}, {c, C1, _, _}) when C /= C1 ->
    unmatchable;
merge({c, C, PList, VarList}, {c, C1, PList1, VarList1}) when C == C1 ->
    PList2 = [merge(P, P1) || {P, P1} <- lists:zip(PList, PList1)],
    {c, C, PList2, VarList++VarList1}.

cclause_list_u(ClauseList, Env, Pos) ->
    {CL1, BList} = cclause_list(ClauseList, Env, pos:left(Pos)),
    make_clauses_uniform(CL1, BList, pos:right(Pos)).
    
%: +func cclause_list :: list(absyn:clause()) * env() * pos() 
%:                    -> {list(cclause()), list(bind())}.
cclause_list([Clause|ClauseList], Env, Pos) ->
    case cclause(Clause, Env, pos:left(Pos)) of
	{Clause1, B} -> 
	    case ClauseList of
		[] ->
		    {[Clause1], [B]};
		_ ->
		    {ClauseList1, BList} 
                        = cclause_list(ClauseList, Env, pos:right(Pos)),
		    {[Clause1|ClauseList1], [B|BList]}
	    end
    end;
cclause_list([], _, _) ->           
    {[], []}.

%: +func cclause :: absyn:clause() * env() * pos() -> {cclause(), bind()}.
cclause({clause,Line, Head, Guard, Body}, Env, Pos) ->
    {PairList, B1} = 
        split_head_guard(Head, Guard, Env, pos:left(pos:left(Pos))),    

    Env1 = env_add_binding(Env, B1),
    {Body1, B3} = body(Body, Env1, pos:right(Pos)),
    B = binding_union(B1, B3),
    Clause = {clause, Line, PairList, Body1},

    {Clause, B}.


%% %: +func pclause :: absyn:line() * env() * pos() -> line().
%% pclause(Line, Env, Pos) ->
%%     Clause = {clause, Line, [], []},
%%     Line.


% env

%: +func env_new :: atom() * record_env() -> !env().		      
env_new(Module, Records) ->
    {env, Module, Records, gb_trees:empty()}.

%: +func env_module :: env() -> !atom().			
env_module({env, Module, _, _Map}) ->
    Module.

env_insert(Var, K, {env, M, R, Map}) ->
    Map1 = gb_trees:insert(Var, K, Map),
    {env, M, R, Map1}.

%: +func env_lookup :: name() * env() -> {value, name()}+none.

env_lookup(Var, {env, _, _, Map}) ->
    gb_trees:lookup(Var, Map).

%: +func env_add_binding :: env() * bind() -> !env().     
env_add_binding(Env, B) ->
    env_add_blist(Env, gb_trees:to_list(B)).

%: +func env_record :: name() * env() -> program:recorddef().
env_record(Name, {env, M, Records, _}) ->
    Rid = {rid, M, Name},	   
    gb_trees:get(Rid, Records).


env_add_blist(Env, []) ->
    Env;
env_add_blist(Env, [{K, V}|B]) ->
    Env1 = env_insert(K, V, Env),
    env_add_blist(Env1, B).

%: +func binding_empty :: () -> bind().
binding_empty() ->
    gb_trees:empty().


%: +func binding_singleton :: name() * name() -> bind().
binding_singleton(K, V) ->
    B = binding_empty(),
    binding_insert(K, V, B).

%: +func binding_insert :: name() * name() * bind()-> bind().
binding_insert(K, V, B) ->
    gb_trees:insert(K, V, B).

%: +func binding_union :: bind() * bind() -> !bind().	
binding_union(B1, B2) ->
    {B, _} = binding_insert_list(gb_trees:to_list(B1), B2, []),
    B.

binding_insert_list([], B, M) ->
    {B, M};
binding_insert_list([{K, V}|L], B, M) ->
    case gb_trees:lookup(K, B) of
	{value, _V1} ->
	    B1 = B,
	    M1 = [{check, {c, conn:atom_fixed(true),[]}}]++M;
	none->
	    B1 = binding_insert(K, V, B),
	    M1 = M
    end,
    binding_insert_list(L, B1, M1).

bkeylist_intersect(BKeyList1, BKeyList2) ->
    ordsets:intersection(BKeyList1, BKeyList2).

binding_to_keylist(B) ->
    gb_trees:values(B).

% Make clauses uniform
%: +func make_clauses_uniform :: list(cclause()) * list(bind()) * pos() ->
%:				   {list(cclause()), bind()}.
make_clauses_uniform(ClauseList, BList, Pos) ->
    ExportListList = [gb_trees:keys(B) || B <- BList], 
    ExportList = case ExportListList of
		     [] ->
			 [];
		     _ ->
			 ordsets:intersection(ExportListList)
		 end,
    NewList = [{Name, make_name_p(Name, Pos)} || Name <- ExportList],       
    BNew = make_bnew(NewList),
    ClauseList1 = [translate_clause(Clause, B, BNew) || 
                      {Clause, B} <- lists:zip(ClauseList, BList)],
    {ClauseList1, BNew}.
    
make_bnew([]) ->
    gb_trees:empty();
make_bnew([{Name, New}|L]) ->
    B = make_bnew(L),
    gb_trees:insert(Name, New, B).                

%: +type pclause() 
%:     = {clause, line(), list({list(pattern()), list(exp())}), list(exp())}.
% +func translate_clause :: pclause() * bind() * gb_trees:tree(name(), name()) ->
%			       !pclause().
translate_clause(Clause, B, BNew) ->
    Map = make_translate_map(B, gb_trees:to_list(BNew)),
    C = rename_clause(Clause, Map),
    C.
   

make_translate_map(_, []) ->
    gb_trees:empty();
make_translate_map(B, [{Var, New}|L]) ->
    Old = gb_trees:get(Var, B),
    Map = make_translate_map(B, L),
    gb_trees:insert(Old, New, Map).

%: +func rename_clause :: cclause() * gb_trees:tree(name(), name()) ->
%: 			    !cclause().
rename_clause({clause, Line, PairList, Body}, Map) ->
    PairList1 = 
        [{rename_head(Head, Map),rename_guard(Guard, Map)}
         || {Head, Guard} <- PairList],
    Body1 = rename_body(Body, Map),
    {clause, Line, PairList1, Body1}.

rename_clause_list(CList, Map) -> 
    [rename_clause(C, Map) || C <- CList].


rename_head(PList, Map) ->
    [rename_pattern(P, Map) || P<-PList].

rename_pattern({empty, VarList}, Map) ->
    VarList1 = rename_var_list(VarList, Map),
    {empty, VarList1};
rename_pattern({c, C, PList, VarList}, Map) ->
    PList1 = [rename_pattern(P, Map) || P<-PList],
    VarList1 = rename_var_list(VarList, Map),
    {c, C, PList1, VarList1}.


%: +func rename_var_list :: list(name()) * gb_trees:tree(name(), name()) ->
%:			      !list(name()).
			      
rename_var_list(VarList, Map) ->
    [rename_var(Var, Map) || Var <- VarList].

rename_var(Var, Map) ->
    case gb_trees:lookup(Var, Map) of
        {value, New} ->
            New;
        none ->
            Var
    end.

rename_guard([], _Map) ->
    [];
rename_guard([E | EList], Map) ->
    E1 = rename_guard_exp(E, Map),
    EList1 = rename_guard(EList, Map),
    [E1 | EList1].

rename_guard_exp(E, _Map) ->
    E.

%: +func rename_body :: list(exp()) * gb_trees:tree(name(), name()) ->
%:			 !list(!exp()).
rename_body([], _Map) ->
    [];
rename_body([E | EList], Map) ->
    E1 = rename_exp(E, Map),	    
    EList1 = rename_body(EList, Map),
    [E1 | EList1].

%: +func rename_exp :: exp() * gb_trees:tree(name(), name()) ->
%: 			    !exp().

rename_exp({c, C, EList}, Map) ->
    {c, C, [rename_exp(E, Map) || E <- EList]};
rename_exp({var, Name}, Map) ->
    Name1 = rename_var(Name, Map),
    {var, Name1};
rename_exp(LFun = {lfun, _, _}, _) ->
    LFun;
rename_exp({match, Line, P, E}, Map) ->
    P1 = rename_pattern(P, Map),
    E1 = rename_exp(E, Map),
    {match, Line, P1, E1};
rename_exp({lambda, ArgList, E}, Map) ->
    ArgListR = rename_var_list(ArgList, Map),
    ER = rename_exp(E, Map),
    {lambda, ArgListR, ER};
rename_exp({remote, Line, M, F}, Map) ->
    [MR, FR] = rename_exp_list([M, F], Map),     
    {remote, Line, MR, FR};
rename_exp({apply, Line, E, EList}, Map) ->
    [ER|EListR] = rename_exp_list([E|EList], Map),
    {apply, Line, ER, EListR};
rename_exp({local_call, Line, F, EList}, Map) ->
    EListR = rename_exp_list(EList, Map),
    {local_call, Line, F, EListR};
rename_exp({call, Line, Fnid, EList}, Map) ->
    EListR = rename_exp_list(EList, Map),
    {call, Line, Fnid, EListR};
rename_exp({ccase, Line, EList, ClauseList, ExportList}, Map) ->
    EListR = rename_exp_list(EList, Map),
    ClauseListR = rename_clause_list(ClauseList, Map),
    ExportListR = rename_var_list(ExportList, Map),
    {ccase, Line, EListR, ClauseListR, ExportListR};
rename_exp({lc, Line, E, GenList}, Map) ->
    ER = rename_exp(E, Map),
    GenListR = [rename_gen(Gen, Map) || Gen <- GenList],
    {lc, Line, ER, GenListR};
rename_exp({'catch', Line, E}, Map) ->
    {'catch', Line, rename_exp(E, Map)};
rename_exp({'receive', Line, ClauseList, ExportList}, Map) ->
    ClauseListR = rename_clause_list(ClauseList, Map),
    ExportListR = rename_var_list(ExportList, Map),
    {'receive', Line, ClauseListR, ExportListR};   
rename_exp({'receive', Line, ClauseList, ExportList, After, Final}, Map) ->
    ClauseListR = rename_clause_list(ClauseList, Map),
    ExportListR = rename_var_list(ExportList, Map),
    AfterR = rename_exp(After, Map),
    FinalR = rename_exp_list(Final, Map),
    {'receive', Line, ClauseListR, ExportListR, AfterR, FinalR};
rename_exp({'try', Line, Body, CRClauseList, ClauseList, After}, Map) ->
    BodyR = rename_exp_list(Body, Map),
    CRClauseListR = rename_clause_list(CRClauseList, Map),
    ClauseListR = rename_clause_list(ClauseList, Map),
    AfterR = rename_exp_list(After, Map),
    {'try', Line, BodyR, CRClauseListR, ClauseListR, AfterR};
rename_exp({block, Line, ExpList}, Map) ->
    {block, Line, rename_exp_list(ExpList, Map)};
rename_exp({check, E}, Map) ->
    {check, rename_exp(E, Map)};
rename_exp({promise, Line, E1, TExp}, Map) ->
    E1R = rename_exp(E1, Map),
    {promise, Line, E1R, TExp}.

   

rename_exp_list(EList, Map) ->
    [rename_exp(E, Map) || E <- EList].

rename_gen({generate, Line, P, E}, Map) ->
    {generate, Line, rename_pattern(P, Map), rename_exp(E, Map)};
rename_gen({b_generate, Line, P, E}, Map) ->
    {b_generate, Line, rename_pattern(P, Map), rename_exp(E, Map)};
rename_gen(E, Map) ->
    rename_exp(E, Map).


% Convert type expressions too
%: +func type_def :: absyn:form() * atom() 
%:                -> !program:typedef().
type_def({scfa_eq, Line, {atom, _, _Type}, {atom, _, Name}, ArgList, TE}, M) ->
    % Type may be aftom type or opaque.
    Pos = pos:pos_zero(),
    RArgList = type_exp_list(ArgList, M, pos:left(Pos)),
    RTE = type_exp(TE, M, pos:right(Pos)),
    Tyid = {tyid, M, Name},
    {typedef, Line, Tyid, RArgList, RTE}.

type_def_external({scfa_eq, Line, {atom, _, opaque},
			    	  {atom, _, Name}, ArgList, _TE}, Module) ->
    Pos = pos:pos_zero(),
    RArgList = type_exp_list(ArgList, Module, pos:left(Pos)),
    RTE = {c, conn:abstract({Module, Name}), RArgList},
    Tyid = {tyid, Module, Name},
    {typedef, Line, Tyid, RArgList, RTE}.

%: +func type_exp :: absyn:texp() * atom() -> texp().		       
type_exp(TE, M) ->
    Pos = pos:pos_zero(),
    type_exp(TE, M, Pos).

%: +func type_exp :: absyn:texp() * atom() * pos:pos() -> texp().		       
type_exp({var, _, Name}, _M, _Pos) ->
    {var, Name};
type_exp({alias, Line, Var, Exp}, M, Pos) ->
    RVar = type_exp(Var, M, pos:left(Pos)),
    RTE = type_exp(Exp, M, pos:right(Pos)),
    {alias, Line, RVar, RTE};
type_exp({nil, _}, _, _) ->
    {c, conn:nil(), []};
type_exp({atom, _, Name}, _, _) ->
    {c, conn:atom_fixed(Name), []};
type_exp({cons, _, TE1, TE2}, M, Pos) ->
    RTE1 = type_exp(TE1, M, pos:left(Pos)),
    RTE2 = type_exp(TE2, M, pos:right(Pos)),
    {c, conn:cons(), [RTE1, RTE2]};
type_exp({tuple, _, [{atom, _, Name}|TEList]}, M, Pos) ->
    RTEList = type_exp_list(TEList, M, Pos),
    N = length(TEList) +1,
    {c, conn:tuple_tagged(Name, N), RTEList};
type_exp({tuple, _, TEList}, M, Pos) ->
    RTEList = type_exp_list(TEList, M, Pos),
    N = length(TEList),
    {c, conn:tuple_fixed(N), RTEList};
type_exp({arrow, _, TEList, TE}, M, Pos) ->
    RTEList = type_exp_list(TEList, M, pos:left(Pos)),
    RTE = type_exp(TE, M, pos:right(Pos)),
    {c, conn:fun_fixed(length(TEList)), [RTE | RTEList]};
type_exp({union, Line, TE1, TE2}, M, Pos) ->
    RTE1 = type_exp(TE1, M, pos:left(Pos)),
    RTE2 = type_exp(TE2, M, pos:right(Pos)),
    {union, Line, RTE1, RTE2};
type_exp({named_type,Line,{atom,_,Name}, TList}, M, Pos) ->
    RTList = type_exp_list(TList, M, pos:left(Pos)),
    case is_primitive_type(Name, RTList, Line, pos:right(Pos)) of
	{value, RT} ->
	    RT;
	none ->
            {type_call, Line, {tyid, M, Name}, local, RTList}
    end;
type_exp({extern_type, Line,{atom,_,DM}, {atom, _, Name}, TList}, M, Pos) ->
    RTList = type_exp_list(TList, M, pos:left(Pos)),
    {type_call, Line, {tyid, DM, Name}, extern, RTList};
type_exp({watch, Line, TE}, M, Pos) ->
    RTE = type_exp(TE, M, Pos),
    {watch, Line, RTE}. % Needs information on the module too?


type_exp_list([], _, _) ->
    [];
type_exp_list([TE|TEList], M, Pos) ->
    RTE = type_exp(TE, M, pos:left(Pos)),
    RTEList = type_exp_list(TEList, M, pos:right(Pos)),
    [RTE|RTEList].

%: +func is_primitive_type :: name() 
%:                          * list(texp()) 
%:                          * absyn:line() 
%:                          * pos:pos() 
%:                          -> none + {value, texp()}.
is_primitive_type(atom, [], _, _) ->
    {value, {c, conn:atom(), []}};
is_primitive_type(float, [], _, _) ->
%    {value, {c, integer, [{c, float, []}]}};
    {value, {c, conn:float(), []}};
is_primitive_type(integer, [], _, _Pos) ->
%    {value, {c, integer, [{var, name(anonymous, Pos)}]}};
%    {value, {c, integer, [{var, zero}]}};
    {value, {c, conn:integer(), []}};
%% is_primitive_type(tuple, [], _, _) ->
%%     {value, {c, conn:tuple(), []}};
is_primitive_type(tuple, [X], _, _Pos) ->
    {value, {c, conn:tuple(), [X]}};
%% is_primitive_type(tuple, [X, Y], _, _) ->
%%     {value, {c, conn:tuple(), [X, Y]}};
is_primitive_type(any, [], _, _Pos) ->
    {value, {c, conn:any(), []}};
is_primitive_type(none, [], _, _Pos) ->
    {value, {none}};
is_primitive_type(pid, [], _, _) ->
    {value, {c, conn:pid(), []}};
is_primitive_type(binary, [], _, _) ->
    {value, {c, conn:bitstring(), []}};
is_primitive_type(reference, [], _, _) ->
    {value, {c, conn:reference(), []}};
is_primitive_type(number, [Var], Line, _) ->
    {value, {number, Line, Var}};
is_primitive_type('fun', [TExp], _, _) ->
    {value, {c, conn:fun_any(), [TExp]}};
is_primitive_type(map, [K, V], _, _Pos) ->
    {value, {c, conn:map(), [K, V]}};
is_primitive_type(set, [K], _, _Pos) ->
    {value, {c, conn:set(), [K]}};
is_primitive_type(_, _, _, _) ->
    none.

make_arg(N) ->
    make_name_i("a", N).

make_name_p(P, Pos) ->
    String = lists:flatten(io_lib:format("~s#~s", [P, pos:name(Pos)])),
    list_to_atom(String).

make_name_i(P, I) ->
    String = lists:flatten(io_lib:format("~s%~w", [P, integer_to_list(I)])),
    list_to_atom(String).


%% name(N, Pos) ->
%%     {N, Pos}.
%% %    make_name("nn", Pos).
