%%% File    : graph.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : Create call graph
%%% Created : 28 Oct 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

%: +type bool() = erlang:bool().
%: +type list(X) = erlang:list(X).
%: +type graph(Node) = 
%:       gb_trees:tree(Node, gb_trees:tree(Node, integer())).

%: +func empty_graph :: () -> !graph(X).
%: +func node_is_defined :: Node*graph(Node) -> !bool().
%: +func new_node :: Node*graph(Node) -> !graph(!Node).
%: +func insert_son :: Node*Node*graph(Node) -> !graph(!Node).
%: +func insert_weighted_son :: Node*Node*integer()*graph(Node) -> !graph(!Node).
%: +func node_list :: graph(Node) -> !list(!Node).
%: +func son_weight_list :: Node*graph(Node) -> !list({!Node, !integer()}).
%: +func foldl :: (Node*Node*integer()*V -> V)*V*graph(Node) -> !V.
%: +func cycles :: graph(Node) -> !list(!list(!Node)).

-module(graph).
-author('svenolof@harpo.csd.uu.se').

-export([empty_graph/0, new_node/2, node_is_defined/2]).

-export([insert_son/3, insert_weighted_son/4, cycles/1]).
-export([foldl/3, node_list/1, son_weight_list/2]).

empty_graph()->
    gb_trees:empty().

new_node(Node, Graph) ->
    case gb_trees:is_defined(Node, Graph) of
	true ->
	    Graph;
	false ->
	    gb_trees:insert(Node, gb_trees:insert(Node, 0, gb_trees:empty()), Graph)
%	    gb_trees:insert(Node, gb_trees:empty(), Graph)
    end.

node_is_defined(Node, Graph) ->
    gb_trees:is_defined(Node, Graph).
    
insert_son(Father, Son, Graph) ->
    Graph1 = new_node(Father, Graph),
    Graph2 = new_node(Son, Graph1),
    {value, SonTable} = gb_trees:lookup(Father, Graph2),
    case gb_trees:lookup(Son, SonTable) of
	{value, N} ->
	    SonTable1 = gb_trees:update(Son, N+1, SonTable),
	    gb_trees:update(Father, SonTable1, Graph2);
	none ->
	    SonTable1 = gb_trees:insert(Son, 1, SonTable),
	    gb_trees:update(Father, SonTable1, Graph2)
    end.

insert_weighted_son(Father, Son, Weight, Graph) ->
    Graph1 = new_node(Father, Graph),
    Graph2 = new_node(Son, Graph1),
    {value, SonTable} = gb_trees:lookup(Father, Graph2),
    case gb_trees:lookup(Son, SonTable) of
	{value, N} ->
	    SonTable1 = gb_trees:update(Son, N+Weight, SonTable),
	    gb_trees:update(Father, SonTable1, Graph2);
	none ->
	    SonTable1 = gb_trees:insert(Son, Weight, SonTable),
	    gb_trees:update(Father, SonTable1, Graph2)
    end.

node_list(Graph) ->
    lists:map(fun({Key,_}) -> Key end, gb_trees:to_list(Graph)).
			 

son_weight_list(Father, Graph) ->
    {value, SonTable} = gb_trees:lookup(Father, Graph),
    gb_trees:to_list(SonTable).


foldl(F, Init, Graph) ->
    GraphList = gb_trees:to_list(Graph),
    G = fun({Father, SonTable}, Init0) ->
		F1 = fun({Son, Weight}, Init1) ->
			     F(Father, Son, Weight, Init1)
		     end,
		lists:foldl(F1, Init0, gb_trees:to_list(SonTable))
	end,
    lists:foldl(G, Init, GraphList).


cycles(Graph) ->
    Finish = finishing(Graph),
    GraphT = transpose(Graph),
    collect_cycles(GraphT, Finish).



finishing(Graph) ->
    NodeList = lists:map(fun ({Key, _}) -> Key end,
			 gb_trees:to_list(Graph)),
    {_, _, Finish} = 
	finishing(NodeList, Graph, 0, gb_trees:empty(), gb_trees:empty()),
    Finish.

finishing([Node | List], Graph, Count, Visited, Finish) ->
    case gb_trees:is_defined(Node, Visited) of 
	true ->
	    finishing(List, Graph, Count, Visited, Finish);
	false ->
	    Visited1 = gb_trees:insert(Node, true, Visited),
	    {value, SonTable} = gb_trees:lookup(Node, Graph),
	    SonList = lists:map(fun ({Key, _}) -> Key end,
				gb_trees:to_list(SonTable)),
	    {Visited2, Count2, Finish2} = 
		finishing(SonList, Graph, Count, Visited1, Finish),
	    Finish3 = gb_trees:insert(Node, Count2, Finish2),
	    finishing(List, Graph, Count2 + 1, Visited2, Finish3)
    end;
finishing([], _, Count, Visited, Finish) ->
    {Visited, Count, Finish}.

transpose(Graph) ->
    T = gb_trees:empty(),
    List = 
	lists:map(fun({Node, SonTable}) ->
			  {Node,
			   lists:map(fun({Son, _}) ->
					     Son 
				     end,
				     gb_trees:to_list(SonTable))}
		  end,
		  gb_trees:to_list(Graph)),
    iterate1(List, T).

iterate1([{Node, SonList} | NodeList], Graph) ->
    Graph1 = iterate2(Node, SonList, Graph),
    iterate1(NodeList, Graph1);
iterate1([], Graph) ->
    Graph.

iterate2(Node, [Son | SonList], Graph) ->
    Graph1 = insert_son(Son, Node, Graph),
    iterate2(Node, SonList, Graph1);
iterate2(_, [], Graph) ->
    Graph.

collect_cycles(GraphT, Finish) ->
    NodeList = lists:map(fun({Node, _}) -> Node end, 
			 lists:reverse(lists:keysort(2, gb_trees:to_list(Finish)))),
    Visited = gb_trees:empty(),
    CycleList = collect_cycles(NodeList, GraphT, Visited),
    CycleList.

collect_cycles([Node | List], Graph, Visited) ->
    case gb_trees:is_defined(Node, Visited) of 
	true ->
	    collect_cycles(List, Graph, Visited);
	false ->
	    Visited1 = gb_trees:insert(Node, true, Visited),
	    {value, SonTable} = gb_trees:lookup(Node, Graph),
	    SonList = lists:map(fun ({Key, _}) -> Key end,
				gb_trees:to_list(SonTable)),
	    {Component, Visited2} = collect_cycle(SonList, Graph, [Node], Visited1),
	    ComponentList = collect_cycles(List, Graph, Visited2),
	    [Component | ComponentList]
    end;
collect_cycles([], _,  _) ->
    [].

collect_cycle([Node | List], Graph, Acc, Visited) ->
    case gb_trees:is_defined(Node, Visited) of 
	true ->
	    collect_cycle(List, Graph, Acc, Visited);
	false ->
	    Visited1 = gb_trees:insert(Node, true, Visited),
	    {value, SonTable} = gb_trees:lookup(Node, Graph),
	    SonList = lists:map(fun ({Key, _}) -> Key end,
				gb_trees:to_list(SonTable)),
	    {Acc1, Visited2} = collect_cycle(SonList, Graph, [Node | Acc], Visited1),
	    collect_cycle(List, Graph, Acc1, Visited2)
    end;
collect_cycle([], _,  Acc, Visited) ->
    {Acc, Visited}.



    
