%%% @author Sven-Olof Nystrom <svenolof@wumpus>
%%% @copyright (C) 2020, Sven-Olof Nystrom
%%% Optimizations local to a predicate
%%% Created : 14 Jan 2020 by Sven-Olof Nystrom <svenolof@wumpus>

-module(local).

-export([process/1]).
-compile(export_all).

%: +func process_predicates :: walker:theory() -> walker:theory().
process_predicates(Predicates) ->
    PList = 
        [{Name, process(Predicate)} || 
            {Name, Predicate} <- gb_trees:to_list(Predicates)],
    gb_trees:from_orddict(PList).


process(Predicate) ->
    Env = info_predicate(Predicate),
    Predicate1 = propagate_predicate(Predicate, Env),
    LLive = live_predicate(Predicate1),
    P1 = remove_dead_predicate(Predicate1, LLive),
    case Predicate /= P1 of 
        true -> 
            io:format("~nBefor ~p", [Predicate]),
            io:format("~nAfter ~p", [P1]);
        false ->
            true
    end,
    P1.

%%%% Info

info_predicate({predicate, _, ParameterList, Body}) ->
    Env = info_add_global_list(ParameterList, gb_trees:empty()),
    info_body(Body, Env).

info_body([], Env) ->
    Env;    
info_body([Body1|Body2], Env) ->
    Env1 = info_body(Body1, Env),
    info_body(Body2, Env1);
info_body(Phi, Env) ->
    info_constraint(Phi, Env).

info_constraint({call, _, ArgList}, Env) ->
    info_add_global_list(ArgList, Env);
info_constraint({include, {sift, Name1, F}, {var, Name2}, Source}, Env) ->
    add_info(Name2, {filter, Name1, F, Source}, Env);
info_constraint({include, {var, Name1}, {var, Name2}, Source}, Env) ->
    add_info(Name2, {filter, Name1, rfilter:any(), Source}, Env);
info_constraint({include, E1, E2, Source}, Env) ->
    {_, L1} = sensitive(E1),
    {L2, _} = sensitive(E2),
    Env1 = info_add_global_list(L1, Env),
    info_add_global_list(L2, Env1);      
info_constraint({watch, _, _}, Env) ->
    Env.


add_info(Name, K, Env) ->
    case gb_trees:lookup(Name, Env) of
        {value, K1} ->
            gb_trees:update(Name, combine(K, K1), Env);             
        none ->
            gb_trees:insert(Name, K, Env)
    end.    

info_add_global_list([], Env) ->
    Env;
info_add_global_list([Name|L], Env) ->
    Env1 = gb_trees:enter(Name, any, Env),
    info_add_global_list(L, Env1).

combine(none, K) -> K;
combine(K, none) -> K;
combine(any, _) -> any;
combine(_, any) -> any;
combine(K, K) -> K;
combine(_, _) -> any.

%%%% Propagate
propagate_predicate({predicate, Name, PList, Body}, Env) ->
    Body1 = propagate_body(Body, Env),
    {predicate, Name, PList, Body1}.


propagate_body([BodyA|BodyB], Env) ->
    BodyA1 = propagate_body(BodyA, Env),
    BodyB1 = propagate_body(BodyB, Env),
    [BodyA1|BodyB1];
propagate_body([], _) ->
    [];
propagate_body(Phi, Env) ->
    propagate_constraint(Phi, Env).

propagate_constraint(Phi = {include, {sift, Name, Filter}, E, Source}, 
                     Env) ->
    case collapse(Name, Filter, Source, Env) of
        {filter, Name, Filter} ->
            Phi;
        {filter, Name1, Filter1, Source1} ->
            {include, {sift, Name1, Filter1}, E, Source1};
        none ->
            []
    end;
propagate_constraint({include, {var, Name}, E, Source}, 
                     Env) ->
    case collapse(Name, rfilter:any(), Source, Env) of
        {filter, Name1, Filter, Source1} ->
            {include, {sift, Name1, Filter}, E, Source1};
        none ->
            []
    end;
propagate_constraint(Phi, _) ->
    Phi.

collapse(Name, Filter, Source, Env) ->
    case gb_trees:lookup(Name, Env) of
        {value, any} ->
            {filter, Name, Filter, Source};
        {value, {filter, Name1, Filter1, Source1}} ->
            collapse(Name1, Filter1, Source1, Env);
        {value, none} ->
            {filter, Name, rfilter:empty(), Source};
        none ->
            {filter, Name, rfilter:empty(), Source}
    end.

%%%% left live

llive_predicate({predicate, _, ParameterList, Body}) ->
    Env = gb_sets:from_list(ParameterList),
    llive_body(Body, Env).

llive_body([B1|B2], Env) ->
    Env1 = llive_body(B1, Env),
    llive_body(B2, Env1);
llive_body([], Env) ->
    Env;
llive_body(Phi, Env) ->
    llive_constraint(Phi, Env).

llive_constraint({include, {sift, _, F}, TE, _}, LLive) ->
    case rfilter:is_empty(F) of 
        true ->
            LLive;
        false ->
            {LList, _} = sensitive(TE),
            add_llive_list(LList, LLive)
    end;
llive_constraint({include, E1, E2, _}, LLive) ->
    {_, LList1} = sensitive(E1),
    {LList2, _} = sensitive(E2),
    LLive1 = add_llive_list(LList1, LLive),
    add_llive_list(LList2, LLive1);
llive_constraint({call, _, List}, LLive) ->
    add_llive_list(List, LLive);
llive_constraint({watch, Name, _}, LLive) ->
    add_llive_list([Name], LLive).

add_llive_list([], LLive) ->
    LLive;
add_llive_list([Name|List], LLive) ->
    LLive1 = gb_sets:add(Name, LLive),
    add_llive_list(List, LLive1).    

%%%% Left and right liveness. 


live_predicate({predicate, _, ParameterList, Body}) ->
    Env = gb_sets:from_list(ParameterList),
    live_body(Body, {Env, Env}).

live_body([B1|B2], Live) ->
    Live1 = live_body(B1, Live),
    live_body(B2, Live1);
live_body([], Live) ->
    Live;
live_body(Phi, Live) ->
    live_constraint(Phi, Live).

live_constraint({include, {sift, Name, F}, TE, _}, Live={L,R}) ->
    case rfilter:is_empty(F) of 
        true ->
            Live;
        false ->
            {LList, RList} = sensitive(TE),
            L1 = add_live_list(LList, L),
            R1 = add_live_list([Name]++RList, R),
            {L1, R1}
    end;
live_constraint({include, E1, E2, _}, {L,R}) ->
    {PList1, NList1} = sensitive(E1),
    {PList2, NList2} = sensitive(E2),
    L1 = add_live_list(NList1++PList2, L),
    R1 = add_live_list(PList1++NList2, R),
    {L1, R1};
live_constraint({call, _, List}, {L, R}) ->
    L1 = add_live_list(List, L),
    R1 = add_live_list(List, R),
    {L1, R1};
live_constraint({watch, Name, _}, {L, R}) ->
    R1 = add_live_list([Name], L),
    {L, R1}.
    
add_live_list([], LLive) ->
    LLive;
add_live_list([Name|List], LLive) ->
    LLive1 = gb_sets:add(Name, LLive),
    add_live_list(List, LLive1).    

%%%

sensitive(TE) ->
    sensitive(TE, [], []).

sensitive({var, Name}, LAcc, RAcc) ->
    {[Name|LAcc], RAcc}; % This can't be right.
sensitive({sift, Name, _}, LAcc, RAcc) ->
    {[Name|LAcc], RAcc};
sensitive({c, C, EList}, LAcc, RAcc) ->
    {L1, L2} = conn:variance(C, EList),
    {LAcc1, RAcc1} = sensitive_list(L1, LAcc, RAcc),
    {RAcc2, LAcc2} = sensitive_list(L2, RAcc1, LAcc1),
    {LAcc2, RAcc2}.

sensitive_list([], LAcc, RAcc) ->
    {LAcc, RAcc};
sensitive_list([E|List], LAcc, RAcc) ->
    {LAcc1, RAcc1} = sensitive(E, LAcc, RAcc),
    sensitive_list(List, LAcc1, RAcc1).


%%%% Dead

active({include, E1, E2, _}, Live) ->
    left_active(E1, Live) and
        right_active(E2, Live);
active(_, _) ->
    true.

left_active({c, _, _}, _) ->
    true;
left_active({var, Name}, {L, _}) ->
    gb_sets:is_member(Name, L);
left_active({sift, Name, _}, {L, _}) ->
    gb_sets:is_member(Name, L).

right_active({c, _, _}, _) ->
    true;
right_active({var, Name}, {_, R}) ->
    gb_sets:is_member(Name, R).


remove_dead_predicate({predicate, Name, PList, Body}, LLive) ->
    Body1 = remove_dead_body(Body, LLive),
    {predicate, Name, PList, Body1}.

remove_dead_body([B1|B2], LLive) ->
    B11 = remove_dead_body(B1, LLive),
    B21 = remove_dead_body(B2, LLive),
    [B11|B21];
remove_dead_body([], _) ->
    [];
remove_dead_body(Phi, LLive) ->
    remove_dead_constraint(Phi, LLive).


remove_dead_constraint(Phi, Live) ->
    case active(Phi, Live) of
        true ->
            Phi;
        false ->
            []
    end.
            
%% remove_dead_constraint(Phi = {include, {var, Name}, _, _}, LLive) ->
%%     case gb_sets:is_member(Name, LLive) of
%%         true ->
%%             Phi;
%%         false ->
%%             []
%%     end;
%% remove_dead_constraint(Phi = {include, {sift, Name, F}, _, _}, LLive) ->
%%     case gb_sets:is_member(Name, LLive) and not rfilter:is_empty(F) of
%%         true ->
%%             Phi;
%%         false ->
%%             []
%%     end;
%% remove_dead_constraint(Phi, _) ->
%%     Phi.
