-module(poly).

-export([top/1, top/3, shorten_predicate_list/1]).
%-compile(export_all).
%: +type list(X) = erlang:list(X).
%: +type string() = erlang:string().
%: +type bool() = erlang:bool().
%: +type name() = any().
%: +type result(T) = {success, T} + {fail, solver:source()}.

%: +type predicate() = walker:predicate().
%: +type env() = map(name(), predicate()).
%: +type pos() = pos:pos().
%: +type proplist() = proplists:proplist().

%: +func top :: atom() * list(string()) * proplist() 
%:           -> result(solver:source()).
top(Module) ->
    Path = flags:get_flag(path),
    Options = flags:get(),
    top(Module, Path, Options).


top(Module, Path, Options) ->
    F = fun() ->
                run(Module, Path, Options)
        end,

    R = maybe_time(F, "\nTime", Module, proplists:get_value(time, Options, false)),
    report(Module, R).

report(M, R = {success, _}) ->   
    io:format("~n~nType checking of ~w succeeded.~n~n", [M]),
    R;
report(M, R = {fail, _}) -> 
    io:format("~n~nType checking of ~w failed.~n~n", [M]),
    R.

maybe_time(F, _, _, false) ->
    F();
maybe_time(F, Text, Data, true) ->
    {Time, R} = timer:tc(F),
    io:format("~s ~w: ~.2f seconds", [Text, Data, Time/1.0E6]),
    R.

run(Module, Path, Options) ->    
    {FnidList, P} = sane_and_complete(Module, Path, Options),
    case proplists:get_value(separate, Options, false) of
       false ->
           check_fnid_list(FnidList, Options, P);
       true ->
           check_fnid_list_individually(FnidList, Options, P)
    end.

check_fnid_list_individually([], _Options, _P) ->
    {success, []};
check_fnid_list_individually([Fnid|FnidList], Options, P) ->
    F = fun() ->
                check_fnid_list([Fnid], Options, P)
        end,
    {Time, R} = timer:tc(F),
    io:format("~nSpecification ~w: ~.2f seconds", [Fnid, Time/1.0E6]),
    case R of
        {fail, S} -> 
	    {fail, S};
        {success, _} ->
             check_fnid_list_individually(FnidList, Options, P)
    end.

check_fnid_list(FnidList, Options, P) ->
    A = agenda:top_list(FnidList, P),
    Predicates = walker:agenda(A, P),
    case proplists:get_value(predicates, Options, false) of
	true ->
            io:format("~n~ts", [prettypr:format(
                                 display:predicates(Predicates), 80, 40)]);
	false ->
	    true
    end,
    case proplists:get_value(local, Options, false) of
        true ->
            Predicates1 = local:process_predicates(Predicates);
        false ->
            Predicates1 = Predicates
    end,
    TimeSpec = proplists:get_value(time_spec, Options, false),
    Dag = coalesce:coalesce(Predicates1),
    case proplists:get_value(predicates, Options, false) of
	true ->
	    io:format("~n=== DAG ===~n", []),
            io:format("~n~ts", [prettypr:format(
                                 display:predicate_list(Dag), 80, 40)]);
	false ->
	    true
    end,
    case collapse_dag(Dag, maps:new(), pos:pos_zero(), Options) of
        {fail, S} -> 
	    {fail, S};
        {success, Tree} ->
	    solve_fnid_list(FnidList, Tree, TimeSpec, false, false)
    end.

sane_and_complete(Module, Path, Options) ->
    P = program:new(Module, Path, Options),
    P1 = program:ensure_loaded(Module, P),
    FnidList = program:find_specs(Module, P1),   
    io:format("~nSpecs: ~w", [length(FnidList)]),
    P2 = sanity:ensure_specs(FnidList, P1),
    P3 = sanity:ensure_defs(FnidList, P2),
    {FnidList, P3}.


solve_fnid_list([], _Env, _, _, _) ->
    {success, []};
solve_fnid_list([Fnid|List], Env, TimeSpec, Verbose, Views) ->    
    TName = walker:fnid_to_predicate_name(Fnid, top),
    {predicate, _, [], G} = maps:get(TName, Env),
    F = fun() -> solver:solve(G) end,
    case maybe_time(F, "\nSpec", Fnid, TimeSpec) of
	{success, _} ->
	    solve_fnid_list(List, Env, TimeSpec, Verbose, Views);
	{fail, S} ->
            S1 = lists:usort(lists:flatten([S])),
	    {fail, S1}
    end.


%: +func collapse_dag :: list(predicate()) * env() * pos() * proplist()
%:       -> result(env()).
                           
collapse_dag([Predicate|PList], Env, Pos, Options) ->
    {predicate, Name, ArgList, Phi} = Predicate,			 
    Phi1 = expand_calls(Phi, Env),

    Predicate1 = {predicate, Name, ArgList, Phi1},
    case maybe_simplify(Predicate1, Options) of
        {success, Predicate2} ->

            Env1 = maps:put(Name, Predicate2, Env),
                    collapse_dag(PList, Env1, pos:right(Pos), Options);
        {fail, S} ->
	    {fail, S}
    end;
collapse_dag([], Env, _, _) ->
    {success, Env}.

%: +func maybe_simplify :: predicate() * proplist() -> result(predicate()). 
maybe_simplify(Predicate, Options) -> 
    %TimeSpec = proplists:get_value(time_spec, Options, false),
    {predicate, _Name, _ArgList, _Phi} = Predicate,
    Predicate1 =
	shorten_predicate(Predicate),
    case proplists:get_value(simplify, Options, false) of
	true ->
	    solve_and_simplify(Predicate1, Options);
	false ->
	    {success, Predicate1}
    end.

%: +func solve_and_simplify :: predicate() * proplist() -> result(predicate()). 
solve_and_simplify(Predicate, Options) ->
    {predicate, Name, ArgList, Phi} = Predicate,
    case solve_0(Name, Phi, Options) of
	{success, Graph} ->
            %% io:format("~nSOLVED ~w", [Name]),
            %% display:show_graph(Graph),
	    case simplify(Name, ArgList, Graph, Options) of
		{success, Phi2} ->
		    P = {predicate, Name, ArgList, Phi2},
		    {success, P};
		{fail, S} ->
		    S1 = lists:usort(lists:flatten([S])),
                    {fail, S1} 
	    end;
	{fail, S} ->
	    S1 = lists:usort(lists:flatten([S])),
	    {fail, S1}
    end.


solve_0(Name, Phi, Options) ->
    F = fun() -> solver:solve_0(Phi) end,
    TimeSolver = proplists:get_value(time_solver, Options, false),
    case TimeSolver of
	true ->
	    io:format("~n~w # ~w", [Name, length(lists:flatten([Phi]))]);
	_ -> 
	    true
    end,

    maybe_time(F, "... time:", "", TimeSolver).    

simplify(Name, ArgList, Graph, Options) ->
    Nostable = proplists:get_value(nostable, Options, false),
%:  promise Nostable::bool(),
    case Nostable of
	false ->
	    F = fun() -> reach:simplifyx(Name, ArgList, Graph, Options) end;
	true ->
	    F = fun() -> reach:simplify_nostable(Name, ArgList, Graph) end
    end,
    TimeSimp = proplists:get_value(time_simp, Options, false),
    maybe_time(F, "\nSimplify:", "", TimeSimp).    



%%%% Expand

%: +func expand_calls :: walker:constraints() * env()  
%:                    -> walker:constraints().
expand_calls(Phi, Env) ->
    {Psi, _} = expand_calls_1(Phi, Env, 0),
    [Psi].

expand_calls_1([], _, N) ->
    {[], N};
expand_calls_1([Phi1|Phi2], Env, N)-> 
    {Psi1, N1} = expand_calls_1(Phi1, Env, N), 
    {Psi2, N2} = expand_calls_1(Phi2, Env, N1),
    {[Psi1| Psi2], N2};
expand_calls_1(Phi={include, _, _, _}, _, N)->
    {Phi, N};
expand_calls_1({call, Name, ArgList}, Env, N)->
    {predicate, _, Params, Phi} = maps:get(Name, Env),
    Map = make_map([zero|Params], [zero|ArgList], maps:new()),
    {Psi, N1, _Map1} = substitute(Phi, N, Map),
    {[Psi], N1};
expand_calls_1(Phi={watch, _, _}, _, N)->
    {Phi, N}.

make_map([], [], Map)  ->    
    Map;
make_map([_X|_L1], [{sift, _, _}|_L2], _Map) ->  
    error("bad");   
make_map([X|L1], [Y|L2], Map) ->  
    case maps:is_key(X, Map) of
	true ->
	    error("Repeated argument in predicate");
	false ->
	    Map1 = maps:put(X, Y, Map),
	    make_map(L1, L2, Map1)
    end.

%% %: +func substitute_exp :: list(walker:dl(walker:  
%% %:                          * gb_trees:tree(name(), name()) 
%% %:                          * pos() 
%% %:                          -> walker:texp().
substitute([], N, Map) ->
    {[], N, Map};
substitute([Phi1|Phi2], N, Map) ->
    {Psi1, N1, Map1} = substitute(Phi1, N, Map),
    {Psi2, N2, Map2} = substitute(Phi2, N1, Map1),
    {[Psi1|Psi2], N2, Map2};
substitute({include, E1, E2, Source}, N, Map)->
    {E1s, N1, Map1} = substitute_exp(E1, N, Map),
    {E2s, N2, Map2} = substitute_exp0(E2, N1, Map1),
    {{include, E1s, E2s, Source}, N2, Map2};
substitute({watch, Name, Message}, N, Map) ->
    {Name1, N1, Map1} = substitute_name(Name, N, Map),
    {{watch, Name1, Message}, N1, Map1}.


substitute_exp({var, Name}, N, Map)  ->
    {Name1, N1, Map1} = substitute_name(Name, N, Map),
    {{var, Name1}, N1, Map1};
substitute_exp({c, C, EList}, N, Map) ->
    {EListS, N1, Map1} = substitute_exp0_list(EList, N, Map),
    {{c, C, EListS}, N1, Map1};
substitute_exp({sift, Name, F}, N, Map) ->
    {NameS, N1, Map1} = substitute_name(Name, N, Map),
    {{sift, NameS, F}, N1, Map1}.

substitute_exp0({var, Name}, N, Map)  ->
    {Name1, N1, Map1} = substitute_name(Name, N, Map),
    {{var, Name1}, N1, Map1};
substitute_exp0({c, C, EList}, N, Map) ->
    {EListS, N1, Map1} = substitute_exp0_list(EList, N, Map),
    {{c, C, EListS}, N1, Map1}.


substitute_exp0_list([], N, Map)   ->
    {[], N, Map};			 
substitute_exp0_list([E|EList], N, Map)   ->
   {Es, N1, Map1} = substitute_exp0(E, N, Map),
   {EListS, N2, Map2} = 
       substitute_exp0_list(EList, N1, Map1),
   {[Es|EListS], N2, Map2}.
   
substitute_name(Name={stable, _, _}, N, Map) -> 
    {Name, N, Map};
substitute_name(Name, N, Map) -> 
    case maps:find(Name, Map) of
	{ok, Name1} ->
	    {Name1, N, Map};
	error ->
	    NameNew = list_to_atom("p"++integer_to_list(N)),
	    {NameNew, N+1, maps:put(Name, NameNew, Map)}
    end.

%% %: +func substitute_exp :: walker:texp() 
%% %:                          * gb_trees:tree(name(), name()) 
%% %:                          * pos() 
%% %:   


%%%% Shorten

%: +func shorten_predicate :: predicate() -> predicate().
shorten_predicate({predicate, Name, ArgList, Phi}) ->
    Map0 = maps:put(zero, zero, maps:new()),
    {Map, N} = short_map_name_list(ArgList, Map0, 0),
    {Map1, _} = short_map(Phi, Map, N),
    ArgList1 = shorten_name_list(ArgList, Map1),
    Phi1 = shorten(Phi,Map1),
    {predicate, Name, ArgList1, Phi1}.

%: +func shorten_predicate_list :: list(predicate()) -> list(predicate()).
shorten_predicate_list(PredicateList) ->
    [shorten_predicate(P) || P <- PredicateList].

shorten(Phi, Map) ->
    [shorten_1(Phi, Map)].


shorten_1([], _Map) ->
    [];
shorten_1([Phi1|Phi2], Map) ->
    [shorten_1(Phi1, Map)|shorten_1(Phi2, Map)];
shorten_1({include, E1, E2, Source}, Map) ->
    {include, 
     shorten_exp(E1, Map),
     shorten_exp0(E2, Map),
     Source};
shorten_1({watch, Name, Message}, Map) ->
    {watch, maps:get(Name, Map), Message};
shorten_1({call, PName, NameList}, Map) ->
    NameList1 = shorten_name_list(NameList, Map),
    {call, PName, NameList1}.

shorten_exp({var, Name}, Map) ->    
    {var, maps:get(Name, Map)};
shorten_exp({c, C, EList}, Map) ->
    {c, C, [shorten_exp0(E, Map) || E <- EList]};
shorten_exp({sift, Name, F}, Map) ->
    {sift, maps:get(Name, Map), F}.

shorten_exp0({var, Name}, Map) ->    
    {var, maps:get(Name, Map)};
shorten_exp0({c, C, EList}, Map) ->
    {c, C, [shorten_exp0(E, Map) || E <- EList]}.

shorten_name_list(L, Map) ->
    [maps:get(Name, Map) || Name <- L].

%%: +func short_map :: constraints() * gb_trees:tree(name() 
short_map([], Map, N) ->
    {Map, N};
short_map([Phi1|Phi2], Map, N) ->
    {Map1, N1} = short_map(Phi1, Map, N),
    short_map(Phi2, Map1, N1);
short_map({include, E1, E2, _}, Map, N) ->
    {Map1, N1} = short_map_exp(E1, Map, N),
    short_map_exp(E2, Map1, N1);
short_map({watch, Name, _}, Map, N) ->
    short_map_name(Name, Map, N);
short_map({call, _PName, NameList}, Map, N) ->
    short_map_name_list(NameList, Map, N).
short_map_exp({var, Name}, Map, N) ->
    short_map_name(Name, Map, N);
short_map_exp({zero}, Map, N) ->
    {Map, N};
short_map_exp({c, _C, EList}, Map, N) ->
    short_map_exp_list(EList, Map, N);
short_map_exp({sift, Name, _}, Map, N) ->
    short_map_name(Name, Map, N).

short_map_exp_list([], Map, N) ->
    {Map, N};
short_map_exp_list([E|EList], Map, N) ->
    {Map1, N1} = short_map_exp(E, Map, N),
    short_map_exp_list(EList, Map1, N1).

short_map_name(Name={stable, _, _}, Map, N) ->
    case maps:is_key(Name, Map) of
        true ->
            {Map, N};
        false ->
            {maps:put(Name, Name, Map), N+1}
    end;
short_map_name(Name={view, _}, Map, N) ->
    case maps:is_key(Name, Map) of
        true ->
            {Map, N};
        false ->
            {maps:put(Name, Name, Map), N+1}
    end;
short_map_name(Name, Map, N) ->
    case maps:is_key(Name, Map) of
        true ->
            {Map, N};
        false ->
            {maps:put(Name, N, Map), N+1}
    end.

short_map_name_list([], Map, N) ->
    {Map, N};
short_map_name_list([Name|NameList], Map, N) ->
    {Map1, N1} = short_map_name(Name, Map, N),
    short_map_name_list(NameList, Map1, N1).

 
