
-module(program).

-export([new/3, load_source_module/2, ensure_loaded/2]).
-export([def_or_spec/2]).
-export([keys/1]).
-export([def/2, typedef/2, spec/2]).
-export([typedef_is_defined/2]).
-export([find_specs/2]).
-export([def_is_sane/3, spec_is_sane/2, typedef_is_sane/2]).
-export([checked_def/2, checked_spec/2, checked_typedef/2]).

%: +type list(X) = erlang:list(X).
%: +type string() = erlang:string().
%: +type bool() = erlang:bool().
%: +type line() = absyn:line().
%: +type name() = any().
%: +type path() = io_lib:iolist().
%: +type options() = proplists:proplist().
%: +type fnid() = {fnid, atom(), atom(), integer()}.
%: +type tyid() = {tyid, atom(), atom()}.
%: +type rid() = {rid, atom(), atom()}.

%: +type def() = {def, line(), fnid(), list(convert:exp()), convert:exp()}
%:             + {import, line(), fnid(), fnid()}.
%: +type spec() = {spec, line(), fnid(), convert:texp()}.
%: +type typedef() = {typedef, line(), tyid(), list(convert:texp()), convert:texp()}.

%: +type recorddef() = 
%:    {gb_trees:tree(atom(), integer()), 
%:     gb_trees:tree(integer(), absyn:exp())}.

%: +opaque program() = 
%:       {program, 
%:            {atom(), path(), options()}, 
%:            gb_sets:tree(atom()), 
%:            gb_trees:tree(fnid(), {bool(), def()}),
%:            gb_trees:tree(fnid(), {bool(), spec()}),
%:            gb_trees:tree(tyid(), {bool(), typedef()}),
%:            gb_trees:tree(rid(), recorddef())}.


%: +func new :: atom() * path() * options() -> !program().
new(CurrentModule, Path, Options) ->
    {program, 
     {CurrentModule, Path, Options}, 
     gb_sets:empty(),
     gb_trees:empty(),     
     gb_trees:empty(),     
     gb_trees:empty(),
     gb_trees:empty()}.

%: +func load_source_module :: atom() * program() ->  !program().
load_source_module(Name, Program)  ->
    Path = get_path(Program),
    case is_module_loaded(Name, Program) of
	true ->
	    Program;
	false ->
	    case scfa_file:find_module(get_path(Program), Name, espec) of
		{ok, Absname} ->
		    Program1 = file(Absname, Name, Program),
		    Flag1 = true;            
		not_found ->
		    Program1 = Program,
		    Flag1 = false
	    end,
	    case scfa_file:find_module(Path, Name, erl) of
		{ok, Absname1} ->
    		    Program2 = file(Absname1, Name, Program1),
		    Flag2 = true;            
		not_found ->
		    Program2 = Program1,
		    Flag2 = false
	    end,

            case Flag1 or Flag2 of
                false ->
                    io:format("~nCannot find module ~w", [Name]),
                    error("Module not found");
                true ->
                    set_module_loaded(Name, Program2)
            end
    end.

get_current_module({program, {Module, _Path, _Options}, 
	  _Modules, _Defs, _Specs, _Typedefs, _Records1}) ->
    Module.

get_path({program, {_Module, Path, _Options}, 
	  _Modules, _Defs, _Specs, _Typedefs, _Records1}) ->
    Path.

get_options({program, {_Module, _Path, Options}, 
	     _Modules, _Defs, _Specs, _Typedefs, _Records1}) ->
    Options.

get_modules({program, _Path, Modules, _Defs, _Specs, _Typedefs, _Records1}) ->
    Modules.

set_modules(Modules, {program, Path, _Modules, Defs, Specs, Typedefs, Records1}) ->
    {program, Path, Modules, Defs, Specs, Typedefs, Records1}.


is_module_loaded(Name, Program) when is_atom(Name) ->
    gb_sets:is_member(Name, get_modules(Program)).

set_module_loaded(Name, Program) when is_atom(Name) ->
    Modules = get_modules(Program),
    Modules1 = gb_sets:add(Name, Modules),
    set_modules(Modules1, Program).

%: +func ensure_loaded :: atom() * program() -> !program().
ensure_loaded(M, Program) ->
    case is_module_loaded(M, Program) of
        true ->
            Program;
        false ->
            load_source_module(M, Program)
    end.

%: +func def :: fnid() * program() -> !def().		 
def(Fnid, 
    {program, _Path, _Modules, Defs, _Specs, _Typedefs, _Records}) ->
    {_, Def} = gb_trees:get(Fnid, Defs),
    Def.
    
%: +func spec :: fnid() * program() -> !spec().	
spec(Fnid,
     {program, _Path, _Modules, _Defs, Specs, _Typedefs, _Records}) ->
    {_, Spec} = gb_trees:get(Fnid, Specs),
    Spec.

%% record(Rid,
%%      {program, _Path, _Modules, _Defs, _Specs, _Typedefs, Records}) ->
%%     gb_trees:get(Rid, Records).

%: +func def_or_spec :: fnid() * program() -> def + spec + !none.
def_or_spec(Fnid = {fnid, M, _, _}, 
            Program = 
                {program, _Path, _Modules, Defs, Specs, _Typedefs, _Records}) ->
    Options = get_options(Program),
    PreferCode = 
        lists:member(M, proplists:get_all_values(prefer_code, Options)),
    Spec = gb_trees:is_defined(Fnid, Specs),
    Def  = gb_trees:is_defined(Fnid, Defs),
    if PreferCode ->
       if Def ->
              def;
          Spec ->
              spec;
          true ->
              none
       end;
       true ->
        if Spec ->
             spec;
           Def ->
             def;
          true ->
             none
        end
    end.

    %% UseSpec = 
    %%     lists:member(M, proplists:get_all_values(use_spec, Options)),
    %% PreferSpec = proplists:get_value(prefer_spec, Options, false),
    %% 
    %% Spec = gb_trees:is_defined(Fnid, Specs),
    %% Def  = gb_trees:is_defined(Fnid, Defs),
    %% if UseSpec  ->
    %%         if Spec ->
    %%                 spec;
    %%            true ->
    %%                 none
    %%         end;
    %%    true ->
    %%         if PreferSpec ->
    %%                 if Spec ->
    %%                         spec;
    %%                    Def ->
    %%                         def;
    %%                    true ->
    %%                         none
    %%                 end;
    %%            true ->
    %%                 if Def ->
    %%                         def;
    %%                    Spec ->
    %%                         spec;
    %%                    true ->
    %%                         none
    %%                 end
    %%         end
    %% end.
           
%: +func find_specs :: atom() * program() -> !list(!fnid()).
find_specs(Module,
           {program, _Path, _Modules, _Defs, Specs, _Typedefs, _Records}) ->
    S = [Fnid || Fnid = {fnid, Module1, _, _} 
                   <- gb_trees:keys(Specs), Module==Module1],
    S.

get_records(
  {program, _Path, _Modules, _Defs, _Specs, _Typedefs, Records}) ->
    Records.


%: +func typedef :: tyid() * program() -> !typedef().
typedef(Tyid,
     {program, _Path, _Modules, _Defs, _Specs, Typedefs, _Records}) ->
    {_, Typedef} = gb_trees:get(Tyid, Typedefs),
    Typedef.

%: +func keys :: program() 
%:            -> list({fnid, atom(), atom(), integer()} + {tyid, atom(), atom()}).
keys({program, _Path, _Modules, Defs, Specs, Typedefs, _Records}) ->
    gb_trees:keys(Defs) 
	++ gb_trees:keys(Specs) 
	++ gb_trees:keys(Typedefs).

add_def(Fnid, Def, {program, Path, Modules, Defs, Specs, Typedefs, Records}) ->
    Defs1 = gb_trees:insert(Fnid, {false, Def}, Defs),
    {program, Path, Modules, Defs1, Specs, Typedefs, Records}.

add_spec(Fnid, Spec, {program, Path, Modules, Defs, Specs, Typedefs, Records}) ->
    Specs1 = gb_trees:insert(Fnid, {false, Spec}, Specs),
    {program, Path, Modules, Defs, Specs1, Typedefs, Records}.

add_typedef(Tyid, Typedef, {program, Path, Modules, Defs, Specs, Typedefs, Records}) ->
    Typedefs1 = gb_trees:insert(Tyid, {false, Typedef}, Typedefs),
    {program, Path, Modules, Defs, Specs, Typedefs1, Records}.

add_record(Rid, MM, {program, Path, Modules, Defs, Specs, Typedefs, Records}) ->
    Records1 = gb_trees:insert(Rid, MM, Records),
    {program, Path, Modules, Defs, Specs, Typedefs, Records1}.


%: +func typedef_is_defined :: tyid() * program() -> bool().
typedef_is_defined(Tyid, 
    {program, _Path, _Modules, _Defs, _Specs, Typedefs, _Records}) ->
    gb_trees:is_defined(Tyid, Typedefs).

%: +func def_is_sane :: fnid() * atom() * program() -> bool().
def_is_sane(Fnid, M, 
	    {program, _Path, _Modules, Defs, _Specs, _Typedefs, _}) ->
    case gb_trees:lookup(Fnid, Defs) of
	{value, {Sane, _}} -> 
	    Sane;
	none ->
	    io:format("~nCannot find definition of ~w in module ~w", [Fnid, M]),
	    error("Missing def")
    end.

%: +func spec_is_sane :: fnid() * program() -> !bool().
spec_is_sane(Fnid,
     {program, _Path, _Modules, _Defs, Specs, _Typedefs, _Records}) ->
    {Sane, _} = gb_trees:get(Fnid, Specs),
    Sane.

%: +func typedef_is_sane :: tyid() * program() -> !bool().
typedef_is_sane(Tyid,
     {program, _Path, _Modules, _Defs, _Specs, Typedefs, _Records}) ->
    {Sane, _} = gb_trees:get(Tyid, Typedefs),
    Sane.

%: +func checked_def :: fnid() * program() -> !program().
checked_def(Fnid, {program, Path, Modules, Defs, Specs, Typedefs, Records}) ->
    {_, Def} = gb_trees:get(Fnid, Defs),
    Defs1 = gb_trees:update(Fnid, {true, Def}, Defs),
    {program, Path, Modules, Defs1, Specs, Typedefs, Records}.

%: +func checked_spec :: fnid() * program() -> !program().
checked_spec(Fnid, {program, Path, Modules, Defs, Specs, Typedefs, Records}) ->
    {_, Spec} = gb_trees:get(Fnid, Specs),
    Specs1 = gb_trees:update(Fnid, {true, Spec}, Specs),
    {program, Path, Modules, Defs, Specs1, Typedefs, Records}.

%: +func checked_typedef :: tyid() * program() -> !program().
checked_typedef(Tyid, {program, Path, Modules, Defs, Specs, Typedefs, Records}) ->
    {_, Typedef} = gb_trees:get(Tyid, Typedefs),
    Typedefs1 = gb_trees:update(Tyid, {true, Typedef}, Typedefs),
    {program, Path, Modules, Defs, Specs, Typedefs1, Records}.




file(FileName,  Module, Program) ->
    Options = get_options(Program),
    IncludePath = proplists:get_value(include_path, Options, []),
    %: promise IncludePath :: list(string()),
    {ok, Epp} = scfa_epp:open(FileName, IncludePath),

    case proplists:get_value(verbose, Options, false) of
        true ->
            io:format("~nOpen successful ~s", [FileName]);
        false ->
            true
    end,

    Program1 = process(Epp, Module, Program),
    scfa_epp:close(Epp),
    Program1.


process(Epp, Module, Program) ->
    case scfa_epp:parse_erl_form(Epp) of
	{ok, AbsForm} ->
            Program1 = form(AbsForm, Module, Program),
	    process(Epp, Module, Program1);
        {eof, _} ->
	    Program;
        {error, ErrorInfo} ->
            {Line, _M, Message} = ErrorInfo,
            io:format("~nError parsing module ~w, line ~w: ~s~n", 
                      [Module, Line, Message]),
            error("Parsing failed")            
    end.

%: +func form :: absyn:form() * atom() * program() -> !program().
form({attribute, _, author, _}, _, Program) ->
    Program;
form({attribute, _, behaviour, _}, _, Program) ->
    Program;
form({attribute, _, callback, _}, _, Program) ->
    Program;
form({attribute, _, compile, export_all}, _, Program) ->
    Program;
form({attribute, _, compile, {inline, _}}, _, Program) ->
    Program;
form({attribute, _, compile, {no_auto_import, _}}, _, Program) ->
    Program;
form({attribute, _, date, _}, _, Program) ->
    Program;
form({attribute, _, deprecated, _}, _, Program) ->
    Program;
form({attribute, _, dialyzer, _}, _, Program) ->
    Program;
form({attribute, _, export, _}, _, Program) ->
    Program;
form({attribute, _, export_type, _}, _, Program) -> % Interesting but not now...
    Program;
form({attribute, _, file, {_, _}}, _Module, Program) ->
    Program;
form({attribute, Line, import, {M0, FAList}}, M, Program) when is_atom(M0) ->
    import_fa_list(M0, FAList, Line, M, Program);
form({attribute, _, module, _}, _Module, Program) ->
    % assert?
    Program;
form({attribute, _, opaque, _}, _, Program) -> 
    Program;
form({attribute, _, optional_callbacks, _}, _, Program) ->
    Program;
form({attribute, _, record, Record}, M, Program) ->
    record_record(Record, M, Program);
form({attribute, _, spec, _}, _, Program) -> 
    Program;
form({attribute, _, type, _}, _, Program) -> 
    Program;
form({attribute, _, vsn, _}, _, Program) ->
    Program;
form(Typedef = {scfa_eq, _, {atom, _, type}, {atom, _, Name}, _, _TExp}, 
     Module, Program) ->
    Tyid = {tyid, Module, Name},
    add_typedef(Tyid, convert:type_def(Typedef, Module), Program);
form(Typedef = {scfa_eq, _, {atom, _, opaque}, {atom, _, Name}, _, _TExp},
    Module, Program) ->
    Tyid = {tyid, Module, Name},
    case get_current_module(Program) == Module of
        true -> 
	    io:format("~nCurrent...", []),
	    CTypeDef = convert:type_def(Typedef, Module);
	false -> CTypeDef = convert:type_def_external(Typedef, Module)
    end,
    add_typedef(Tyid, CTypeDef, Program);
form(Form = {scfa_ann, _, {atom, _,func}, {atom, _, Name}, TExp}, 
     Module, Program) ->
    Fnid = {fnid, Module, Name, fun_arity(TExp)},
    Spec = convert_spec(Fnid, Form, Module),
    add_spec(Fnid, Spec, Program);
%% form(Form = {scfa_ann, _, {atom, _,func}, {atom, _, Name}, _Phi, TExp}, 
%%      Module, Program) ->
%%     Fnid = {fnid, Module, Name, fun_arity(TExp)},
%%     Spec = convert_bounded_spec(Fnid, Form, Module),
%%     add_spec(Fnid, Spec, Program);

form(F={function, _Line, _Name, _Arity, _Clauses}, Module, Program) ->
    Records = get_records(Program),
    Def = {def, _, Fnid, _, _} = convert:fun_def(F, Module, Records),
    add_def(Fnid, Def, Program).


import_fa_list(_, [], _, _, Program) -> 
    Program;
import_fa_list(M0, [{F,A0}|L], Line, M, Program) 
  when is_atom(F), is_integer(A0) -> 
    A = A0,
%:  promise A :: integer(),
    Fnid = {fnid, M, F, A},
    Fnid0 = {fnid, M0, F, A},
    Program1 = add_def(Fnid, {import, Line, Fnid, Fnid0}, Program),
    import_fa_list(M0, L, Line, M, Program1).

convert_spec(Fnid, {scfa_ann, Line, _, _, TExp}, Module) ->
    RTE = convert:type_exp(TExp, Module),
    {spec, Line, Fnid, RTE}.

%% convert_bounded_spec(Fnid, {scfa_ann, Line, _, _, Phi, TExp}, Module) ->
%%     RPhi = convert:constraint(Phi, Module, pos:pos_zero()),
%%     RTE = convert:type_exp(TExp, Module),
%%     {spec_bounded, Line, Fnid, RPhi, RTE}.



%: +type record_field() = 
%:    {record_field, line(), {atom, line(), atom()}}
%:  + {record_field, line(), {atom, line(), atom()}, absyn:exp()}.

record_record({Name, FieldList0}, M, Program) when is_atom(Name) ->
    FieldList = FieldList0,
%:  promise FieldList :: list(record_field()),
    Rid = {rid, M, Name},
    FMapDMap = record_field_map(2, FieldList, gb_trees:empty(), gb_trees:empty()),
    add_record(Rid, FMapDMap, Program).

record_field_map(Index, [{record_field, Line, {atom, _, Huh}} | List], 
		 FMap, DMap) ->
    FMap1 = gb_trees:insert(Huh, Index, FMap),
    DMap1 = gb_trees:insert(Index, {atom, Line, undefined}, DMap),
    record_field_map(Index+1, List, FMap1, DMap1);
record_field_map(Index, 
                 [{record_field, _, {atom, _, Huh}, Dud} | List], 
                 FMap, DMap) ->
    FMap1 = gb_trees:insert(Huh, Index, FMap),
    DMap1 = gb_trees:insert(Index, Dud, DMap),
    record_field_map(Index+1, List, FMap1, DMap1);
record_field_map(Index, 
                 [{typed_record_field, 
		   {record_field, Line, {atom, _, Huh}}, _} | List], 
                 FMap, DMap) ->
    FMap1 = gb_trees:insert(Huh, Index, FMap),
    DMap1 = gb_trees:insert(Index, {atom, Line, undefined}, DMap),
    record_field_map(Index+1, List, FMap1, DMap1);
record_field_map(Index, 
                 [{typed_record_field, 
		   {record_field, _Line, {atom, _, Huh}, Dud}, _} | List], 
                 FMap, DMap) ->
    FMap1 = gb_trees:insert(Huh, Index, FMap),
    DMap1 = gb_trees:insert(Index, Dud, DMap),
    record_field_map(Index+1, List, FMap1, DMap1);
record_field_map(_, [], FMap, DMap) ->
    {FMap, DMap}.
    

% More later

fun_arity({arrow, _, ExpList, _}) ->
    length(ExpList);
fun_arity({'when', _, TExp, _}) ->
    fun_arity(TExp).

% Belongs in a different module...


