-module(reach).

%: +type list(X) = erlang:list(X).
%: +type deeplist(X) = erlang:deeplist(X).

%-compile(export_all).
-export([simplifyx/4, simplify_nostable/3]).

%: +func simplifyx 
%:     :: walker:pname() 
%:      * list(walker:name()) 
%:      * solver:graph() 
%:     -> poly:result(deeplist(walker:constraint())).
%: +func simplifyx 
%:     :: walker:pname() 
%:      * walker:list(walker:name()) 
%:      * solver:graph() 
%:      * any()
%:     -> poly:result(deeplist(walker:constraint())).
                       
simplifyx(Pred, NameList, Graph, _) ->
    simplifyx(Pred, NameList, Graph).
simplifyx(Pred, NameList, Graph) ->
    PList = [{var, Name} || Name <- NameList],
    {L, R} = reachability(PList, Graph),

    LS = left_supported(Graph, my_sets:to_list(L)),
    RC = right_covered(Graph, my_sets:to_list(R)),

    L1 = my_sets:from_list([Name || Name <- my_sets:to_list(L),
                                    my_sets:is_member(Name, RC)]),
    R1 = my_sets:from_list([Name || Name <- my_sets:to_list(R),
                                    my_sets:is_member(Name, LS)]),
    case solver:scan_watches(L, Graph) of
        {success, WL} ->
            G1 = [{watch, Name, M} || {Name, M} <- WL], 
            G2 = copy_graph(L1, R1, Graph),

            Stability = stability(NameList, {L, R}, Graph),
            G3 = stabilize(Pred, [G1, G2], Stability),
            {success, G3};
       {fail, Source} -> 
            {fail, Source}
    end.

views_are_reachable(Graph) ->
    [{var, Name} || Name = {view, _} <- solver:all_vars(Graph)].

%: +func simplify_nostable
%:     :: walker:name() 
%:      * walker:list(walker:name()) 
%:      * solver:graph() 
%:     -> poly:result(walker:dl(walker:constraint())).

simplify_nostable(_Pred, NameList, Graph) ->
    PList = [{var, Name} || Name <- NameList],
    {L, R} = reachability(PList, Graph),
    R_V = R, %views_are_reachable(R, Graph),

    LS = left_supported(Graph, my_sets:to_list(L)),
    RC = right_covered(Graph, my_sets:to_list(R_V)),

    L1 = my_sets:from_list([Name || Name <- my_sets:to_list(L),
				    my_sets:is_member(Name, RC)]),
    R1 = my_sets:from_list([Name || Name <- my_sets:to_list(R_V),
				    my_sets:is_member(Name, LS)]),
    case solver:scan_watches(L, Graph) of
        {success, WL} ->
            G1 = [{watch, Name, M} || {Name, M} <- WL], 
            G2 = copy_graph(L1, R1, Graph),
            {success, [G1, G2]};
       {fail, Source} -> 
            {fail, Source}
    end.


%: +func reachability 
%:  :: list(walker:texp0()) * solver:graph() 
%:  -> {my_sets:set(walker:name()), my_sets:set(walker:name())}.
reachability(PList, Graph) ->
    LVisit = my_sets:empty(), RVisit = my_sets:empty(),
    {LVisit1, RVisit1} = left_reachable_covers(PList, LVisit, RVisit, Graph),
    {LVisit2, RVisit2} =  right_reachable_supports(PList, LVisit1, RVisit1, Graph),
    V = views_are_reachable(Graph),
    right_reachable_supports(V, LVisit2, RVisit2, Graph).
    

left_reachable_var(Name, LVisit, RVisit, Graph) ->
    case my_sets:is_member(Name, LVisit) of
        true ->
            {LVisit, RVisit};
         false ->
            case solver:lookup_entry(Name, Graph) of
                none -> 
                    LVisit1 = my_sets:add(Name, LVisit),
                    {LVisit1, RVisit};
                {value, Entry} ->
		    LVisit1 = my_sets:add(Name, LVisit),

                    Uplinks = 
                        maps:keys(solver:entry_uplinks(Entry)),
                    {LVisit2, RVisit2} =
                        left_reachable_uplinks(Uplinks, LVisit1, RVisit, Graph),

                    Covers = [Cover||
                                 {Cover, _} <- solver:entry_covers(Entry)],
                    left_reachable_covers(Covers, LVisit2, RVisit2, Graph)
            end
    end.

left_reachable_uplinks([], LVisit, RVisit, _) ->
    {LVisit, RVisit};
left_reachable_uplinks([Name|Uplinks], LVisit, RVisit, Graph) ->    
    {LVisit1, RVisit1} = left_reachable_var(Name, LVisit, RVisit, Graph),
    left_reachable_uplinks(Uplinks, LVisit1, RVisit1, Graph).

left_reachable_covers([], LVisit, RVisit, _) ->
    {LVisit, RVisit};
left_reachable_covers([Cover|Covers], LVisit, RVisit, Graph) -> 
    {LVisit1, RVisit1} = left_reachable_cover(Cover, LVisit, RVisit, Graph),
    left_reachable_covers(Covers, LVisit1, RVisit1, Graph).

left_reachable_cover({c, C, EList}, LVisit, RVisit, Graph) ->
    {Covers, Supports} = conn:variance(C, EList),
    {LVisit1, RVisit1} = left_reachable_covers(Covers, LVisit, RVisit, Graph),
    right_reachable_supports(Supports, LVisit1, RVisit1, Graph);
left_reachable_cover({var, Name}, LVisit, RVisit, Graph) ->
    left_reachable_var(Name, LVisit, RVisit, Graph);
left_reachable_cover({sift, Name, _}, LVisit, RVisit, Graph) ->
    left_reachable_var(Name, LVisit, RVisit, Graph).


right_reachable_var(Name, LVisit, RVisit, Graph) ->
    case my_sets:is_member(Name, RVisit) of
        true ->
            {LVisit, RVisit};
        false ->
            case solver:lookup_entry(Name, Graph) of
                none ->
		    RVisit1 = my_sets:add(Name, RVisit),
                    {LVisit, RVisit1};
                {value, Entry} -> 
		    RVisit1 = my_sets:add(Name, RVisit),

                    Downlinks = 
                        maps:keys(solver:entry_downlinks(Entry)),
                
                    {LVisit2, RVisit2} =
                        right_reachable_downlinks(Downlinks, 
                                                  LVisit, RVisit1, Graph),
                    Supports = [Support||
                                   {Support, _} <- solver:entry_supports(Entry)],
                
                    right_reachable_supports(Supports, LVisit2, RVisit2, Graph)
            end
    end.

right_reachable_downlinks([], LVisit, RVisit, _) ->
    {LVisit, RVisit};
right_reachable_downlinks([Name|Downlinks], LVisit, RVisit, Graph) ->    
    {LVisit1, RVisit1} = right_reachable_var(Name, LVisit, RVisit, Graph),
    right_reachable_downlinks(Downlinks, LVisit1, RVisit1, Graph).

right_reachable_supports([], LVisit, RVisit, _) ->
    {LVisit, RVisit};
right_reachable_supports([Support|Supports], LVisit, RVisit, Graph) ->    
    {LVisit1, RVisit1} = right_reachable_support(Support, LVisit, RVisit, Graph),
    right_reachable_supports(Supports, LVisit1, RVisit1, Graph).

right_reachable_support({c, C, EList}, LVisit, RVisit, Graph) ->
    {Supports, Covers} = conn:variance(C, EList),
    {LVisit1, RVisit1} = left_reachable_covers(Covers, LVisit, RVisit, Graph),
    % io:format("~nVariance rr / RR supports: ~w", [Supports]),
    right_reachable_supports(Supports, LVisit1, RVisit1, Graph);
right_reachable_support({var, Name}, LVisit, RVisit, Graph) ->
    right_reachable_var(Name, LVisit, RVisit, Graph);
right_reachable_support({sift, Name, _}, LVisit, RVisit, Graph) ->
    right_reachable_var(Name, LVisit, RVisit, Graph).

%%%%

copy_graph(LVisit, RVisit, Graph) ->
    lists:flatten(
      [copy_left_reachable(my_sets:to_list(LVisit), Graph),
       copy_right_reachable(my_sets:to_list(RVisit), Graph)]).

copy_left_reachable([Name|LVisit], Graph) ->
    case solver:is_present_var(Name, Graph) of
        false ->
            copy_left_reachable(LVisit, Graph);
        true ->
            [copy_uplinks(Name, Graph),
             copy_covers(Name, Graph) |
             copy_left_reachable(LVisit, Graph)]
    end;
copy_left_reachable([], _) ->
    [].

copy_downlinks(Name, Graph) ->
     Entry = solver:get_entry(Name, Graph),
     Downlinks = maps:to_list(solver:entry_downlinks(Entry)),
     [{include, sift( Down, F), {var, Name}, Source} || 
         {Down, {F, Source}} <- Downlinks].
copy_supports(Name, Graph) ->
     Entry = solver:get_entry(Name, Graph),
     Supports = solver:entry_supports(Entry),
     [{include, E, {var, Name}, Source} || {E, Source} <- Supports].


copy_right_reachable([Name|RVisit], Graph) ->
    case solver:is_present_var(Name, Graph) of
        false ->
            copy_right_reachable(RVisit, Graph);
        true ->
            [copy_downlinks(Name, Graph),
             copy_supports(Name, Graph),
             copy_right_reachable(RVisit, Graph)]
    end;
copy_right_reachable([], _) ->
    [].

copy_uplinks(Name, Graph) ->
     Entry = solver:get_entry(Name, Graph),
     Uplinks = maps:to_list(solver:entry_uplinks(Entry)),
     [{include, sift(Name, F), {var, Up}, Source} ||
         {Up, {F, Source}} <- Uplinks].

copy_covers(Name, Graph) ->
     Entry = solver:get_entry(Name, Graph),
     Covers = solver:entry_covers(Entry),
     [{include, {sift, Name, rfilter:singleton(solver:cexp_c(E))}, E, Source} || 
         {E, Source} <- Covers].


sift(Name, one) ->
    {var, Name};
sift(Name, F) ->
    {sift, Name, F}.

%%% Cover and support

right_covered(Graph, RR) ->
    SC = simply_right_covered(Graph),                   
    right_covered(Graph, RR++SC, my_sets:empty()).

simply_right_covered({WL, Map}) ->
    L = maps:to_list(Map),
    [Name||{Name, Entry} <- L,
           ([] /= solver:entry_covers(Entry)
            orelse Name == zero) 
               orelse is_view(Name)]
        ++ [W || {W, _} <- WL].

is_view({view, _}) ->
    true;
is_view(_) ->
    false.

right_covered(_, [], Set) ->
    Set;
right_covered(Graph, [Name|Queue], Set) ->
    case my_sets:is_member(Name, Set) of
        true ->
            right_covered(Graph, Queue, Set);
        false ->
            Set1 = my_sets:add(Name, Set),
            case 
            solver:lookup_entry(Name, Graph) of
                none ->
                    right_covered(Graph, Queue, Set1);
                {value, Entry} ->
                    LeftList = 
                        [Left ||
                            {Left, _} <- 
                                maps:to_list(solver:entry_downlinks(Entry))],
                    right_covered(Graph, LeftList++Queue, Set1)  
            end
  end.


left_supported(Graph, RR) ->
    SC = simply_left_supported(Graph),                   
    left_supported(Graph, RR++SC, my_sets:empty()).


simply_left_supported({_, Map}) ->
    L = maps:to_list(Map),
    [Name||{Name, Entry} <- L,
           [] /= solver:entry_supports(Entry)
    ].

left_supported(_, [], Set) ->
    Set;
left_supported(Graph, [Name|Queue], Set) ->
    case my_sets:is_member(Name, Set) of
        true ->
            left_supported(Graph, Queue, Set);
        false ->
            Set1 = my_sets:add(Name, Set),
            case solver:lookup_entry(Name, Graph) of
                none ->
                    left_supported(Graph, Queue, Set1);
                {value, Entry} ->
                    RightList = 
                        [Right ||
                            {Right, _} 
                                <- maps:to_list(solver:entry_uplinks(Entry))],
                    left_supported(Graph, RightList++Queue, Set1)  
            end
    end.

% Activity and stability


stability(_NameList, Reachability, Graph) ->
    {LV, RV} = Reachability,
    stability1(LV, RV, Graph).   

stability1(LV, RV, Graph) ->
    %% {LA, RA} = activity(LV, RV, Graph),
    %% io:format("~nLA: ~w", [gb_sets:to_list(LA)]),
    %% io:format("~nRA: ~w", [gb_sets:to_list(RA)]),

    {LA1, RA1} = compute_activity(LV, RV, Graph),

    LS=stability_names(my_sets:to_list(LV), my_sets:empty(), LA1, RA1, LV, RV),
    RS=stability_names(my_sets:to_list(RV), my_sets:empty(), LA1, RA1, LV, RV),
    my_sets:union(LS, RS).

stability_names([], S, _, _, _, _) ->
    S;
stability_names([Name|Names], S, LA, RA, LV, RV) ->
    case (my_sets:is_member(Name, RA) 
          andalso my_sets:is_member(Name, LV))
        orelse
          (my_sets:is_member(Name, LA) 
           andalso my_sets:is_member(Name, RV)) of
        true ->
            S1 = S;
        false ->
            S1 = my_sets:add(Name, S)
    end,
    stability_names(Names, S1, LA, RA, LV, RV).


stabilize_name(_, Name = {stable, _, _}, _) ->
    Name;
stabilize_name(_, zero, _) ->
    zero;
stabilize_name(Pred, Name, Stable) ->
    case my_sets:is_member(Name, Stable) of
        true ->
            {stable, Pred, Name};
        false ->
            Name
    end.

stabilize(Pred, Phi, Stable) ->
    [stabilize_1(Pred, Phi, Stable)].

stabilize_1(_, [], _) ->
    [];
stabilize_1(Pred, [Phi1|Phi2], Stable) ->
    [stabilize_1(Pred, Phi1, Stable) | 
     stabilize_1(Pred, Phi2, Stable)];
stabilize_1(Pred, {include, E1, E2, Source}, Stable)->
    E1s = stabilize_exp(Pred, E1, Stable),
    E2s = stabilize_exp0(Pred, E2, Stable),
    {include, E1s, E2s, Source};
stabilize_1(Pred, {watch, Name, Message}, Stable) ->
    Name1 = stabilize_name(Pred, Name, Stable),
    {watch, Name1, Message}.


stabilize_exp(Pred, {var, Name}, Stable)  ->
    Name1 = stabilize_name(Pred, Name, Stable),
    {var, Name1};
stabilize_exp(Pred, {c, C, EList}, Stable) ->
    EListS = stabilize_exp_list(Pred, EList, Stable),
    {c, C, EListS};
stabilize_exp(Pred, {sift, N, F}, Stable) ->
    Ns = stabilize_name(Pred, N, Stable),
    {sift, Ns, F}.


stabilize_exp0(Pred, {var, Name}, Stable)  ->
    Name1 = stabilize_name(Pred, Name, Stable),
    {var, Name1};
stabilize_exp0(Pred, {c, C, EList}, Stable) ->
    EListS = stabilize_exp_list(Pred, EList, Stable),
    {c, C, EListS}.

stabilize_exp_list(Pred, EList, Stable)   ->
    [stabilize_exp0(Pred, E, Stable) || E <- EList].


% activity, take two.

compute_activity(LV, RV, Graph) ->
    NameList = solver:all_vars(Graph),
    AG = make_activity(NameList, Graph),
    A = my_sets:empty(),
    A1 = left_active_vars(LV, A, AG),
    A2 = right_active_vars(RV, A1, AG),

    AList = my_sets:to_list(A2),
    LA = [Name || {la, Name} <- AList],
    RA = [Name || {ra, Name} <- AList],
    {my_sets:from_list(LA), my_sets:from_list(RA)}.

left_active_vars(Vars, A, AG) ->
     Nodes = [{la, Name} || Name <- my_sets:to_list(Vars)],
     ag_traverse_nodes(Nodes, A, AG).    

right_active_vars(Vars, A, AG) ->
     Nodes = [{ra, Name} || Name <- my_sets:to_list(Vars)],
     ag_traverse_nodes(Nodes, A, AG).
     
ag_traverse_node(Node, A, AG) ->
   case my_sets:is_member(Node, A) of
       true ->
           A;
       false ->
           A1 = my_sets:add(Node, A),
           case maps:find(Node, AG) of
               error ->
                   A1;
               {ok, Children} ->
                   List = my_sets:to_list(Children),
                   ag_traverse_nodes(List, A1, AG)
           end
   end.

ag_traverse_nodes([], A, _) ->
    A;
ag_traverse_nodes([Node|Nodes], A, AG) ->
    A1 = ag_traverse_node(Node, A, AG),
    ag_traverse_nodes(Nodes, A1, AG).

%: +type activity_node() = {la, walker:name()} + {ra, walker:name()}.
%: +type activity_graph() 
%:    = map(activity_node(), my_sets:set(activity_node())).
%: +func make_activity :: list(walker:name()) * solver:graph() -> activity_graph().
make_activity(VarList, Graph) ->
    Activity = maps:new(),
    make_activity(VarList, Activity, Graph).

%: +func make_activity :: list(walker:name()) * activity_graph() * solver:graph() -> activity_graph().
make_activity([], Activity, _) ->
    Activity;
make_activity([Var|VarList], Activity, Graph) ->
    {value, Entry} = solver:lookup_entry(Var, Graph),
    Supports = [Support||
                {Support, _} <- solver:entry_supports(Entry)],
    Activity1 = activity_lefts(Supports, {la, Var}, Activity),
    Downlinks = maps:keys(solver:entry_downlinks(Entry)),
    Activity2 = activity_left_downlinks(Downlinks, {la, Var}, Activity1),
    Covers = [Cover ||
                    {Cover, _} <- solver:entry_covers(Entry)],
    Activity3 = activity_rights({ra, Var}, Covers, Activity2),
    Uplinks = maps:keys(solver:entry_uplinks(Entry)),

    Activity4 = activity_right_uplinks(Uplinks, {ra, Var}, Activity3),

    make_activity(VarList, Activity4, Graph).


activity_lefts([], _, Activity) ->
   Activity;
activity_lefts([Support|Supports], Node, Activity) ->
   Activity1 = activity_left(Support, Node, Activity),
   activity_lefts(Supports, Node, Activity1).

activity_left({var, X}, Node, Activity) ->
    activity_add_edge({la, X}, Node, Activity);
activity_left({sift, X, _}, Node, Activity) ->
   activity_add_edge({la, X}, Node, Activity);
activity_left({c, C, EList}, Node, Activity) ->
    {Supports, Covers} = conn:variance(C, EList),
    Activity1 = activity_lefts(Supports, Node, Activity),
    activity_rights(Node, Covers, Activity1).

activity_left_downlinks([], _, Activity) ->
    Activity;
activity_left_downlinks([Name|NameList], Node, Activity) ->
    Activity1=activity_add_edge({la, Name}, Node, Activity),
    activity_left_downlinks(NameList, Node, Activity1).



activity_rights(_, [], Activity) ->
   Activity;
activity_rights(Node, [Cover|Covers], Activity) ->
   Activity1 = activity_right(Node, Cover, Activity),
   activity_rights(Node, Covers, Activity1).

activity_right(Node, {var, Y}, Activity) ->
   activity_add_edge({ra, Y}, Node, Activity);
activity_right(Node, {sift, X, _}, Activity) ->
   activity_add_edge({ra, X}, Node, Activity);
activity_right(Node, {c, C, EList}, Activity) ->
    {Covers, Supports} = conn:variance(C, EList),
    Activity1 = activity_rights(Node, Covers, Activity),
    activity_lefts(Supports, Node, Activity1).

activity_right_uplinks([], _, Activity) ->
    Activity;
activity_right_uplinks([Name|NameList], Node, Activity) ->
    Activity1=activity_add_edge({ra, Name}, Node, Activity),
    activity_right_uplinks(NameList, Node, Activity1).


activity_add_edge(Node1, Node2, Activity) ->
    case maps:find(Node1, Activity) of
        error ->
            Children = my_sets:add(Node2, my_sets:empty()),
            maps:put(Node1, Children, Activity);
        {ok, Children} ->
            Children1 = my_sets:add(Node2, Children),
            maps:put(Node1, Children1, Activity)
    end.                                  %




            



            


