-module(record_expand).

-export([exp/3, pattern/3]).

%: +type list(X) = erlang:list(X).

% Build a record.

%: +func exp :: absyn:exp() * program:recorddef() * convert:pos() 
%:              -> absyn:exp().
                  
exp({record, Line, Name, FieldList}, {FMap, DMap}, _Pos) ->
    Size = gb_trees:size(DMap)+1,	    
    Atom = {atom, Line, Name},
    IPMap = fields(FieldList, FMap, gb_trees:empty()),
    FList = record_tuple(2, Size, DMap, IPMap),
    {tuple, Line, [Atom|FList]};
exp({record, Line, Exp, Name, FieldList}, {FMap, DMap}, Pos) ->
    IPMap = make_var_map(2, gb_trees:keys(FMap), Line, Pos, gb_trees:empty()),
    Size = gb_trees:size(DMap)+1,	    
    Atom = {atom, Line, Name},

    FList = record_tuple(2, Size, IPMap),
    %: promise FList :: list(absyn:pattern()),
    Match = {match, Line, {tuple, Line, [Atom|FList]}, Exp},
    IPMap1 = update_fields(FieldList,  FMap, IPMap),
    FList1 = record_tuple(2, Size, IPMap1),
    Res = {tuple, Line, [Atom|FList1]},
    {block, Line,  [Match, Res]};

% Access a field of a record.

exp({record_field, Line, Exp, Name, {atom, _, Field}}, {FMap, _}, Pos) ->
    New = new_var(Line, "record", pos:left(Pos)),
    Index = gb_trees:get(Field, FMap),
    Size = gb_trees:size(FMap)+1,
    L = make_tuple_pattern_list(2, Index, Size, New, Line, Pos),
    {block, Line, 
        [{match, Line, {tuple, Line, [{atom, Line, Name} | L]}, Exp},
    	 New]}.


%: +func make_var_map 
%:     :: integer()
%:      * list(absyn:name())
%:      * absyn:line()
%:      * convert:pos()
%:      * gb_trees:tree(integer(), absyn:exp())
%:     -> gb_trees:tree(integer(), absyn:exp()).
           
make_var_map(_, [], _, _, IPMap) ->
    IPMap;
make_var_map(Index, [Name|L], Line, Pos, IPMap) ->
    IPMap1 = gb_trees:insert(Index, new_var(Line, Name, pos:left(Pos)), IPMap),
    make_var_map(Index+1, L, Line, pos:right(Pos), IPMap1).


% Build a record or match a record.

fields([{record_field, _, {atom, _, Name}, T}|FL], FMap, IPMap) ->
    Index = gb_trees:get(Name, FMap),
    IPMap1 = gb_trees:insert(Index, T, IPMap),
    fields(FL,  FMap, IPMap1);
fields([], _, IPMap) ->
    IPMap.

% Make a list of tuple elements.

record_tuple(Index, Size, DMap, IPMap) when Index =< Size ->
    case gb_trees:lookup(Index, IPMap) of
        {value, Exp} ->
            true;
        none ->
            Exp = gb_trees:get(Index, DMap)
    end,
    [Exp|record_tuple(Index+1, Size, DMap, IPMap)];
record_tuple(Index, Size, _, _) when Size < Index ->    
    [].


% Access a field of a record.

%: +func make_tuple_pattern_list 
%:       :: integer() 
%:        * integer() 
%:        * integer() 
%:        * absyn:pattern() 
%:        * absyn:line() 
%:        * convert:pos()    
%:     -> list(absyn:pattern()).
  
make_tuple_pattern_list(I, Place, Size, New, Line, Pos)
  when I =/= Place, I=<Size ->
    [{var, Line, '_'}|make_tuple_pattern_list(I+1, Place, Size, New, Line, Pos)];
make_tuple_pattern_list(I, Place, Size, New, Line, Pos)
  when I == Place, I=<Size ->
    [New|make_tuple_pattern_list(I+1, Place, Size, New, Line, Pos)];
make_tuple_pattern_list(I, _Place, Size, _New, _Line, _Pos) 
  when I>Size ->
    [].

% Records in patterns

%: +func pattern :: absyn:pattern()*program:recorddef()*convert:pos() 
%: -> absyn:pattern(). 
pattern({record, Line, Name, FieldList}, {FMap, _}, Pos) ->
    Size = gb_trees:size(FMap) +1,
    Atom = {atom, Line, Name},
    IPMap = fields(FieldList, FMap, gb_trees:empty()),
    Var = new_var(Line, '_', Pos),
    FList = pattern_record_tuple(2, Size, Var, IPMap),
    {tuple, Line, [Atom|FList]}.

pattern_record_tuple(Index, Size, Var, IPMap) when Index =< Size ->
    case gb_trees:lookup(Index, IPMap) of
        {value, Exp} ->
            true;
        none ->
	    Exp = Var
    end,
    [Exp|pattern_record_tuple(Index+1, Size, Var, IPMap)];
pattern_record_tuple(Index, Size, _, _) when Size < Index ->    
    [].

%
record_tuple(Index, Size, IPMap) when Index =< Size ->
    Exp = gb_trees:get(Index, IPMap),
    [Exp|record_tuple(Index+1, Size, IPMap)];
record_tuple(Index, Size, _) when Size < Index ->    
    [].

% update a record

update_fields([{record_field, _, {atom, _, Name}, T}|FL], FMap, IPMap) ->
    Index = gb_trees:get(Name, FMap),
    IPMap1 = gb_trees:update(Index, T, IPMap),
    update_fields(FL,  FMap, IPMap1);
update_fields([], _, IPMap) ->
    IPMap.

%

new_var(Line, S, Pos) ->
    String = lists:flatten(io_lib:format("~s$~w", [S, Pos])),
    %: promise String :: list(integer()),
    Name = list_to_atom(String),
    {var, Line, Name}.



