-module(rfilter).

%-compile(export_all).
%: +opaque filter() = list(item()).
%: +type item() = {conn:con(), filter()}.
%: +type list(X) = erlang:list(X).
%: +type bool() = erlang:bool().

-export([singleton/1, singleton_inverse/1, empty/0, any/0]).
-export([union/2, intersect/2, included/2]).
-export([minus/2, is_empty/1]).
-export([overlap/2]).
%-export([contains/2]).

%: +func singleton :: conn:con() -> filter().
singleton(C) ->
    [{C, []}].

%: +func singleton_inverse :: conn:con() -> filter().
singleton_inverse(C) ->
    case conn:is_any(C) of
	true ->
	    [];
	false ->
	    [{conn:any(), [{C, []}]}]
    end.

%: +func empty :: () -> filter().
empty() ->
    [].

%: +func any :: () -> filter().
any() ->
    singleton(conn:any()).

%: +func is_empty :: filter() -> bool().
is_empty([]) ->
    true;
is_empty(_) ->
    false.

%: +func union :: filter() * filter() -> filter().
union([], F) -> F;
union(F, []) -> F;
union([I1 = {C1, D1}|F1], [I2 = {C2, D2}|F2]) ->
%    io:format("~nunion ~w ~w", [[I1|F1], [I2|F2]]),
   case {conn:con_included(C1, C2), conn:con_included(C2, C1)} of
       {true, true} -> % identical
           [{C1, intersect(D1, D2)} | union(F1, F2)];
       {true, false} ->
       	   union(F1, [{C2, minus(D2, [I1])}|F2]);
       {false, true} ->
       	   union([{C1, minus(D1, [I2])}|F1], F2);
       {false, false} ->
           case conn:con_ordered(C1, C2) of
               true ->
                    [{C1, D1} | union(F1, [{C2, D2} | F2])];
               false ->
                    [{C2, D2} | union([{C1, D1} | F1], F2)]
           end
    end.

%: +func intersect :: filter() * filter() -> filter().
intersect([], _) -> [];
intersect(_, []) -> [];
intersect([I1 = {C1, D1}|F1], [I2 = {C2, D2}|F2]) ->
%    io:format("~nintersect ~w ~w", [[I1|F1], [I2|F2]]),
    case {conn:con_included(C1, C2), conn:con_included(C2, C1)} of
	{true, true} -> % identical
	    [{C1, union(D1, D2)} | intersect(F1, F2)];
	{true, false} ->
	    minus([I1], D2, intersect(F1, [I2|F2]));
	{false, true} ->
	    minus([I2], D1, intersect([I1|F1], F2));
	{false, false} ->
            case conn:con_ordered(C1, C2) of
               true ->
                    intersect(F1, [{C2, D2} | F2]);
               false ->
                    intersect([{C1, D1}|F1], F2)
           end
     end.

%: +func minus :: filter() * filter() -> filter().
minus(F, []) ->
    F;
minus([], _) ->
    [];
minus([I1={C1, D1}|F1], [I2 = {C2, D2}|F2]) ->
%    io:format("~nminus ~w ~w", [[I1|F1], [I2|F2]]),
    case {conn:con_included(C1, C2), conn:con_included(C2, C1)} of
	{true, true} -> % identical
	    minus(D2, D1, minus(F1, F2));
	{true, false} ->
            minus(F1, [I2|F2]);
	{false, true} ->
            minus([{C1, union(D1, [I2])}| F1], F2);
	{false, false} ->
            case conn:con_ordered(C1, C2) of
               true ->
                    [I1| minus(F1, [I2 | F2])];
               false ->
                    minus([I1|F1], F2)
           end
    end. 
% Keep in mind that the accumulator should only contain constructors
% not in the first argument.
minus(F, [], Acc) ->
    F++Acc;
minus([], _, Acc) ->
    Acc;
minus([I1={C1, D1}|F1], [I2 = {C2, D2}|F2], Acc) ->
%    io:format("~nminus ~w ~w", [[I1|F1], [I2|F2]]),
    case {conn:con_included(C1, C2), conn:con_included(C2, C1)} of
	{true, true} -> % identical
	    minus(D2, D1, minus(F1, F2, Acc));
	{true, false} ->
            minus(F1, [I2|F2], Acc);
	{false, true} ->
            minus([{C1, union(D1, [I2])}| F1], F2, Acc);
	{false, false} ->
            case conn:con_ordered(C1, C2) of
               true ->
                    [I1| minus(F1, [I2 | F2], Acc)];
               false ->
                    minus([I1|F1], F2, Acc)
           end
    end.


%: +func included :: filter()*filter() -> bool().
included([], _) -> true;
included(_, []) -> false;
included([I1={C1, D1}|F1], [I2={C2, D2}|F2]) ->
%    io:format("~nincluded ~w ~w", [[I1|F1], [I2|F2]]),
   case {conn:con_included(C1, C2), conn:con_included(C2, C1)} of
       {true, true} -> % identical
           included(D2, D1) 
	       andalso included(F1, F2);
       {true, false} ->
           included(intersect([{C1, []}], D2), D1) andalso
	       included(F1,[I2|F2]);             
       {false, true} ->
           false;
       {false, false} ->
           case conn:con_ordered(C1, C2) of
               true ->
                   false;
               false ->
                    included([I1 | F1],  F2)
           end
    end.

%% %: +func contains :: conn:con() * filter()-> bool().
%% contains(_, []) ->
%%     false;
%% contains(C1, [{C2, D1}|D2]) ->
%%     case {conn:con_included(C1, C2), conn:con_included(C2, C1)} of
%%        {true, true} -> % identical
%%            true;
%%        {true, false} ->
%%            not contains(C1, D1);
%%        {false, true} ->
%%            false;
%%        {false, false} ->
%%            case conn:con_ordered(C1, C2) of
%%                true ->
%%                    false;
%%                false ->
%%                     contains(C1,  D2)
%%            end
%%     end.

overlap(F1, F2) ->
    not is_empty(intersect(F1, F2)).
