-module(sanity).

-export([complete_module/3, ensure_def/2, ensure_defs/2,
         ensure_spec/2, ensure_specs/2, ensure_typedef/2]).

%-compile(export_all).

%: +type fnid() = {fnid, atom(), atom(), integer()}.
%: +type list(X) = erlang:list(X).
%: +type options() = program:options().
%: +type path() = program:path().
%: +type program() = program:program().

%: +func complete_module :: atom() * path() * options() 
%:                            -> {list(fnid()), program()}.
complete_module(Module, Path, Options) ->
    P = program:new(Module, Path, Options),
    P1 = program:ensure_loaded(Module, P),
    FnidList = program:find_specs(Module, P1),   
    P2 = sanity:ensure_specs(FnidList, P1),
    P3 = sanity:ensure_defs(FnidList, P2),
    {FnidList, P3}.

%: +func ensure_def :: program:fnid() * program:program() -> program:program().
ensure_def(Fnid={fnid, M, _, _}, Program) ->
    Program1 = program:ensure_loaded(M, Program),
    case program:def_is_sane(Fnid, M, Program1) of
	true ->
	    Program1;
	false ->
	    case program:def(Fnid, Program1) of
		Def = {def, _, _, _, _} ->
		    Program2 = program:checked_def(Fnid, Program1),
		    ensure_def(Fnid, Def, Program2)
	    end
    end.

%: +func ensure_spec :: program:fnid() * program:program() -> program:program().
ensure_spec(Fnid={fnid, M, _, _}, Program) ->
    Program1 = program:ensure_loaded(M, Program),
    case program:spec_is_sane(Fnid, Program1) of
	true ->
	    Program1;
	false ->

	    Spec = program:spec(Fnid, Program1),
	    Program2 = program:checked_spec(Fnid, Program1),
	    ensure_spec(Fnid, Spec, Program2)
    end.
    
ensure_def_or_spec(Fnid={fnid, M, _, _}, Line, Module, Program) ->
    Program1 = program:ensure_loaded(M, Program),
%    io:format("~nDef or spec ~w ~w", [Fnid, program:def_or_spec(Fnid, Program1)]),
    case program:def_or_spec(Fnid, Program1) of
        def ->
	    ensure_def(Fnid, Program1);
	spec ->
	    ensure_spec(Fnid, Program1);
	none ->
            io:format("~n Function ~w referenced in module ~w, line ~w is "
                      ++ "not defined or specified",          
                      [Fnid, Module, Line]),                  
	    error("missing definition")
   end.

%: +func ensure_specs :: program:list(program:fnid()) * program:program() -> program:program().
ensure_specs([], Program) ->
    Program;
ensure_specs([Fnid|FnidList], Program) ->
    Program1 = ensure_spec(Fnid, Program),
    ensure_specs(FnidList, Program1).

%: +func ensure_defs :: program:list(program:fnid()) * program:program() -> program:program().
ensure_defs([], Program) ->
    Program;
ensure_defs([Fnid|FnidList], Program) ->
    Program1 = ensure_def(Fnid, Program),
    ensure_defs(FnidList, Program1).

%: +func ensure_typedef :: program:tyid() * program:program() -> program:program().
ensure_typedef(Tyid={tyid, M, _}, Program) ->
    Program1 = program:ensure_loaded(M, Program),
    case program:typedef_is_defined(Tyid, Program1) of
        false ->
            myerror("Type ~s is not defined", 
                    [show_tyid(Tyid)]);
        true ->
            case program:typedef_is_sane(Tyid, Program1) of
	        true ->
	            Program1;
	        false ->
	            Typedef = program:typedef(Tyid, Program1),
	            Program2 = program:checked_typedef(Tyid, Program1),
	            ensure_typedef(Tyid, Typedef, Program2)
            end
    end.

myerror(M, A) ->
    error(lists:flatten(io_lib:format(M, A))).

show_tyid({tyid, M, F}) ->
    io_lib:format("~w:~w", [M, F]).

ensure_def({fnid, M, _F, _A}, {def, _, _, _, E}, Program) ->
    ensure_exp(E, M, Program).

ensure_spec({fnid, M, _, A}, {spec, _, _, TExp}, Program) ->
    ensure_spec_body(TExp, A, M, Program);
ensure_spec({fnid, M, _, A}, {spec_bounded, _, _, Phi, TExp}, Program) ->
    Program1 = ensure_constraint(Phi, M, Program),
    ensure_spec_body(TExp, A, M, Program1).

ensure_constraint({'=<', _, TExp1, TExp2}, M, Program) -> 
    Program1 = ensure_texp(TExp1, M, Program),    
    ensure_texp(TExp2, M, Program1);
ensure_constraint({'and', _, Constraint1, Constraint2}, M, Program) -> 
    Program1 = ensure_constraint(Constraint1, M, Program),
    ensure_constraint(Constraint2, M, Program1).

ensure_spec_body(TExp={c, {fun_fixed, A}, _}, A, M, Program) ->
    ensure_texp(TExp, M, Program).    



ensure_typedef({tyid, M, _}, {typedef, _, _Name, _, TExp},
	    Program) ->
    ensure_texp(TExp, M, Program).




% Program code

ensure_body([E], M, Program) ->
    ensure_exp(E, M, Program);
ensure_body([E|Body], M, Program) ->
    Program1 = ensure_exp(E, M, Program),
    ensure_body(Body, M, Program1).

%% Type checker is faster with the stuff below commented out.
%% %: +func ensure_exp :: convert:exp() * atom() * program:program() 
%% %:    -> program:program().
ensure_exp({var, _}, _M, Program) ->
    Program;
ensure_exp({c,_, ExpList}, M, Program) ->
    ensure_exp_list(ExpList, M, Program);
ensure_exp({op, _, Exp1, Exp2}, M, Program) ->
    ensure_exp_list([Exp1, Exp2], M, Program);
ensure_exp({op, _, _, Exp1, Exp2}, M, Program) ->
    ensure_exp_list([Exp1, Exp2], M, Program);
ensure_exp({eq, Exp1, Exp2}, M, Program) ->
    ensure_exp_list([Exp1, Exp2], M, Program);
ensure_exp({match, _, _, Exp}, M, Program) ->
    ensure_exp(Exp, M, Program);
ensure_exp({ccase, _Line, KeyList, ClauseList, _}, M, Program) ->
    Program1 = ensure_exp_list(KeyList, M, Program),
    ensure_clause_list(ClauseList, M, Program1);
ensure_exp({lambda, _, Exp}, M, Program) ->
    ensure_exp(Exp, M, Program);   
ensure_exp({lfun, _, Fnid}, _, Program) ->
    ensure_def(Fnid, Program);
ensure_exp({local_call, Line, Fnid, ExpList}, M, Program) ->
    Program1 = ensure_exp_list(ExpList, M, Program),
    ensure_def_or_spec(Fnid, Line, M, Program1);
ensure_exp({call, Line, Fnid, ExpList}, M, Program) ->
    Program1 = ensure_exp_list(ExpList, M, Program),
    ensure_def_or_spec(Fnid, Line, M, Program1);
ensure_exp({apply, _, Exp, ExpList}, M, Program) ->
    ensure_exp_list([Exp|ExpList], M, Program);
ensure_exp({lc, _, Exp, GenList}, M, Program) ->
    Program1 = ensure_exp(Exp, M, Program),
    ensure_gen_list(GenList, M, Program1);
ensure_exp({promise, _, Exp, TExp}, M, Program) ->
    Program1 = ensure_exp(Exp, M, Program), 
    ensure_texp(TExp, M, Program1);
ensure_exp({block, _, Block}, M, Program) ->
    ensure_exp_list(Block, M, Program);   
ensure_exp({'try', _, Body, CRClauseList, ClauseList, After}, M, Program) ->
    Program1 = ensure_exp_list(Body, M, Program),
    Program2 = ensure_clause_list(CRClauseList, M, Program1),
    Program3 = ensure_clause_list(ClauseList, M, Program2),
    ensure_exp_list(After, M, Program3);
ensure_exp({'catch', _, Exp}, M, Program) ->
    ensure_exp(Exp, M, Program);
ensure_exp({'receive', _, ClauseList, _}, M, Program) ->
    ensure_clause_list(ClauseList, M, Program);
ensure_exp({'receive', _, ClauseList, _, After, Final}, M, Program) ->
    Program1 = ensure_clause_list(ClauseList, M, Program),
    Program2 = ensure_exp(After, M, Program1),
    ensure_exp_list(Final, M, Program2);
ensure_exp({check, E}, M, Program) ->
    ensure_exp(E, M, Program).


ensure_exp_list([], _, Program) ->
    Program;
ensure_exp_list([E|EList], M, Program) ->
    Program1 = ensure_exp(E, M, Program),
    ensure_exp_list(EList, M, Program1).

ensure_gen_list([], _, Program) ->
    Program;
ensure_gen_list([Gen|GenList], M, Program) ->
    Program1 = ensure_gen(Gen, M, Program),
    ensure_gen_list(GenList, M, Program1).

ensure_gen({generate, _, _Pattern, Exp}, M, Program) ->
    ensure_exp(Exp, M, Program);
ensure_gen({b_generate, _, _Pattern, Exp}, M, Program) ->
    ensure_exp(Exp, M, Program);
ensure_gen(Exp, M, Program) ->
    ensure_exp(Exp, M, Program).


ensure_clause({clause, _, _PairList, Body}, M, Program) ->
    ensure_body(Body, M, Program).

%% %: +func ensure_clause_list 
%% %:    :: convert:list(convert:cclause()) * atom() * program:program() 
%% %:    -> program:program().
ensure_clause_list([], _M, Program) ->
    Program;
ensure_clause_list([Clause|ClauseList], M, Program) ->
    Program1 = ensure_clause(Clause, M, Program),
    ensure_clause_list(ClauseList, M, Program1).

% Type expressions

ensure_texp({var, _}, _, Program) ->
    Program;
ensure_texp({alias, _, Var, TE}, M, Program) ->
    ensure_texp(Var, M, Program),
    ensure_texp(TE, M, Program);
ensure_texp({c, _, TEList}, M, Program) ->
    ensure_texp_list(TEList, M, Program);
ensure_texp({none}, _, Program) ->
    Program;
ensure_texp({any}, _, Program) ->
    Program;
ensure_texp({number, _, TE}, M, Program) ->
    ensure_texp(TE, M, Program);
ensure_texp({union, _, Exp1, Exp2}, M, Program) ->
    Program1 = ensure_texp(Exp1, M, Program),
    ensure_texp(Exp2, M, Program1);
ensure_texp({type_call, _, Tyid, _, ExpList}, M, Program) ->
    Program1 = ensure_texp_list(ExpList, M, Program),
    ensure_typedef(Tyid, Program1);
%% ensure_texp({type_extern_call, _, Tyid, ExpList}, M, Program) ->
%%     Program1 = ensure_texp_list(ExpList, M, Program),
%%     ensure_typedef(Tyid, Program1);
ensure_texp({watch, _, Exp}, M, Program) ->
    ensure_texp(Exp, M, Program).

ensure_texp_list([], _M, Program) ->
    Program;
ensure_texp_list([E|List], M, Program) ->
    Program1=ensure_texp(E, M, Program),
    ensure_texp_list(List, M, Program1).   




