%%% @author Sven-Olof Nystrom <svenolof@hyena>
%%% @copyright (C) 2015, Sven-Olof Nystrom
%%% Created : 20 Aug 2015 by Sven-Olof Nystrom <svenolof@hyena>

-module(solver).

% -compile(export_all).
-include_lib("stdlib/include/assert.hrl").

-export([solve/1, solve_0/1, scan_watches/2]).
-export([all_vars/1, is_present_var/2]).
-export([cexp_c/1, lookup_entry/2, get_entry/2]).
-export([entry_supports/1, entry_covers/1]).
-export([entry_uplinks/1, entry_downlinks/1]).
-export([collect_supports/2]).
-export([source_comb/2, source_empty/0]).

%: +type con() = conn:con().
%: +type constraint() = walker:constraint().
%: +type dl(X) = erlang:deeplist(X).
%: +type work() = 
%:     {fail, source()}
%:   + {include, walker:texp(), walker:texp0(), source()}.

%: 
%: +type name() =
%:     walker:name().
%: 
%: +type graph() =
%:     {list({name(), io_lib:iolist()}),
%:      map(name(), entry())}.
%: +type bool() = erlang:bool().
%: +type list(X) = erlang:list(X).
%: 
%: +type entry() =
%:     {entry,
%:      map(name(), {rfilter:filter(), source()}),  % downlinks    
%:      map(name(), {rfilter:filter(), source()}),  % uplinks    
%:      list({walker:texp0(), source()}),                     % supports
%:      list({walker:texp0(), source()})}.                    % covers
%:
%: +type source() = erlang:deep_entry({atom(), any()} + {atom(), any(), any()}). 
%:
%: +type string() = erlang:string().
%: +func solve :: dl(constraint()) -> !poly:result(!graph()).

%: +type deeplist(X) = erlang:deeplist(X).

%: +func solve_0 :: dl(constraint()) -> !poly:result(!graph()).

%: +func make_theory :: () -> !graph().

make_theory() ->
    {[], maps:new()}.

%: +func get_entry :: name() * graph() -> !entry().
get_entry(X, {_, G}) ->
    maps:get(X, G).

%: +func lookup_entry :: name() * graph() -> {value, !entry()} + !none.
lookup_entry(X, {_, G}) ->
    case maps:find(X, G) of
        {ok, Entry} ->
            {value, Entry};
        error ->
            none
    end.

update_entry(X, Entry, {W, G}) ->
    {W, maps:update(X, Entry, G)}.

%: +func is_present_var :: name() * graph() -> bool().
is_present_var(X, {_, G}) ->
    maps:is_key(X, G).

ensure_var(X, G={W, M}) ->
    case is_name(X) of
	false -> io:format("~nX=~w", [X]);
	true  -> true
    end,
    ?assert(is_name(X), "Variable expected"),
    case maps:is_key(X, M) of
	true ->
	    G;
	false ->
	    {W, maps:put(X, empty_entry(), M)}
    end.

%%% +func add_watch :: name() * s

add_watch(X, Watch, {W, M}) ->
    {[{X, Watch} | W], M}.

%%%%

make_wl(G) ->    
    WL = worklist:empty(),
    make_wl(WL, G).

make_wl(WL, []) ->
    WL;
make_wl(WL, [G1|G2]) ->
    WL1 = make_wl(WL, G1),
    make_wl(WL1, G2);
make_wl(WL, G) ->
    add_work(G, WL).

solve(G) ->
   Theory = make_theory(),
   WL = make_wl(G),
   case loop(WL, Theory) of
       {fail, S} ->
           {fail, S};
       {success, Theory1} ->
           case scan_watches(my_sets:from_list([]), Theory1) of
                {success, _} ->
                    {success, Theory1};
                {fail, Source} ->
                   {fail, Source} 
           end
    end.

solve_0(G) ->
   Theory = make_theory(),
   WL = make_wl(G),
   loop(WL, Theory).

loop(WL, G) ->
    case worklist:get_work(WL) of
        none ->
            {success, G};
        {value, {{fail, S}, _WL1}} ->
            {fail, S};
        {value, {Work, WL1}} ->
            {WorkList, G1} = add_constraint(Work, G),
            WL2 = add_work_list(WorkList, WL1),
            loop(WL2, G1)
    end.

add_work(Work = {include, E1, E2, _Source}, WL) ->
   Key = {E1, E2},
    worklist:insert(Key, Work, WL);
add_work(Work = {watch, _, _}, WL) ->
    worklist:insert(Work, Work, WL);
add_work(Work = {fail, _}, WL) ->
    worklist:insert(Work, Work, WL).


add_work_list([], WL) ->
    WL;
add_work_list([Work|WorkList], WL) ->
    WL1 = add_work(Work, WL),
    add_work_list(WorkList, WL1).

%: +func scan_watches :: my_sets:set(name()) * graph() -> 
%:                       !poly:result(!list(!{name(), !io_lib:iolist()})).

scan_watches(Left, Graph = {WL, _}) ->
    scan_watches(WL, Left, Graph, []).

scan_watches([], _Left, _Graph, Acc) ->
    {success, Acc};
scan_watches([W = {Name, M}|WL], Left, Graph, NL) ->
    case triggered(Name, Graph) of
        true ->
            scan_watches(WL, Left, Graph, NL);
        false ->
            case my_sets:is_member(Name, Left) of
                true ->
                    scan_watches(WL, Left, Graph, [W|NL]);
                false ->
                    io:format("~nUntriggered: ~w", [Name]),
                    io:format("~n~s ", [M]),
                    {fail, []}
            end
   end.

triggered(Name, Graph) ->
   is_present_var(Name, Graph) andalso
        collect_supports(Name, Graph, rfilter:any(), source_empty()) /= [].


%: +func all_vars :: graph() -> !list(!name()).                       
all_vars({_, Map}) ->
    maps:keys(Map).

%%%

ensure_var_list([], G) ->
    G;
ensure_var_list([X|L], G) ->
    G1 = ensure_var(X, G),
    ensure_var_list(L, G1).

empty_entry() ->
    {entry, 
     maps:new(), % Downlinks
     maps:new(), % Uplinks
     ordsets:new(), % Supports
     ordsets:new()  % Covers
    }.

%: +func entry_downlinks 
%:       :: entry()   
%:       -> !map(name(), !{!rfilter:filter(), !source()}).
entry_downlinks({entry, Downlinks, _Uplinks, _Supports, _Covers}) ->
    Downlinks.

%: +func entry_uplinks 
%:       :: entry()   
%:       -> !map(!name(), !{!rfilter:filter(), !source()}).
entry_uplinks({entry, _Downlinks, Uplinks, _Supports, _Covers}) ->
    Uplinks.

%: +func entry_supports :: entry() -> !list({!walker:texp0(), !source()}).
entry_supports({entry, _Downlinks, _Uplinks, Supports, _Covers}) ->
    Supports.

%: +func entry_covers :: entry() -> !list({!walker:texp0(), !source()}).
entry_covers({entry, _Downlinks, _Uplinks, _Supports, Covers}) ->
    Covers.

entry_update_downlinks({entry, _Downlinks, Uplinks, Supports, Covers},
		     Downlinks) ->
    {entry, Downlinks, Uplinks, Supports, Covers}.


entry_update_uplinks({entry, Downlinks, _Uplinks, Supports, Covers},
		     Uplinks) ->
    {entry, Downlinks, Uplinks, Supports, Covers}.

entry_update_supports({entry, Downlinks, Uplinks, _Supports, Covers},
		     Supports) ->
    {entry, Downlinks, Uplinks, Supports, Covers}.

entry_update_covers({entry, Downlinks, Uplinks, Supports, _Covers},
		     Covers) ->
%%     if length(Covers) >= 20 -> io:format("~nCovers: ~w", [ length(Covers)]), 
%% io:format("~p", [Covers]);
%%        true -> true
%%     end,
    {entry, Downlinks, Uplinks, Supports, Covers}.



is_zero_name(zero) ->
    true;
is_zero_name({var, zero}) ->
    error("unwanted zero");
is_zero_name(_) ->
    false.

%: +func add_constraint :: constraint() * !graph() -> 
%:                         {!list(!work()), !graph()}.
                             
add_constraint({include, {var, zero}, _, _}, G) ->
    {[], G};
add_constraint({include, E={c, C, _}, R={c, C1, _}, S}, G) ->
    case conn:con_included(C, C1) of
        true ->
            {match(E, R, S), G};
	false ->			 
            {[{fail, S}], G}
    end;
add_constraint({include, {var, X}, R={c, C, _}, S}, G) ->
    F = rfilter:singleton_inverse(C),
    {WorkList1, G1} = add_exclude(X, F, G, [{exclude,C, F} , S]),
    {WorkList2, G2} = add_cover(X, R, G1, S),
    {WorkList1++WorkList2, G2};
add_constraint({include, {sift, X, F}, R={c, C, _}, S}, G) ->
    Fe = rfilter:minus(F, 
		       rfilter:singleton(C)),
    {WorkList1, G1} = 
        add_exclude(X, Fe, G, [{exclude,C, Fe}, S]),
    {WorkList2, G2} = 
	add_cover(X, R, G1, S),
    {WorkList1++WorkList2, G2};
add_constraint({include, {sift, X, F}, {var, Y}, S}, G) ->
    add_link(X, F, Y, G, S);
add_constraint({include, {var, X}, {var, Y}, S}, G) ->
    add_link(X, rfilter:any(), Y, G, S);
add_constraint({include, E={c, _, _}, {var, X}, S}, G) ->
    add_support(E, X, G, S);
add_constraint({watch, Name, M}, G) ->
    {[], add_watch(Name, M, G)}.

%: +func add_support :: walker:texp0() * name() * graph() * source() 
%:                   -> {!list(!work()), !graph()}.

add_support(Support, X, G, Source) ->
    SS = {Support, Source},
    G1 = ensure_var(X, G),
    XEntry = get_entry(X, G1),
    Supports = entry_supports(XEntry),
    case ordsets:is_element(SS, Supports) of
	true ->
	    {[], G1};
	false ->
	    Supports1 = ordsets:add_element(SS, Supports),
	    XEntry1 = entry_update_supports(XEntry, Supports1),
	    G2 = update_entry(X, XEntry1, G1),
	    push_support(X, Support, G2, Source)
    end.

%: +func add_cover :: name() * walker:texp0() * graph() * source() -> 
%:                        !{!list(!work()), !graph()}.

add_cover(X, Cover, G, Source) ->
    CS = {Cover, Source},
    G1 = ensure_var(X, G),
    XEntry = get_entry(X, G1),
    Covers = entry_covers(XEntry),
    case ordsets:is_element(CS, Covers) of
	true ->
            {[], G1};
	false ->
	    Covers1 = ordsets:add_element(CS, Covers),
	    XEntry1 = entry_update_covers(XEntry, Covers1),
	    G2 = update_entry(X, XEntry1, G1),

	    Supports = collect_supports(X, G2, 
					rfilter:singleton(cexp_c(Cover)),
					Source),
	    {match_list_list(Supports, [CS]), G2}
    end.


add_exclude(X, F, G, Source) ->
    add_link(X, F, zero, G, Source).

%: +func add_link :: name() * rfilter:filter() * name() 
%:                   * graph() * source() -> 
%:                        !{!list(!work()), !graph()}.

add_link(X, F, Y, G, Source) ->
    FS = {F, Source},
    case rfilter:is_empty(F) 
    orelse is_present_link(X, F, Y, G) 
    orelse X==Y of
        true -> 

	    {[], G};
	false -> 
	    G1 = ensure_var_list([X, Y], G),
  	    XEntry = get_entry(X, G1),
            XUplinks = entry_uplinks(XEntry),
	    case maps:find(Y, XUplinks) of
	        error ->
		    XUplinks1 = maps:put(Y, FS, XUplinks);
		{ok, {F1, S1}} ->
                    FS1 = {rfilter:union(F, F1), source_comb(Source, S1)},
		    XUplinks1 = maps:update(Y, FS1, XUplinks)
            end,
	    XEntry1 = entry_update_uplinks(XEntry, XUplinks1),

	    G2 = update_entry(X, XEntry1, G1), 

  	    YEntry = get_entry(Y, G2),
            YDownlinks = entry_downlinks(YEntry),
	    case maps:find(X, YDownlinks) of
	        error ->
		    YDownlinks1 = maps:put(X, FS, YDownlinks);
		{ok, {F2, S2}} ->
		    io:format("~nRfilter: ~p ~p", [F, F2]),
                    FS2 = {rfilter:union(F, F2), source_comb(Source, S2)},
		    YDownlinks1 = 
			maps:update(X, FS2, YDownlinks)
            end,
	    YEntry1 = entry_update_downlinks(YEntry, YDownlinks1),	
	    G3 = update_entry(Y, YEntry1, G2),

	    XSupports  = collect_supports(X, G3, F, Source),
            XSupports1 = 
                [{XSupport, source_comb(Source, Source1)}
                 || 
                    {XSupport, Source1} <- XSupports,
                    rfilter:overlap(rfilter:singleton(cexp_c(XSupport)), F)],
%                    rfilter:contains(cexp_c(XSupport), F)],
            {WL, G4} = push_support_list(XSupports1, Y, G3),
            {WL, G4}
    end.

is_present_link(X, F1, Y, G) ->
    case is_present_var(X, G) of
	true ->
	    XEntry = get_entry(X, G),
	    XUplinks = entry_uplinks(XEntry),
	    case maps:find(Y, XUplinks) of
		error ->
		    false;
		{ok, {F2, _}} ->
		    rfilter:included(F1, F2)
	    end;
	false ->
	    false
    end.

%: +func match_list_list :: list({walker:texp0(), source()}) 
%:                          * list({walker:texp0(), source()}) -> 
%:                          list(work()).

match_list_list(SList, RList) -> 
    match_list_list(SList, RList, []).
    

match_list_list(_, [], Acc) ->
    Acc;
match_list_list(SList, [R|RList], Acc) ->
    Acc1 = match_list(SList, R, Acc),
    match_list_list(SList, RList, Acc1).

match_list([], _, Acc) ->
    Acc;
match_list([S|SList], R, Acc) ->
    Acc1 = match_1(S, R, Acc),
    match_list(SList, R, Acc1).

%% %: +func match :: {walker:texp0(), source()} * {walker:texp0(), source()} -> 
%% %:                !list(!work()).
%% match(SList, RList) ->
%%     match_1(SList, RList, []).

match_1({Left, S1}, {Right, S2}, Acc) ->
    case conn:match(Left, Right) of
        {value, L} ->
            S = source_comb(S1, S2),
            [{include, A, B, S} || {include, A, B} <- L ]++Acc;
        none ->        
            Acc
    end.
%: +func match :: walker:texp0() * walker:texp0() * source() -> 
%:                !list(!work()).

match(SList, RList, S) ->
    match_1(SList, RList, S, []).

match_1(Left, Right, S, Acc) ->
    case conn:match(Left, Right) of
        {value, L} ->
            [{include, A, B, S} || {include, A, B} <- L ]++Acc;
        none ->
            Acc
    end.

% push support

%: +func push_support_list :: list({walker:texp0(), source()}) * name() * graph() 
%:                         ->  !{!list(!work()), !graph()}.

push_support_list([], _, G) ->
    {[], G};
push_support_list([{E, Source}|L], X, G) ->
    {WorkList1, G1} = push_support(X, E, G, Source),
    {WorkList2, G2} = push_support_list(L, X, G1),
    {WorkList1++WorkList2, G2}.

%: +func push_support :: name() * walker:texp0() * graph() * source() -> 
%:                       !{!list(!work()), !graph()}.

push_support(X, E, G, Source) ->
    Visited = maps:new(),
    Filter = rfilter:any(),
    {WorkList, _, G1} = push_support(X, E, Filter, G, Visited, Source),
    {WorkList, G1}.

push_support(X, E, Filter, G, Visited, Source) ->
    case is_visited(X, Filter, Visited) of
	true ->
	    {[], Visited, G};
	false ->
	    case is_zero_name(X) of 
		false ->
		    G0 = ensure_var(X, G),
		    Visited1 = add_visited(X, Filter, Visited),
		    XEntry = get_entry(X,  G0),
		    XCovers = [{Cover, FS}
			       || {Cover, FS} <- entry_covers(XEntry),
				  rfilter:included(
				    rfilter:singleton(cexp_c(Cover)),
                                     Filter)],
%                                  rfilter:contains(cexp_c(Cover),
%                                    Filter)],
                    WorkList1  =
                        match_list_list([{E, Source}], XCovers),
                    
                    XUplinks = 
                        maps:to_list(entry_uplinks(XEntry)),

                    {WorkList2, Visited2, G2} =
                        push_support_uplinks(XUplinks, E, Filter,
                                             G, Visited1, Source),
                    {WorkList1++WorkList2, Visited2, G2};
                true ->
                    {[{fail, [{adding, E} | Source]}],  Visited, G}
	    end
    end.

push_support_uplinks([], _, _, G, Visited, _Source) ->
    {[], Visited, G};
push_support_uplinks([{X, {Filter0, Source1}}|Uplinks], E, F,
                     G, Visited, Source2) ->
    Source = source_comb(Source1, Source2),
    Filter = rfilter:intersect(F, Filter0),
    case rfilter:overlap(rfilter:singleton(cexp_c(E)), Filter) of
        false ->
	    push_support_uplinks(Uplinks, E, F, G, Visited, Source2);
        true ->
            {WorkList1, Visited1, G1} = 
                push_support(X, E, Filter, G, Visited, Source),
            {WorkList2, Visited2, G2} =
                push_support_uplinks(Uplinks, E, F, G1, Visited1, Source2),
            {WorkList1++WorkList2, Visited2, G2}
    end.

is_visited(X, F, Visited) ->
    case maps:find(X, Visited) of
        error ->
	    false;
	{ok, F1} ->
            rfilter:included(F, F1)
    end.

add_visited(X, F, Visited) ->
    case maps:find(X, Visited) of
        error ->
	    maps:put(X, F, Visited);
	{ok, F1} ->
            maps:update(X, rfilter:union( F1, F), Visited)
    end.


collect_supports(Name, Graph) ->
    case is_present_var(Name, Graph) of
	true ->
	    collect_supports(Name, Graph, rfilter:any(), source_empty());
	false ->
	    []
    end.

%: +func collect_supports :: name() * graph() * rfilter:filter() * source() 
%:                        -> !list(!{!walker:texp0(), !source()}).

collect_supports(X, G, Filter, Source) ->
    Visited = my_sets:empty(),
    {_, Acc} = collect_supports(X, G, Filter, Visited, [], Source),
    Acc.

collect_supports(X, G, F, Visited, Acc, Source) ->
    case rfilter:is_empty(F) orelse is_visited(X, Visited) of
	true ->
	    {Visited, Acc};
	false ->
	    Visited1 = add_visited(X, Visited),
	    XEntry = get_entry(X,  G),

	    XSupports = 
                [{Support, source_comb(Source, Source1)} || 
                    {Support, Source1} <- entry_supports(XEntry),
                    rfilter:overlap(                        
		       rfilter:singleton(cexp_c(Support)), F)],
%                    rfilter:contains(cexp_c(Support), F)],
	    Acc1 = XSupports ++ Acc, 
	    XDownlinks = maps:to_list(entry_downlinks(XEntry)),
	    collect_supports_downlinks(XDownlinks, G, F, Visited1, Acc1, Source)
    end.

collect_supports_downlinks([], _, _, Visited, Acc, _) ->
    {Visited, Acc};
collect_supports_downlinks([{X, {Filter, Source1}}|Downlinks], 
                           G, F, Visited, Acc, Source) ->
    Source2 = source_comb(Source, Source1),
    F2 = rfilter:intersect(Filter,F),
    {Visited1, Acc1} =				
        collect_supports(X, G, F2, Visited, Acc, Source2),
    collect_supports_downlinks(Downlinks, G, F, Visited1, Acc1, Source).


is_visited(X, Visited) ->  
    my_sets:is_member(X, Visited).

add_visited(X, Visited) ->
    my_sets:add(X, Visited).


is_name({var, _}) ->
    false;
is_name(N) when is_integer(N)->
    true;
is_name({stable, A, B}) when is_atom(A), is_integer(B) ->
    true;
is_name(zero) -> true.
%% is_name(S) ->
%%     io:format("~nsolver: name ~p", [S]),
%%     true.

%: +func cexp_c :: walker:texp0() -> !con().  
cexp_c({c, C, _}) ->
    C.

%%

source_empty() ->
    [].

source_comb(S1, S2) ->
    [S1,S2].
