%% @author Sven-Olof Nystrom <svenolof@asgam>
%%% @copyright (C) 2015, Sven-Olof Nystrom
%%% @doc
%%%
%%% @end
%%% Created :  8 Sep 2015 by Sven-Olof Nystrom <svenolof@asgam>

% Generate constraints from Erlang code.

%: +type exp() = convert:exp().
%: +type pattern() = convert:pattern().
%: +type list(X) = erlang:list(X).

%: +type tyid() = convert:tyid().
%: +type fnid() = convert:fnid().

%: +type con() = conn:con().
%: +type name() = atom()+integer()+{stable, pname(), atom()+integer()}.
%: +type line() = program:line().

%: +type pos() = pos:pos().
%: +type env() = {env, integer(), atom(), gb_trees:tree(name(), texp0())}.
%: +type tenv() = {tenv, atom(), gb_trees:tree(name(), texp0())}.
%: +type bind() = gb_trees:tree(name(), texp0()).

%: +type pname() = atom().
%: +type constraint() = 
%:     {call, pname(), list(name())} 
%:   + {include, texp(), texp0(), solver:source()}
%:   + {watch, name(), io_lib:iolist()}.
%: +type constraints() =  dl(constraint()).
%: +type predicate() = {predicate, pname(), list(name()), dl(constraint())}.
%: +type dl(X) = erlang:deeplist(X).
%: +type dir() = lower + upper.
%: +type texp() = 
%:     {var, name()}
%:   + {c, con(), list(texp0())}
%:   + {sift, name(), rfilter:filter()}.
%: +type texp0() = 
%:     {var, name()}
%:   + {c, con(), list(texp0())}.
%:
%: +type theory() = gb_trees:tree(pname(), predicate()).

-module(walker).
%-compile(export_all).

-export([agenda/2, top/2, def/2, spec/3, typedef/3]).
-export([reverse/1, source_empty/0, apply_filter/2]).
-export([bind_empty/0, bind_singleton/2, bind_combine/2]).
-export([fnid_to_predicate_name/2, name_to_string/1]).


%: +func agenda :: agenda:agenda() 
%:               * program:program() 
%:               -> theory().
agenda(A, P) ->
    L = agenda:to_list(A),
    Items = agenda_items(L, theory_new(), P),
    Items.

agenda_items([A | L], T, P) ->
    PredicateList = item(A, P),
    T1 = theory_insert_list(PredicateList, T),
    agenda_items(L, T1, P);
agenda_items([], T, _) ->
    T.

%: +func item :: agenda:item() 
%:             * program:program() 
%:             -> !list(predicate()).
item({top, F}, P) ->
    [top(F, P)];
item({def, F}, P) ->
    [def(F, P)];
item({spec, F, Dir}, P) ->
    [spec(F, Dir, P)];
item({typedef, Tyid, Dir}, P) ->
    typedef(Tyid, Dir, P);
item({callee, Fnid}, P) ->
    [callee(Fnid, P)].

%: +func top :: fnid() * program:program() -> !predicate().
top(F, _P) ->
    Pos = pos:pos_zero(),
    Name = name_new(Pos),
    TName =  fnid_to_predicate_name(F, top),
    {predicate, TName, [],  
     [{call, fnid_to_predicate_name(F, def), [Name]},
      {call, fnid_to_predicate_name(F, {spec, upper}), [Name]}]}.


%: +func def :: fnid() * program:program() -> !predicate().
def(Fnid = {fnid, Module, _, Arity}, P) -> 
    {def, Line, _, ArgList, E} = program:def(Fnid, P),
    Source = source(Module, Line, {function_definition, Fnid}),
    ArgList1 = [{var, Name} || {var, Name} <- ArgList], %to keep scfa happy
    Pos = pos:pos_zero(),
    Name = name_new(pos:left(Pos)),
    B = bind_insert_arg_list(ArgList, bind_empty()),
    Env = env_add_binding(env_new(Module), B),
    {Res, G1, _} = exp(E, Env, pos:right(Pos)),
    G2 = {include ,
    	  {c, conn:fun_fixed(Arity), [Res]++ArgList1}, {var, Name}, Source},
    PName = fnid_to_predicate_name(Fnid, def),
     {predicate, PName, [Name], [G1, G2]}.		

%: +func spec :: fnid() * dir() * program:program() -> !predicate().
spec(Fnid = {fnid, Module, _, _}, Dir, Program) ->
    {spec, Line, _, TE} = program:spec(Fnid, Program),
    Source = source(Module, Line, {function_spec, Fnid}),
    FV = free_variables(TE), 
    Pos = pos:pos_zero(),	   
    [Pos1, Pos2] = pos:pos_list(2, Pos),
    Name = name_new(Pos1),
    AI = alias_introduced(TE),

    case Dir of
        upper ->
    	    TEnv = tenv_new(AI,FV, Module),
            {Type, G1} = type_exp(TE, Dir, TEnv, Pos2),
            G2 = create_include({var, Name}, Type, rfilter:any(), Source);
        lower ->
	    TEnv = tenv_new(AI++FV, [], Module),
            {Type, G1} = type_exp(TE, Dir, TEnv, Pos2),
            G2 = create_include(Type, {var, Name}, rfilter:any(), Source)
	 end,
    PName = fnid_to_predicate_name(Fnid, {spec, Dir}),
    {predicate, PName, [Name], [G1, G2]}.


%: +func typedef :: tyid() * dir() * program:program() ->
%:                  !list(predicate()).
typedef(Tyid = {tyid, Module, _}, Dir, Program) ->
    {typedef, Line, _, ArgList, TE} = program:typedef(Tyid, Program),
    Source = source(Module, Line, {type_definition, Tyid}),
    ArgList1 = [{var, Name} || {var, Name} <- ArgList], %to keep scfa happy
 
    FV = free_variables(TE),
    Pos = pos:pos_zero(),
    [Pos1, Pos2] = pos:pos_list(2, Pos),
    N = length(ArgList),
    assert(length(FV)=<N, "Some type variable not present in the parameter list"),
    Name = name_new(Pos1),
    UnboundList = [Name1||{var, Name1} <- ArgList],
    Env = tenv_new(UnboundList, [], Module),
    {ResType, G1} = type_exp(TE, Dir, Env, Pos2),
    Type = {c, conn:fun_fixed(N), [ResType|ArgList1]},                    
    G2 = case Dir of
    	     upper ->
    		 create_include({var, Name}, Type, rfilter:any(), Source);
    	     lower ->
    		 create_include(Type, {var, Name}, rfilter:any(), Source)
    	 end,
    TNameLocal = tyid_to_predicate_name(Tyid, Dir, local),
    TNameExtern = tyid_to_predicate_name(Tyid, Dir, extern),
    [{predicate, TNameLocal, [Name], [G1, G2]},
     {predicate, TNameExtern, [Name], 
      [{call, TNameLocal, [Name]}]}].

%: +func callee :: fnid() * program:program() -> !predicate().
callee(Fnid, Program) ->
    case program:def_or_spec(Fnid, Program) of
	def ->
	    Callee = fnid_to_predicate_name(Fnid, def);
	spec ->
	    Callee = fnid_to_predicate_name(Fnid, {spec, lower});
	none ->
	    Callee = stupid,
	    error(lists:flatten(
		    io_lib:format("no def or spec for ~w in program ~w", 
				  [Fnid, program:keys(Program)])))
    end,
    Pos = pos:pos_zero(),
    Name = name_new(pos:left(Pos)),
    Name1 = name_new(pos:right(Pos)),
    CName = fnid_to_predicate_name(Fnid, callee),
    {predicate, CName, [Name], 
     [{include, {var, Name1}, {var, Name}, source_empty()},
      {call, Callee, [Name1]}]}.

% New body and block to handle checks.

body([{check, E}|EList], Env, Pos) ->
    {_, G1, B1} = exp(E, Env, Pos),
    {T, G2, B2} = body(EList, env_add_binding(Env, B1), pos:right(Pos)), 
    {T, [G1, G2], bind_combine(B1, B2)};
body([E|EList], Env, Pos) ->
    {T0, G1, B1} = exp(E, Env, pos:left(Pos)),
    {T, G2, B2} = body_1(EList, T0, env_add_binding(Env, B1), pos:right(Pos)),
    {T, [G1, G2], bind_combine(B1, B2)}.

body_1([{check, E}|EList], T0, Env, Pos) ->
    {_, G1, B1} = exp(E, Env, Pos),
    {T, G2, B2} = body_1(EList, T0, env_add_binding(Env, B1), pos:right(Pos)), 
    {T, [G1, G2], bind_combine(B1, B2)};
body_1([E|EList], _, Env, Pos) ->
    {T0, G1, B1} = exp(E, Env, pos:left(Pos)),
    {T, G2, B2} = body_1(EList, T0, env_add_binding(Env, B1), pos:right(Pos)),
    {T, [G1, G2], bind_combine(B1, B2)};
body_1([], T0, _, _) ->
    {T0, [], bind_empty()}.


block([], _, _) ->
    {[], bind_empty()};
block([E|EList], Env, Pos) ->
    {_, G1, B1} = exp(E, Env, pos:left(Pos)), 
    {G2, B2} = block(EList, Env, pos:right(Pos)),
    {[G1, G2], bind_combine(B1, B2)}.

%: +func exp :: convert:exp() * env() * pos() 
%:           -> !{!texp0(), !constraints(), !bind()}.
exp({var, Name}, Env, _Pos) ->
    E = env_get(Name, Env),
    {E, [], bind_empty()};
exp({c, C, ExpList}, Env, Pos) ->
    case C of
	{integer_fixed, _} ->
	    {integer_u(Pos), [], bind_empty()};
	{float_fixed, _} ->
	    {float_l(Pos), [], bind_empty()};
	C1 ->
	    {TList, G, B} = exp_list(ExpList, Env, Pos),
	    {{c, C1, TList}, G, B}
    end;
exp({match, _Line, Pattern, Exp}, Env, Pos) ->
    [Pos1, Pos2] = pos:pos_list(2, Pos),
    {T, G1, B1} = exp(Exp, Env, Pos1),
    {G2, B2} = 
        match:match(Pattern, T, Pos2), 
    {T, [G1, G2], bind_combine(B1, B2)};
exp({lambda, VarList, E}, Env, Pos) ->
    TempList = temp_new(VarList, pos:left(Pos)),          
    B = bind_insert_list(lists:zip(VarList, TempList), bind_empty()),
    Env1 = env_add_binding(Env, B),
    {T, G, _} = exp(E, Env1, pos:right(Pos)),
    {{c, conn:fun_fixed(length(VarList)), [T|TempList]}, G, bind_empty()};
exp({lfun, _, Fnid={fnid, _, _, _}}, _Env, Pos) ->
    Name = name_new(Pos),
    CName = fnid_to_predicate_name(Fnid, callee),
    G = {call, CName, [Name]},
    Res = {var, Name},
    {Res, [G], bind_empty()};
exp(E={ccase, _, _, _, _}, Env, Pos) ->
     ccase_exp(E, Env, Pos);
exp(E={'receive', _, _, _}, Env, Pos) ->
     receive_exp(E, Env, Pos);
exp(E={'receive', _, _, _, _, _}, Env, Pos) ->
     receive_exp(E, Env, Pos);
exp({call, Line, Fnid, ExpList}, Env, Pos) ->
    Source = source(env_module(Env), Line, {call, Fnid}),
    Arity = length(ExpList),
    Name = name_new(pos:left(pos:left(Pos))),
    Res = {var, name_new(pos:right(pos:left(Pos)))},
    {TList, G1, B} = exp_list(ExpList, Env, pos:right(Pos)),
    CName = fnid_to_predicate_name(Fnid, callee),
    G2 = {call, CName, [Name]},
    G3 = {include, {var, Name}, {c, conn:fun_fixed(Arity), [Res|TList]}, Source},
    {Res, [G1, G2, G3], B}; 
exp({local_call, Line, Fnid, ExpList}, Env, Pos) ->
    Source = source(env_module(Env), Line, {local_call, Fnid}),
    Arity = length(ExpList),
    Name = name_new(pos:left(pos:left(Pos))),
    Res = {var, name_new(pos:right(pos:left(Pos)))},
    {TList, G1, B} = exp_list(ExpList, Env, pos:right(Pos)),
    CName = fnid_to_predicate_name(Fnid, callee),
    G2 = {call, CName, [Name]},
    G3 = {include, {var, Name}, {c, conn:fun_fixed(Arity), [Res|TList]}, Source},
    {Res, [G1, G2, G3], B}; 
exp({apply, Line, Exp, ExpList}, Env, Pos) ->
    Source = source(env_module(Env), Line, apply),
    Arity = length(ExpList),
    Res = {var, name_new(pos:right(pos:left(Pos)))},
    {[T|TList], G1, B} = exp_list([Exp|ExpList], Env, pos:right(Pos)),
    G3 = {include, T, {c, conn:fun_fixed(Arity), [Res|TList]}, Source},
    {Res, [G1, G3], B};
exp({lc, Line, Exp, GenList}, Env, Pos) ->
    Source = source(env_module(Env), Line, list_comprehension),
    {G1, B1} = gen_list(GenList, Env, pos:left(Pos)),
    Env1 = env_add_binding(Env, B1),
    {T, G2, _B2} = exp(Exp, Env1, pos:left(pos:right(Pos))),
    Res = {var, name_new(pos:right(pos:right(Pos)))},
    G3 = [{include, {c, conn:cons(), [T, Res]}, Res, Source}, 
          {include, {c, conn:nil(), []}, Res, Source}],
    {Res, [G1, G2, G3], bind_empty()};
exp({promise, _, Exp, TExp}, Env, Pos) ->
    Module = env_module(Env),
    assert(free_variables(TExp) == []),
    [Pos1, Pos2, Pos3] = pos:pos_list(3, Pos),
    {_, G1, B1} = exp(Exp, Env, Pos1),
    {T, G2} = type_exp(TExp, lower, tenv_new(Module), Pos2),
    {T1, G3} = unextend(T, lower, Pos3), 
    B2 = bind_replace(B1, T),
    {T1, [G1, G2, G3], B2};
exp({block, _, Block}, Env, Pos) ->
    body(Block, Env, Pos);
exp({'try', Line, Block, CRClauseList, ClauseList, After}, Env, Pos) ->
    [Pos1, Pos2, Pos3, Pos4, Pos5, Pos6, Pos7] = pos:pos_list(7, Pos),
    Source = source(env_module(Env), Line, 'try'),
    {T, G1, B} = body(Block, Env, Pos1),
    Res1 = {var, name_new(Pos2)},
    {G2, _} = clause_list([T], [], CRClauseList, [], Res1, Env, Pos3),
    Res2 = {var, name_new(Pos4)},
    Any = {c, conn:any(), []},
    {G3, _} = clause_list([Any], [], ClauseList, [], Res2, Env, Pos5),
    {G4, _} = block(After, Env, Pos6),	
    Res = {var, name_new(Pos7)},
    G5 = [{include, Res1, Res, Source},
	  {include, Res2, Res, Source}],
        {Res, [G1, G2, G3, G4, G5], B};
exp({'catch', _, E}, Env, Pos) ->
    {_, G, B} = exp(E, Env, Pos),
    {{c, conn:any(), []}, G, B};
    
exp({check, E}, Env, Pos) ->
    {_, G, B} = exp(E, Env, Pos),
    {{var, zero}, G, B}.

%: +func gen_list :: list(convert:lc_gen()) * env() * pos() 
%:                -> !{constraints(), !bind()}.		       
gen_list([], _, _) ->
    {[], bind_empty()};
gen_list([Gen|GenList], Env, Pos) ->
    {G1, B1} = gen(Gen, Env, pos:left(Pos)),
    {G2, B2} = gen_list(GenList, env_add_binding(Env, B1), pos:right(Pos)),
    {[G1, G2], bind_combine(B1, B2)}.

%: +func gen :: convert:lc_gen() * env() * pos() 
%:           -> !{!dl(constraint()), !bind()}.		       
gen({generate, Line, P, Exp}, Env, Pos) ->
    Source = source(env_module(Env), Line, list_comprehension),
    [Pos1, Pos2, Pos3] = pos:pos_list(3, Pos),
    {T, G1, B1} = exp(Exp, Env, Pos1),
    Name = name_new(Pos2), 
    U = {var, name_new(Pos3)},
    G2 = [{include, T, {var, Name}, Source},
          {include, {sift, Name, rfilter:singleton(conn:cons())}, 
	            {c, conn:cons(), [U, {var, Name}]},
                    Source}],
    {G3, B3} = 
        match:match(P, U, pos:right(pos:right(Pos))), 
        
    {[G1, G2, G3], bind_combine(B1, B3)};
gen({b_generate, _Line, _P, _Exp}, _Env, _Pos) ->
    error("not implemented yet");  % necessary to keep scfa happy
gen(Exp, Env, Pos) ->
    {_T, G, _B} = exp(Exp, Env, Pos),
    {G, bind_empty()}.


%: +func ccase_exp :: convert:exp() * env() * pos() 
%:                -> !{!texp0(), !dl(constraint()), !bind()}.
ccase_exp({ccase, _, KeyList, ClauseList, ExportList}, Env, Pos) ->
    {TList, G1, B1} = exp_list(KeyList, Env, pos:left(pos:left(Pos))),
    Res = {var, name_new(pos:right(pos:left(Pos)))},
    {G2, B2} = 
	clause_list(TList, [], ClauseList, ExportList, Res, Env, pos:right(Pos)),
    {Res, [G1, G2], bind_combine(B1, B2)}.

%: +func clause_list :: list(texp0())
%:                    * list(list(pattern()))
%:                    * list(convert:cclause()) 
%:                    * list(name())
%:                    * texp0()
%:                    * env()
%:                    * pos()
%:                   -> !{!constraints(), !bind()}.
clause_list(KeyList, QList, [Clause|ClauseList], ExportList, Res, Env, Pos) ->
    {G1, B1} = clause(KeyList, QList, Clause, Res, Env, pos:left(Pos)),
    case check_simple_clause(Clause) of
        {value, HeadList} ->
	    QList1 = HeadList++QList;
	none ->
	    QList1 = QList
    end,
    {G2, B2} = clause_list(KeyList, QList1, ClauseList, 
			   ExportList, Res, Env, pos:right(Pos)),
    B11 = bind_exports(B1, ExportList),
    G3 = bind_update_merge(B11, B2),
    {[G1, G2, G3], B2};
clause_list(_, _, [], ExportList, _, _, Pos) ->
    {[], bind_exports_init(ExportList, Pos)}.

%: +func check_simple_clause :: convert:cclause() 
%:                           -> {value, list(list(pattern()))} + none.
check_simple_clause({clause, _, PairList, _Body})->
    case lists:all(fun({_, Guard}) -> Guard==[] end, PairList) of
	true ->
            HeadList = [Head || {Head, _} <- PairList], 
            {value, HeadList};
        false ->
            none
    end.

%: +func clause :: list(texp0())
%:                      * list(list(pattern()))
%:                      * convert:cclause() 
%:                      * texp0()
%:                      * env()
%:                      * pos()
%:                     -> !{!constraints(), !bind()}.
clause(KeyList, QList,  {clause, Line, PairList, Body}, Res, Env, Pos) ->
    Source = source(env_module(Env), Line, clause),
    HeadList = [Head || {Head, _} <- PairList], 
    case simple_headlist(KeyList, QList, HeadList, Env, pos:left(Pos)) of
        [] ->
	    {[], bind_empty()};        
	L = [{_, B1}|_] -> 
            G = lists:map(fun({G0, _}) -> G0 end, L),
            {T, G2, B2} = body(Body, env_add_binding(Env, B1), pos:right(Pos)),
	    G3 = {include, T, Res, Source}, 
	    {[G, G2, G3], bind_combine(B1, B2)}
    end.

%: +func simple_headlist :: list(texp0())
%:                        * list(list(match:epattern()))
%:                        * list(list(match:epattern()))
%:                        * env()
%:                        * walker:pos() 
%:                       -> list({walker:constraints(), walker:bind()}).
simple_headlist(KeyList, QList, [Head|HeadList], Env, Pos) ->
    case match:match_head(Head, KeyList, QList, pos:left(Pos)) of
	{G, B} -> 
            L = simple_headlist(KeyList, 
                                [Head]++QList, 
                                HeadList, 
                                Env, 
                                pos:right(Pos)),
            [{G, B} | L];
        unmatchable ->
            simple_headlist(KeyList, QList, HeadList, Env, pos:right(Pos))
     end;
simple_headlist(_, _, [], _, _) ->
    [].

%: +func receive_exp :: convert:exp() * env() * pos() 
%:                -> !{!texp0(), !dl(constraint()), !bind()}.
receive_exp({'receive', _Line, ClauseList, ExportList}, Env, Pos) ->
    TList = [{c, conn:any(), []}],
    Res = {var, name_new(pos:left(Pos))},
    {G2, B2} = 
	clause_list(TList, [], ClauseList, ExportList, Res, Env, pos:right(Pos)),
    {Res, G2, B2}.



%: +func exp_list :: list(convert:exp()) * env() * pos()
%:                -> !{list(texp0()), !list(constraints()), !bind()}.
exp_list([], _Env, _Pos) ->
    {[], [], bind_empty()};
exp_list([Exp|ExpList], Env, Pos) ->
    {T, G1, B1} = exp(Exp, Env, pos:left(Pos)),
    Env1 = env_add_binding(Env, B1),
    {TList, G2, B2} = exp_list(ExpList, Env1, pos:right(Pos)),
    {[T|TList], [G1, G2], bind_combine(B1, B2)}.

%: +func type_exp :: convert:texp() * dir() * tenv() * pos() ->
%:		       !{!texp0(), !constraints()}.
type_exp(TE, Dir, Env, Pos) ->
    {T1, G1} = type_exp_1(TE, Dir, Env, pos:left(Pos)),
    {T2, G2} = unextend(T1, Dir, pos:right(Pos)),
    {T2, [G1, G2]}.

type_exp_1({var, '_'}, _Dir, _Env, Pos) ->
    Name = name_new(Pos),
    {{var, Name}, []};
type_exp_1({var, Name}, _Dir, Env, _Pos) ->
    Var = tenv_get(Name, Env),
    {Var, []};
type_exp_1({alias, _Line, {var, Name}, _TE}, upper, Env, _Pos) -> 
    Var = tenv_get(Name, Env),
    {Var, []};
type_exp_1({alias, Line, {var, Name}, TE}, lower, Env, Pos) -> 
    {T, G} = type_exp(TE, lower, Env, Pos),
    Var = tenv_get(Name, Env),
    G0 = {include, Var, T, {alias, Name, Line}},
    {Var, [G, G0]};

type_exp_1({c, {fun_fixed, N}, [Exp| ExpList]}, Dir, Env, Pos) ->
    {TList, G1} = type_exp_list(ExpList, reverse(Dir), Env, pos:left(Pos)),
    {T, G2} = type_exp(Exp, Dir, Env, pos:right(Pos)),
    {{c, conn:fun_fixed(N), [T|TList]}, [G1, G2]};
type_exp_1({c, C, TEList}, Dir, Env, Pos) ->
    {TList, G} = type_exp_list(TEList, Dir, Env, Pos),
    {{c, C, TList}, G};
type_exp_1({number, Line, _E}, lower, _Env, Pos) ->
    X = {var, name_new(Pos)},
    G = {include, {c, conn:integer(), []}, X, {number, Line}},
    {X, [G]};
type_exp_1({number, Line, _E}, upper, _Env, Pos) ->
    X = {var, name_new(Pos)},
    G = {include, X, {c, conn:float(), []}, {number, Line}},
    {X, [G]};

%% type_exp({union, _Line, Exp1, Exp2}, Dir, Env, Pos) ->
%%     Source = source_empty(),
%%     {T1, G1} = type_exp(Exp1, Dir, Env, pos:left(pos:left(Pos))),
%%     X = {var, name_new(pos:right(pos:left(Pos)))},
%%     {T2, G2} = type_exp(Exp2, Dir, Env, pos:right(Pos)),
%%     C = solver:cexp_c(T1), % If this fails, the report should be eloquent
%%     case Dir of
%%         lower ->
%%             G3 = [{include, T1, X, Source},
%%                   {include, T2, X, Source}];            
%%         upper ->
%% 	    F = rfilter:singleton_inverse(C),
%%             G3 = [{include, apply_filter(X, rfilter:singleton(C)), T1, Source},
%%                   {include, apply_filter(X, F), T2, Source}]
%%     end,
%%     {X, [G1, G2, G3]}; % Use function make_union defined below?
type_exp_1({union, _Line, Exp1, Exp2}, Dir, Env, Pos) ->
    %Source = source(tenv_module(Env), Line, union),
    [Pos2, Pos3] = pos:pos_list(2, Pos),
    %X = {var, name_new(Pos1)},
    {T1, G1} = type_exp_1(Exp1, Dir, Env, Pos2),
    {T2, G2} = type_exp_1(Exp2, Dir, Env, Pos3),
    case Dir of             
        lower ->
            %% G3 = [{include, T1, X, Source},
            %%       {include, T2, X, Source}],
            %% {X, [G1, G2, G3]};            
            {{union, T1, T2}, [G1, G2]};
        upper ->
            F  = texp_filter(Exp1),
	    Fi = texp_inverse_filter(Exp1),
            {{union, {trim, T1, F}, {trim, T2, Fi}}, [G1, G2]}
    end;
type_exp_1({any}, _, _, _) ->
    {{c, conn:any(), []}, []};
type_exp_1({none}, _, _, _) ->
    {{var, zero}, []};
type_exp_1({type_call, Line, Tyid, Site, ExpList}, Dir, Env, Pos) ->
    SourceL = source(tenv_module(Env), Line, {type_call_lower, Tyid}),
    SourceU = source(tenv_module(Env), Line, {type_call_upper, Tyid}),
    {TList, G0} = type_exp_list(ExpList, Dir, Env, pos:left(Pos)),
    Res = temp_new(pos:left(pos:right(Pos))),
    Name = name_new(pos:right(pos:right(Pos))),
    TName = tyid_to_predicate_name(Tyid, Dir, Site),
    G1 = {call, TName, [ Name]},
    Arrow = {c, conn:fun_fixed(length(ExpList)), [Res|TList]},
    case Dir of 
        lower ->
            G2 = {include, {var, Name}, Arrow, SourceL};
        upper ->
            G2 = {include, Arrow, {var, Name}, SourceU}
    end,
    {Res, [G0, G1, G2]};
%% type_exp({type_extern_call, Line, Tyid, ExpList}, Dir, Env, Pos) -> % hmm
%%     type_exp({type_call, Line, Tyid, ExpList}, Dir, Env, Pos);

% Note: Watches probably only make sense when the direction is
% upper.  Turn into a no-op when Dir is lower.
type_exp_1({watch, Line, Exp}, Dir=upper, Env, Pos) ->
    Source = source(tenv_module(Env), Line, watch),
    {T, G} = type_exp(Exp, Dir, Env, pos:left(Pos)),
    Name = name_new(pos:right(Pos)),
    Temp = {var, Name},
    Message = io_lib:format("Untriggered watch at line ~w", [Line]),
    {Temp, [G, {include, Temp, T, Source}, {watch, Name, Message}]};
type_exp_1({watch, _Line, Exp}, Dir=lower, Env, Pos) ->
    type_exp(Exp, Dir, Env, Pos).

type_exp_list([], _Dir, _Env, _Pos) ->
    {[], []};
type_exp_list([Exp|ExpList], Dir, Env, Pos) ->    
    [Pos1, Pos2] = pos:pos_list(2, Pos),
    {T, G1} = type_exp(Exp, Dir, Env, Pos1),
    {TList, G2} = type_exp_list(ExpList, Dir, Env, Pos2),
    {[T|TList], [G1, G2]}.

%%

texp_filter({union, _, TExp1, TExp2}) ->
    F1 = texp_filter(TExp1),
    F2 = texp_filter(TExp2),
    rfilter:union(F1, F2);

texp_filter({c, C, _}) ->
    rfilter:singleton(C).


texp_inverse_filter({union, _, TExp1, TExp2}) ->
    F1 = texp_inverse_filter(TExp1),
    F2 = texp_inverse_filter(TExp2),
    rfilter:intersect(F1, F2);

texp_inverse_filter({c, C, _}) ->
    rfilter:singleton_inverse(C).

unextend(E = {union, _, _}, Dir, Pos) ->
    unextend_union(E, Dir, Pos);
unextend(E = {trim, _, _}, Dir, Pos) ->
    unextend_trim(E, Dir, Pos);
unextend(E, _, _) ->
    {E, []}.

unextend_union(E, upper, Pos) ->
    Name = name_new(Pos),
    Temp = {var, Name},
    {Temp, create_include(Temp, E, rfilter:any(), source_empty())};
unextend_union(E, lower, Pos) ->
    Name = name_new(Pos),
    Temp = {var, Name},
    {Temp, create_include(E, Temp, rfilter:any(), source_empty())}.

unextend_trim(E, upper, Pos) ->
    Name = name_new(Pos),
    Temp = {var, Name},
    {Temp, create_include(Temp, E, rfilter:any(), source_empty())};
unextend_trim(E, lower, Pos) ->
    Name = name_new(pos:left(Pos)),
    Temp = {var, Name},
    {Temp, create_include(E, Temp, rfilter:any(), source_empty())}.


%% Free variables of type expressions
%: +func free_variables :: convert:texp() -> list(name()).
free_variables(TE) ->
    L = free_variables0(TE),
    L.

free_variables0(TE) ->
    Set = gb_sets:empty(),
    Set1 = free_variables(TE, Set),
    gb_sets:to_list(Set1).

free_variables({var, zero}, Set) ->
    Set;
free_variables({var, Name}, Set) ->
    gb_sets:add(Name, Set);
free_variables({alias, _, _, TE}, Set) ->
    free_variables(TE, Set);
free_variables({c, _, TEList}, Set) ->
    free_variables_list(TEList, Set);
free_variables({none}, Set) ->
    Set; 
free_variables({any}, Set) ->
    Set; 
free_variables({number, _, TE}, Set) ->
    free_variables(TE, Set);    
free_variables({union, _, TE1, TE2}, Set) ->
    free_variables_list([TE1, TE2], Set);
free_variables({type_call, _, _, _, TEList}, Set) ->
    free_variables_list(TEList, Set);
%% free_variables({type_extern_call, _,  _, TEList}, Set) ->
%%     free_variables_list(TEList, Set);
free_variables({watch, _Line, TE}, Set) ->
    free_variables(TE, Set).

free_variables_list([], Set) ->
    Set;
free_variables_list([TE|TEList], Set) ->
    Set1 = free_variables(TE, Set),
    free_variables_list(TEList, Set1).

%% Alias introduced of type expressions
%: +func alias_introduced :: convert:texp() -> list(name()).
alias_introduced(TE) ->
    Set = gb_sets:empty(),
    Set1 = alias_introduced(TE, Set),
    gb_sets:to_list(Set1).

alias_introduced({var, _}, Set) ->
    Set;
alias_introduced({alias, _, {var, Name}, TE}, Set) ->
    Set1 = alias_introduced(TE, Set),
    gb_sets:add(Name, Set1);
alias_introduced({c, _, TEList}, Set) ->
    alias_introduced_list(TEList, Set);
alias_introduced({none}, Set) ->
    Set; 
alias_introduced({any}, Set) ->
    Set; 
alias_introduced({number, _, TE}, Set) ->
    alias_introduced(TE, Set);    
alias_introduced({union, _, TE1, TE2}, Set) ->
    alias_introduced_list([TE1, TE2], Set);
alias_introduced({type_call, _, _, _, TEList}, Set) ->
    alias_introduced_list(TEList, Set);
%% alias_introduced({type_extern_call, _,  _, TEList}, Set) ->
%%     alias_introduced_list(TEList, Set);
alias_introduced({watch, _Line, TE}, Set) ->
    alias_introduced(TE, Set).

alias_introduced_list([], Set) ->
    Set;
alias_introduced_list([TE|TEList], Set) ->
    Set1 = alias_introduced(TE, Set),
    alias_introduced_list(TEList, Set1).

%% apply filter
%: +func apply_filter :: texp0() * rfilter:filter() -> texp().                           

apply_filter({var, Name}, F) ->
    {sift, Name, F};
apply_filter(E = {c, C, _}, F) ->
    Fc = rfilter:singleton(C),
    case rfilter:included(Fc, F) of
        true ->
             E;
        false -> 
             {var, zero}
    end;
apply_filter({sift, N, F1}, F2) ->
    apply_filter({var, N}, 
		 rfilter:intersect(F1, F2)).

create_include(E0, {union, E1, E2}, F, Source) ->
    [create_include(E0, E1, F, Source),
     create_include(E0, E2, F, Source)];
create_include(E0, {trim, E, F1}, F2, Source) ->
    F = rfilter:intersect(F1, F2),
    create_include(E0, E, F, Source);
create_include({union, E1, E2}, E, F, Source) ->
    [create_include(E1, E, F, Source),
     create_include(E2, E, F, Source)];
create_include({trim, E1, F1}, E2, F2, Source) ->
     F = rfilter:intersect(F1, F2),
    create_include(E1, E2, F, Source);
create_include(E1, E2, F, Source) ->
    {include, apply_filter(E1, F), E2, Source}.



%% Env. Maps Erlang variables to constraint expressions.

%: +func env_new :: atom() -> !env().
env_new(Module) ->
    {env, 0, Module, gb_trees:empty()}.

%: +func env_module :: env() -> !atom().
env_module({env, _, Module, _}) ->
    Module.

%: +func env_insert :: name() * texp0() * env() -> !env().
env_insert(Var, K, {env, N, M, Map}) ->
    Map1 = gb_trees:insert(Var, K, Map),
    {env, N, M, Map1}.

env_get(Var, {env, _, _, Map}) ->
    gb_trees:get(Var, Map).


%: +func env_add_binding :: env() * bind() -> !env().
env_add_binding(Env, B) ->
    env_add_blist(Env, gb_trees:to_list(B)).
    
env_add_blist(Env, []) ->
    Env;
env_add_blist(Env, [{K, V}|B]) ->
    Env1 = env_insert(K, V, Env),
    env_add_blist(Env1, B).

%% Bind
%: +func bind_empty :: () -> !bind().   
bind_empty() ->
    gb_trees:empty().

%: +func bind_singleton :: name() * texp0() -> bind().
bind_singleton(V, T) ->
    gb_trees:insert(V, T, gb_trees:empty()).

%: +func bind_combine :: bind() * bind() -> !bind().			   
bind_combine(B1, B2) ->
   bind_insert_list(gb_trees:to_list(B1), B2).

%: +func bind_insert :: name() * texp0() * bind() -> bind().
bind_insert(V, T, B) ->
    gb_trees:insert(V, T, B).

%: +func bind_insert_list :: list({name(),texp0()}) * bind() -> bind().
bind_insert_list([], B) ->
    B;
bind_insert_list([{K, V}|L], B) ->
    B1 = bind_insert(K, V, B),
    bind_insert_list(L, B1).

bind_insert_arg_list(VarList, B) ->
    KVList = [{V,{var, V}}||{var, V} <-VarList],
    bind_insert_list(KVList, B).   

bind_replace(B, T) ->
    [{V, _}] = gb_trees:to_list(B),
    bind_singleton(V, T).

bind_update_merge(B1, B2) ->
    L1 = gb_trees:to_list(B1),    
    L2 = gb_trees:to_list(B2),
    bind_update_merge_lists(L1, L2, []).

bind_update_merge_lists([], [], G) ->
    G;
bind_update_merge_lists([{K, V1}|L1], [{K, V2}|L2], G) ->                     
    G1 = {include, V1, V2, source_empty()},
    bind_update_merge_lists(L1, L2, [G1|G]).

% 

%: +func bind_exports :: bind() * list(name()) -> !bind().
bind_exports(B, ExportList) ->
  bkv_exports(gb_trees:to_list(B), ExportList, []).

bkv_exports(_, [], Acc) ->
    gb_trees:from_orddict(lists:reverse(Acc));
bkv_exports([{K0, _}|BKV], [K|ExportList], Acc ) when K0 < K ->
    bkv_exports(BKV, [K|ExportList], Acc );
bkv_exports([{K, V}|BKV], [K|ExportList], Acc ) ->
    bkv_exports(BKV, ExportList, [{K, V} | Acc] ).



%

bind_exports_init(ExportList, Pos) ->
    bind_exports_init(ExportList, gb_trees:empty(), Pos).

bind_exports_init([], B, _) ->
    B;
bind_exports_init([Var|L], B, Pos) ->
    Temp = {var, name_new(pos:left(Pos))},
    B1 = gb_trees:insert(Var, Temp, B),
    bind_exports_init(L, B1, pos:right(Pos)).



%% tenv

tenv_new(UnboundList, AbstractList, Module) ->
    EntryList = [{Name, {var, Name}} || 
		    Name <- UnboundList]
        ++
    		[case Name of 
                      {anonymous, _} ->
                         {Name, {var, Name}};
                       _ ->  
                         {Name, {c, conn:abstract(Name), []}}
                 end
                 || Name <- AbstractList],
    {tenv, Module, from_list(EntryList)}.


tenv_new(Module) ->
    {tenv, Module, gb_trees:empty()}.

from_list(L) ->
    from_list(L, gb_trees:empty()).

from_list([], Map) ->
    Map;
from_list([{Key, Val}|L], Map) ->
    from_list(L, gb_trees:insert(Key, Val, Map)).

tenv_get(zero, {tenv, _, _TEnv}) ->
    {var, zero};
tenv_get(Var, {tenv, _, TEnv}) ->
    gb_trees:get(Var, TEnv).

tenv_module({tenv, M, _}) ->
    M.

%

name_new(P) ->
    list_to_atom(lists:flatten(io_lib:format("w~w", [P]))).

temp_new(P) ->
    {var, name_new(P)}.

temp_new([], _) ->
    [];
temp_new([_|L], Pos) ->
    [temp_new(pos:left(Pos)) | temp_new(L, pos:right(Pos))].


% Theory

%: +func theory_new :: () -> !theory().			 
theory_new() ->
    gb_trees:empty().

%: +func theory_insert :: predicate() * theory() -> !theory().
theory_insert({predicate, Key, ArgList, G}, Theory) ->
    Predicate = {predicate, Key, ArgList, flatten(G)},
    gb_trees:insert(Key, Predicate, Theory).


flatten(Body) ->
    flatten(Body, []).

flatten([], Acc) ->
    Acc;
flatten([Body1|Body2], Acc) ->
    flatten(Body1, flatten(Body2, Acc));
flatten(Phi, Acc) ->
    [Phi|Acc].


%: +func theory_insert_list :: list(predicate()) * theory() -> !theory().
theory_insert_list([], Theory) ->
    Theory;
theory_insert_list([P|L], Theory) ->
    Theory1 = theory_insert(P, Theory),
    theory_insert_list(L, Theory1).

%% Dir

%: +func reverse :: dir() -> !dir().
reverse(upper) ->
    lower;
reverse(lower) ->
    upper.    


%% assert

assert(true) ->
    true.

assert(true, _) ->
    true.


%% source

%: +func source :: atom() * line() * any() -> !solver:source().		     
source(Module, Line, Activity) ->
    {Module, Line, Activity}.

%: +func source_empty :: () -> !solver:source().		     
source_empty() ->
    [].

%% Numbers

%% integer_l(Pos) ->
%%     {c, integer, [{var, {anonymous, Pos}}]}.

float_l(_) ->
%    {c, integer, [{c, float, []}]}.
    {c, conn:float(), []}.


integer_u(_) ->
%    {c, integer, [{var, zero}]}.
    {c, conn:integer(), []}.


%% float_u(Pos) ->
%%     {c, integer, [{c, float, []}]}.


%%%% Predicate names...
%: +func fnid_to_predicate_name :: fnid()
%:                               * (atom() + {atom(), atom()})
%:                              -> pname().
fnid_to_predicate_name(Fnid, Mode) ->
    list_to_atom(lists:flatten([fnid_to_string(Fnid),"#",fmode_to_string(Mode)])).

fnid_to_string({fnid, Module, Name, Arity}) ->
    io_lib:format("~w:~w/~w", [Module, Name, Arity]).


fmode_to_string(top) ->
    "t";
fmode_to_string(def) ->
    "d";
fmode_to_string({spec, lower}) ->
    "l";
fmode_to_string({spec, upper}) ->
    "u";        
fmode_to_string(callee) ->
    "c".

tyid_to_predicate_name(Tyid, Mode, Site) ->
    list_to_atom(lists:flatten([tyid_to_string(Tyid),
                                "#",
                                tmode_to_string(Mode),
                                site_to_string(Site)])).

tyid_to_string({tyid, Module, Name}) ->
    io_lib:format("~w:~w", [Module, Name]).

tmode_to_string(lower) ->
    "l";
tmode_to_string(upper) ->
    "u".

site_to_string(local) ->
    "h";
site_to_string(extern) ->
    "e".

% Name
%: +func name_to_string :: name() -> erlang:string().
name_to_string(Name) when is_atom(Name) ->
    atom_to_list(Name);
name_to_string(Name) when is_integer(Name) ->
    integer_to_list(Name).

